/********************************************************************************
*                                                                               *
*                     A p p l i c a t i o n   O b j e c t                       *
*                                                                               *
*********************************************************************************
* Copyright (C) 1997 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXApp.h,v 1.79 1999/01/05 20:40:00 jeroen Exp $                          *
********************************************************************************/
#ifndef FXAPP_H
#define FXAPP_H


// Forward declarations
class FXWindow;
class FXIcon;
class FXBitmap;
class FXCursor;
class FXRootWindow;
class FXMainWindow;
class FXFont;
class FXDC;
class FXWDC;

// Opaque FOX objects
struct FXTimer;
struct FXChore;
struct FXRepaint;
struct FXInput;


// Application Object
class FXApp : public FXObject {
  FXDECLARE(FXApp)
    
  // We've got many friends
  friend class FXId;
  friend class FXBitmap;
  friend class FXImage;
  friend class FXIcon;
  friend class FXCursor;
  friend class FXDrawable;
  friend class FXWindow;
  friend class FXRootWindow;
  friend class FXShell;
  friend class FXTopWindow;
  friend class FXMainWindow;
  friend class FXGLCanvas;
  friend class FXSlider;
  friend class FXScrollbar;
  friend class FXFont;
  friend class FXDC;
  friend class FXWDC;
  
private:
  
  // Platform independent private data
  const FXchar *appname;
  FXRegistry    registry;               // Application setting registry
  FXDC         *windowdc;               // Global device context for windows
  FXDC         *bitmapdc;               // Global device context for 1-bit bitmaps
  FXDC         *nulldc;                 // Null device context
  FXWindow     *mainWindow;             // Main window
  FXWindow     *focusWindow;            // Window which has the focus
  FXWindow     *cursorWindow;           // Window under the cursor
  FXWindow     *grabWindow;             // Window which has the grab
  FXWindow     *refresher;              // GUI refresher pointer
  FXRootWindow *root;                   // Root window
  FXTimer      *timers;                 // List of timers, sorted by time
  FXChore      *chores;                 // List of chores
  FXRepaint    *repaints;               // Unhandled repaint rectangles
  FXTimer      *timerrecs;              // List of recycled timer records
  FXChore      *chorerecs;              // List of recycled chore records
  FXRepaint    *repaintrecs;            // List of recycled repaint records
  FXInvocation *invocation;             // Modal loop invocation
  FXInput      *inputs;                 // Input file descriptors being watched
  FXint         ninputs;                // Number of inputs
  FXint         maxinput;               // Maximum input number
  void         *r_fds;                  // Set of file descriptors for read
  void         *w_fds;                  // Set of file descriptors for write
  void         *e_fds;                  // Set of file descriptors for exceptions
  FXuint        done;                   // True if application is done
  FXint         exitcode;               // Exit code
  FXEvent       event;                  // Event
  
private:
  
  // Global data
  static FXApp *app;                    // Application pointer
  
private:

  // Platform dependent private data
#ifndef FX_NATIVE_WIN32
  Display      *display;                // Display we're talking to
  int           connection;             // Connection fd of display
  int           wcontext;               // Window hash context
  Visual       *visual;                 // Application visual
  FXuint        depth;                  // Visual depth
  FXID          colormap;               // Application colormap
  FXID          wmDeleteWindow;         // Catch delete window 
  FXID          wmQuitApp;              // Catch quit application
  FXID          wmProtocols;            // Window manager protocols
  FXID          wmMotifHints;           // Motif hints
  FXID          wmTakeFocus;            // Focus explicitly set by app
  FXID          cbSelection;            // Clipboard selection
  FXID          xdndAware;              // XDND awareness atom
  FXID          xdndEnter;              // XDND enter window message
  FXID          xdndLeave;              // XDND leave window message
  FXID          xdndPosition;           // XDND position update message
  FXID          xdndStatus;             // XDND status feedback message
  FXID          xdndDrop;               // XDND drop message
  FXID          xdndFinished;           // XDND finished message
  FXID          xdndSelection;          // XDND selection atom
  FXID          xdndActionMove;         // XDND Move action
  FXID          xdndActionCopy;         // XDND Copy action
  FXID          xdndActionLink;         // XDND Link action
  FXID          xdndTypes;              // XDND types list atom
  FXID          xdndSource;             // XDND drag source window
  FXID          xdndTarget;             // XDND drop target window
  FXID          xdndAction;             // XDND Drag and Drop action
  FXbool        xdndAccepts;            // XDND true if target accepts
  FXbool        xdndSendPosition;       // XDND send position update when status comes in
  FXbool        xdndStatusPending;      // XDND waiting for status feedback
  FXbool        xdndFinishPending;      // XDND waiting for drop-confirmation
  FXbool        xdndStatusReceived;     // XDND received at least one status
  FXbool        xdndWantUpdates;        // XDND target wants new positions while in rect
  FXuint        xdndVersion;            // XDND version for ongoing transaction
  FXID         *xdndTypeList;           // XDND type list
  FXuint        xdndNumTypes;           // XDND number of types
  FXRectangle   xdndRect;               // XDND rectangle bounding target
  FXint         xdndXPos;               // XDND Cached X position
  FXint         xdndYPos;               // XDND Cached Y position
  FXID          ddeAtom;                // DDE Exchange Atom
  FXID          ddeDelete;              // DDE Delete Target Atom
  FXuchar      *ddeData;                // DDE array
  FXuint        ddeSize;                // DDE array size
  FXID          blackstipple;           // Black stipple pattern
  FXbool        shmi;                   // Use XSHM Image possible
  FXbool        shmp;                   // Use XSHM Pixmap possible
  FXbool        synchronize;            // Synchronized
  FXPalEntry   *palette;                // Palette of colors
  FXuint        ncolors;                // Number of colors in palette
#else
  DWORD		idInst;			// DDE instance handle
  UINT		xdndEnter;		// XDND enter window message
  UINT		xdndLeave;		// XDND leave window message
  UINT		xdndPosition;		// XDND position update message
  UINT		xdndStatus;		// XDND status feedback message
  UINT		xdndDrop;		// XDND drop message
  UINT		xdndFinished;		// XDND finished message
  FXushort      xdndActionMove;         // XDND Move action
  FXushort      xdndActionCopy;         // XDND Copy action
  FXushort      xdndActionLink;         // XDND Link action
  FXID          ddeAtom;                // DDE Exchange Atom
  UINT          ddeDelete;              // DDE Delete Target Atom
  FXuchar      *ddeData;                // DDE array
  FXuint        ddeSize;                // DDE array size
#endif
  
#ifdef FX_NATIVE_WIN32
public:
  // This is public so we can set it globally and get it over with
  static HINSTANCE hInstance;
#endif

public:
  
  // Public platform independent data
  FXFont       *normalFont;             // Normal font
  FXCursor     *arrowCursor;            // Arrow cursor
  FXCursor     *rarrowCursor;           // Reverse arrow cursor
  FXCursor     *textCursor;             // Text cursor
  FXCursor     *hsplitCursor;           // Horizontal split cursor
  FXCursor     *vsplitCursor;           // Vertical split cursor
  FXCursor     *resizeCursor;           // Resize grip cursor
  FXCursor     *swatchCursor;           // Color swatch drag cursor
  FXCursor     *dontdropCursor;         // Cursor indicating no drop
  FXCursor     *moveCursor;             // Move cursor
  FXCursor     *dragHCursor;            // Resize horizontal edge
  FXCursor     *dragVCursor;            // Resize vertical edge
  FXCursor     *dragTLCursor;           // Resize upper-left or bottom-right corner
  FXCursor     *dragTRCursor;           // Resize upper-left or bottom-right corner
  FXCursor     *dndCopyCursor;          // Drag and drop copy
  FXCursor     *dndMoveCursor;          // Drag and drop move
  FXCursor     *dndLinkCursor;          // Drag and drop link
  FXBitmap     *grayBitmap;             // Gray bitmap
  FXuint        clickSpeed;             // Double click speed
  FXuint        scrollSpeed;            // Scroll speed
  FXuint        blinkSpeed;             // Cursor blink speed
  FXuint        menuPause;              // Menu popup delay
  FXuint        tooltipPause;           // Tooltip popup delay
  FXuint        tooltipTime;            // Tooltip display time
  FXint         scrollbarWidth;         // Scrollbar width
  FXint         dragDelta;              // Minimum distance considered a move
  FXColor       backColor;              // Background color
  FXColor       foreColor;              // Foreground color
  FXColor       hiliteColor;            // Hightlight color
  FXColor       shadowColor;            // Shadow color
  FXColor       borderColor;            // Border color
  FXColor       selforeColor;           // Select foreground color
  FXColor       selbackColor;           // Select background color
  
public:

  static const FXuchar version[3];      // Version number


private:
  
  // Private member functions
  FXApp(const FXApp&){}

#ifndef FX_NATIVE_WIN32
  void addRepaint(FXID win,FXint x,FXint y,FXint w,FXint h,FXbool synth=0);
#endif
  
  // Event handling
#ifndef FX_NATIVE_WIN32
  void getNextEvent(XEvent& ev);
  void dispatchEvent(XEvent& ev);
#else
  void getNextEvent(MSG& msg);
  void dispatchEvent(MSG& msg);
  static LRESULT CALLBACK wndproc(HWND hwnd,UINT iMsg,WPARAM wParam,LPARAM lParam);
  LRESULT dispatchEvent(HWND hwnd,UINT iMsg,WPARAM wParam,LPARAM lParam);
#endif

  // Timers for MS-Windows port
#ifdef FX_NATIVE_WIN32
  static void CALLBACK TimerProc(HWND hwnd,UINT uMsg,UINT idEvent,DWORD dwTime);
  void timerProc(UINT idEvent,DWORD dwTime); 
#endif
  
  // Color handling
#ifndef FX_NATIVE_WIN32
  void makePalette();
  void getPalette();
  void freePalette();
#endif
  FXPixel matchColor(FXuint r,FXuint g,FXuint b);
  
  // Setup/tear down stuff
#ifndef FX_NATIVE_WIN32
  int openDisplay(const char* dpyname);
  void closeDisplay();
#else
  void registerWindowClasses();
  static HDDEDATA CALLBACK DdeCallback(UINT iType,UINT iFmt,HCONV hConv,HSZ hsz1,HSZ hsz2,HDDEDATA hData,DWORD dwData1,DWORD dwData2);
  HDDEDATA ddeCallback(UINT iType,UINT iFmt,HCONV hConv,HSZ hsz1,HSZ hsz2,HDDEDATA hData,DWORD dwData1,DWORD dwData2);
#endif

public:
  
  // Message handlers
  long onCmdQuit(FXObject*,FXSelector,void*);

public:
  
  // Messages the application understands
  enum{
    ID_QUIT=0,
    ID_LAST
    };

public:

  // Construct application
  FXApp(const FXchar *name="Application",const FXchar *vendor="Default");

  // Get root Window
  FXRootWindow* getRoot() const { return root; }

  // Get the window under the cursor, if any  
  FXWindow *getCursorWindow() const { return cursorWindow; }

  // Get the window which has the focus, if any  
  FXWindow *getFocusWindow() const { return focusWindow; }

  // Get main window, if any
  FXWindow *getMainWindow() const { return mainWindow; }

  // Add timeout
  FXTimer* addTimeout(FXint ms,FXObject* tgt,FXSelector sel);

  // Remove timeout, returns NULL
  FXTimer* removeTimeout(FXTimer *t);
  
  // Add a chore
  FXChore* addChore(FXObject* tgt,FXSelector sel);
  
  // Remove chore
  FXChore* removeChore(FXChore *c);
  
  // Add input
  void addInput(FXint fd,FXuint mode,FXObject *tgt,FXSelector sel);
  
  // Remove input
  void removeInput(FXint fd,FXuint mode);
  
  // Create application's windows
  void create();

  // Destroy application's windows
  void destroy();

  // Peek for event
  FXbool peekEvent();
  
  // Main application event loop
  FXint run();

  // Break out of main event loop
  void stop(FXint code=0);
    
  // Run till some flag becomes non-zero
  void runUntil(FXuint& condition);
  
  // Run modal for window
  FXint runModalFor(FXWindow* window);

  // Break out of event loop
  void stopModal(FXint code);

  // Run modal while window is shown, or until stopModal is called
  FXint runModalWhileShown(FXWindow* window);
      
  // True if the window is modal
  FXbool isModal(FXWindow* window) const;
  
  // Force GUI refresh
  void forceRefresh();
  
  // Schedule a refresh
  void refresh();

  // Flush pending repaints
  void flush();

  // Initialize application
  void init(int& argc,char** argv);
  
  // Exit application
  void exit(FXint code);

  // Get registry
  FXRegistry& reg(){ return registry; }

  // Register new DND type
  FXDragType registerDragType(const FXString& name) const;
  
  // Get drag type name
  FXString getDragTypeName(FXDragType type) const;
  
  // Show a modal error message
  FXuint showModalErrorBox(FXuint opts,const char* title,const char* message,...);

  // Show a modal warning message
  FXuint showModalWarningBox(FXuint opts,const char* title,const char* message,...);

  // Show a modal question dialog
  FXuint showModalQuestionBox(FXuint opts,const char* title,const char* message,...);

  // Show a modal information dialog
  FXuint showModalInformationBox(FXuint opts,const char* title,const char* message,...);

  // Beep
  void beep();

  // Return application instance
  static inline FXApp* instance(){ return app; }

  // Save
  virtual void save(FXStream& store) const;
  
  // Load
  virtual void load(FXStream& store);
  
  // Virtual destructor
  virtual ~FXApp();
  };


#endif
