/********************************************************************************
*                                                                               *
*                       C o m b o   B o x   O b j e c t                         *
*                                                                               *
*********************************************************************************
* Copyright (C) 1997 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXComboBox.h,v 1.5 1999/01/12 04:17:04 jeroen Exp $                      *
********************************************************************************/
#ifndef FXCOMBOBOX_H
#define FXCOMBOBOX_H



///// WARNING:- API STILL IN FLUX //////

// Combobox
class FXComboBox : public FXPacker {
  FXDECLARE(FXComboBox)
protected:
  FXTextField   *text;
  FXArrowButton *button;
  FXList        *list;
  FXPopup       *pane;
protected:
  FXComboBox(){}
  FXComboBox(const FXComboBox&){}
  virtual void layout();
public:
  enum{
    ID_BUTTON=FXPacker::ID_LAST,
    ID_LIST,
    ID_TEXT,
    ID_LAST
    };
public:
  long onLeftBtnPress(FXObject*,FXSelector,void*);
  long onLeftBtnRelease(FXObject*,FXSelector,void*);
  long onMotion(FXObject*,FXSelector,void*);
  long onCmdPost(FXObject*,FXSelector,void*);
  long onCmdUnpost(FXObject*,FXSelector,void*);
  long onButtonDown(FXObject*,FXSelector,void*);
  long onButtonUp(FXObject*,FXSelector,void*);
  long onCmdText(FXObject*,FXSelector,void*);
  long onListLeftBtnPress(FXObject*,FXSelector,void*);
  long onListLeftBtnRelease(FXObject*,FXSelector,void*);
  long onListClicked(FXObject*,FXSelector,void*);
  long onListCommand(FXObject*,FXSelector,void*);
  long onListSelected(FXObject*,FXSelector,void*);
  long onListDeselected(FXObject*,FXSelector,void*);
public:
  FXComboBox(FXComposite *p,FXint cols,FXObject* tgt=NULL,FXSelector sel=0,FXuint opts=0,FXint x=0,FXint y=0,FXint w=0,FXint h=0,FXint pl=DEFAULT_PAD,FXint pr=DEFAULT_PAD,FXint pt=DEFAULT_PAD,FXint pb=DEFAULT_PAD);
  virtual void create();
  virtual void destroy();
  virtual FXint getDefaultWidth();
  virtual FXint getDefaultHeight();
  virtual FXbool contains(FXint parentx,FXint parenty) const;
  void setCursorPos(FXint pos);
  FXint getCursorPos() const;
  void setAnchorPos(FXint pos);
  FXint getAnchorPos() const;
  void setText(const FXString& txt);
  FXString getText() const;
  void setNumColumns(FXuint cols);
  FXuint getNumColumns() const;
  void killSelection();
  FXint getNumItems() const;
  FXint getNumSelectedItems() const;
  FXListItem** getSelectedItems() const;
  FXListItem* getFirstItem() const;
  FXListItem* getLastItem() const;
  FXListItem* getNextItem(const FXListItem* item) const;
  FXListItem* getPrevItem(const FXListItem* item) const;
  FXListItem* addItemFirst(const FXString& txt,void* ptr=NULL);
  FXListItem* addItemLast(const FXString& txt,void* ptr=NULL);
  FXListItem* addItemAfter(FXListItem* other,const FXString& txt,void* ptr=NULL);
  FXListItem* addItemBefore(FXListItem* other,const FXString& txt,void* ptr=NULL); 
  void removeItem(FXListItem* item);
  void removeItems(FXListItem* fm,FXListItem* to);
  void removeAllItems();
  FXListItem* getItemAt(FXint x,FXint y) const;
  FXint getItemX(const FXListItem* item) const;
  FXint getItemY(const FXListItem* item) const;
  FXint getItemWidth(const FXListItem* item) const;
  FXint getItemHeight(const FXListItem* item) const;
  void updateItem(FXListItem* item);
  void makeItemVisible(FXListItem* item);
  FXbool isItemSelected(const FXListItem* item) const;
  FXbool isItemCurrent(const FXListItem* item) const;
  FXbool isItemVisible(const FXListItem* item) const;
  void markItems();
  FXbool selectItemRange(FXListItem* beg,FXListItem* end);
  FXbool deselectItemRange(FXListItem* beg,FXListItem* end);
  FXbool toggleItemRange(FXListItem* beg,FXListItem* end);
  FXbool restoreItemRange(FXListItem* beg,FXListItem* end);
  FXbool selectItem(FXListItem* item);
  FXbool deselectItem(FXListItem* item);
  FXbool toggleItem(FXListItem* item);
  FXbool restoreItem(FXListItem* item);
  void sortItems();
  void setCurrentItem(FXListItem* item);
  FXListItem* getCurrentItem() const;
  void setAnchorItem(FXListItem* item);
  FXListItem* getAnchorItem() const;
  FXbool extendSelection(FXListItem* item);
  void setItemText(FXListItem* item,const FXString& text);
  FXString getItemText(const FXListItem* item) const;
  void setItemData(FXListItem* item,void* ptr) const;
  void* getItemData(const FXListItem* item) const;
  FXItemSortFunc getSortFunc() const;
  void setSortFunc(FXItemSortFunc func);
  virtual ~FXComboBox();
  };


#endif
