/********************************************************************************
*                                                                               *
*                     D i r e c t o r y   L i s t   O b j e c t                 *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXDirList.h,v 1.22 1999/01/07 07:15:20 jeroen Exp $                      *
********************************************************************************/
#ifndef FXDIRLIST_H
#define FXDIRLIST_H


// Directory List options
enum FXDirListStyle {
  DIRLIST_SHOWFILES     = 0x08000000,           // Show files as well as directories
  DIRLIST_SHOWHIDDEN    = 0x10000000            // Show hidden files or directories
  };


struct FXDirItem;

class FXFileDict;


// Generic tab item
class FXDirList : public FXTreeList {
  FXDECLARE(FXDirList)
protected:
  FXTimer    *refresh;                // Refresh timer
  FXFileDict *associations;           // Association table
  FXGIFIcon  *closed_folder;          // Closed folder icon
  FXGIFIcon  *open_folder;            // Open folder icon
  FXGIFIcon  *mini_doc;               // Document icon
  FXGIFIcon  *mini_app;               // Application icon
  FXuint      counter;                // Refresh counter
protected:
  FXDirList();
  FXDirList(const FXTreeList&){}
  virtual FXTreeItem* createItem();
  virtual void deleteItem(FXTreeItem*);
  FXchar *getpath(const FXTreeItem* item,FXchar* pathname) const;
  FXbool scanRootDir(FXbool relist);
  FXbool scanSubDir(FXDirItem *par,FXchar *pathname,FXbool relist);
  FXbool listSubDir(FXDirItem *par,FXchar *pathname);
  FXTreeItem *getNamedChild(FXTreeItem* par,const FXchar* name,FXuint len) const;
public:
  long onRefresh(FXObject*,FXSelector,void*);
  long onSelectionRequest(FXObject*,FXSelector,void* ptr);
  long onItemExpanded(FXObject*,FXSelector,void*);
  long onItemCollapsed(FXObject*,FXSelector,void*);
  long onCmdSetValue(FXObject*,FXSelector,void*);
  long onCmdSetStringValue(FXObject*,FXSelector,void*);
  long onCmdGetStringValue(FXObject*,FXSelector,void*);
  long onCmdToggleHidden(FXObject*,FXSelector,void*);
  long onUpdToggleHidden(FXObject*,FXSelector,void*);
  long onCmdShowHidden(FXObject*,FXSelector,void*);
  long onUpdShowHidden(FXObject*,FXSelector,void*);
  long onCmdHideHidden(FXObject*,FXSelector,void*);
  long onUpdHideHidden(FXObject*,FXSelector,void*);
  long onCmdToggleFiles(FXObject*,FXSelector,void*);
  long onUpdToggleFiles(FXObject*,FXSelector,void*);
  long onCmdShowFiles(FXObject*,FXSelector,void*);
  long onUpdShowFiles(FXObject*,FXSelector,void*);
  long onCmdHideFiles(FXObject*,FXSelector,void*);
  long onUpdHideFiles(FXObject*,FXSelector,void*);
public:
  static FXbool cmpName(const FXItem* a,const FXItem* b);
public:
  enum {
    ID_REFRESH=FXTreeList::ID_LAST,
    ID_SHOW_FILES,
    ID_HIDE_FILES,
    ID_TOGGLE_FILES,
    ID_SHOW_HIDDEN,
    ID_HIDE_HIDDEN,
    ID_TOGGLE_HIDDEN,
    ID_LAST
    };
public:
  FXDirList(FXComposite *p,FXObject* tgt=NULL,FXSelector sel=0,FXuint opts=0,FXint x=0,FXint y=0,FXint w=0,FXint h=0);
  virtual void create();
  virtual void destroy();
  // Will change
  FXchar* itempath(FXDirItem* item,FXchar* pathname);
  FXbool isItemDirectory(const FXTreeItem* item) const;
  FXbool isItemFile(const FXTreeItem* item) const;
  FXbool isItemExecutable(const FXTreeItem* item) const;
  FXString getItemPathname(const FXTreeItem* item) const;
  FXTreeItem* getPathnameItem(const FXString& file) const;
  FXbool showFiles() const;
  void showFiles(FXbool shown);
  FXbool showHiddenFiles() const;
  void showHiddenFiles(FXbool shown);
  void setAssociations(FXFileDict* assoc);
  FXFileDict* getAssociations() const { return associations; }
  virtual void save(FXStream& store) const;
  virtual void load(FXStream& store);
  virtual ~FXDirList();
  };


#endif
