/********************************************************************************
*                                                                               *
*                               F o n t   O b j e c t                           *
*                                                                               *
*********************************************************************************
* Copyright (C) 1997 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXFont.h,v 1.11 1999/01/12 04:17:04 jeroen Exp $                         *
********************************************************************************/
#ifndef FXFONT_H
#define FXFONT_H


class FXFont : public FXId {
  FXDECLARE(FXFont)
protected:
  FXString      name;       // Name of the font
  void         *font;       // Info about the font
protected:
  FXFont();
  FXFont(const FXFont&){}
public:
  
  // Construct a font with given X font string
  // Note: this will be changed some day 
  FXFont(FXApp* a,const FXString& nm);

  // Create the font
  virtual void create();
  
  // Destroy the font
  virtual void destroy();
  
  // Find out if the font is monotype or proportional
  FXbool isFontMono() const;
  
  // Width of widest character in font
  FXint getFontWidth() const;
  
  // Height of highest character in font
  FXint getFontHeight() const;
  
  // Ascent from baseline
  FXint getFontAscent() const;
  
  // Descent from baseline
  FXint getFontDescent() const;
  
  // Calculate width of given text in this font
  FXint getTextWidth(const FXchar *text,FXuint n) const;
  
  // Calculate height of given text in this font
  FXint getTextHeight(const FXchar *text,FXuint n) const;
  
  // Save font data into stream
  virtual void save(FXStream& store) const;
  
  // Load font data from stream
  virtual void load(FXStream& store);
  
  // Destruct font
  virtual ~FXFont();
  };
  
    
#endif
