/********************************************************************************
*                                                                               *
*                             M e n u   O b j e c t s                           *
*                                                                               *
*********************************************************************************
* Copyright (C) 1997 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXMenu.h,v 1.30 1999/01/12 04:17:04 jeroen Exp $                         *
********************************************************************************/
#ifndef FXMENU_H
#define FXMENU_H


// Button masks for menu buttons
enum FXMenuStyle {
  MENU_DEFAULT  = 0x00008000,
  MENU_AUTOGRAY = 0x00010000,
  MENU_AUTOHIDE = 0x00020000
  };


class FXMenuPane;
class FXMDIButton;
class FXMenuButton;


// Menu item separator
class FXMenuSeparator : public FXFrame {
  FXDECLARE(FXMenuSeparator)
protected:
  FXMenuSeparator(){}
  FXMenuSeparator(const FXMenuSeparator&){}
public:
  long onPaint(FXObject*,FXSelector,void*);
public:
  FXMenuSeparator(FXComposite* p,FXuint opts=0);
  virtual FXint getDefaultWidth();
  virtual FXint getDefaultHeight();
  };


// Text menu item
class FXMenuCaption : public FXFrame {
  FXDECLARE(FXMenuCaption)
protected:
  FXString     label;
  FXString     accel;
  FXString     help;
  FXIcon      *icon;
  FXFont      *font;
  FXint        hotoff;
  FXHotKey     hotkey;
  FXColor      textColor;
  FXColor      selbackColor;
  FXColor      seltextColor;
protected:
  FXMenuCaption(){}
  FXMenuCaption(const FXMenuCaption&){}
public:
  long onPaint(FXObject*,FXSelector,void*);
  long onQueryHelp(FXObject*,FXSelector,void*);
  long onUpdate(FXObject*,FXSelector,void*);
public:
  FXMenuCaption(FXComposite* p,const FXString& text,FXIcon* ic=NULL,FXuint opts=0);
  virtual void create();
  virtual FXint getDefaultWidth();
  virtual FXint getDefaultHeight();
  void setText(const FXString& text);
  FXString getText() const { return label; }
  void setIcon(FXIcon* ic);
  FXIcon* getIcon() const { return icon; }
  void setFont(FXFont* fnt);
  FXFont* getFont() const { return font; }
  FXColor getTextColor() const { return textColor; }
  void setTextColor(FXColor clr);
  FXColor getSelBackColor() const { return selbackColor; }
  void setSelBackColor(FXColor clr);
  FXColor getSelTextColor() const { return seltextColor; }
  void setSelTextColor(FXColor clr);
  void setAccelText(const FXString& text);
  FXString getAccelText() const { return accel; }
  void setHelpText(const FXString& text);
  FXString getHelpText() const { return help; }
  virtual void save(FXStream& store) const;
  virtual void load(FXStream& store);
  virtual ~FXMenuCaption();
  };


// Command menu item
class FXMenuCommand : public FXMenuCaption {
  FXDECLARE(FXMenuCommand)
protected:
  FXMenuCommand(){}
  FXMenuCommand(const FXMenuCommand&){}
  void drawCheck(FXint x,FXint y);
  void drawBullit(FXint x,FXint y);
public:
  long onPaint(FXObject*,FXSelector,void*);
  long onEnter(FXObject*,FXSelector,void*);
  long onLeave(FXObject*,FXSelector,void*);
  long onLeftBtnPress(FXObject*,FXSelector,void*);
  long onLeftBtnRelease(FXObject*,FXSelector,void*);
  long onActivate(FXObject*,FXSelector,void*);
  long onDeactivate(FXObject*,FXSelector,void*);
  long onCheck(FXObject*,FXSelector,void*);
  long onUncheck(FXObject*,FXSelector,void*);
public:
  FXMenuCommand(FXComposite* p,const FXString& text,FXIcon* ic=NULL,FXObject* tgt=NULL,FXSelector sel=0,FXuint opts=0);
  virtual void create();
  virtual FXbool canFocus() const;
  virtual void setFocus();
  virtual void killFocus();
  void setDefault();
  void setOther();
  FXint isDefault() const;
  virtual void enable();
  virtual void disable();
  void check();
  void uncheck();
  FXint isChecked() const;
  void checkRadio();
  void uncheckRadio();
  FXint isRadioChecked() const;
  };


// Cascade menu item
class FXMenuCascade : public FXMenuCaption {
  FXDECLARE(FXMenuCascade)
protected:
  FXMenuPane *pane;
  FXTimer*    timer;
protected:
  FXMenuCascade(){}
  FXMenuCascade(const FXMenuCascade&){}
  void drawTriangle(FXint l,FXint t,FXint r,FXint b);
public:
  long onPaint(FXObject*,FXSelector,void*);
  long onEnter(FXObject*,FXSelector,void*);
  long onLeave(FXObject*,FXSelector,void*);
  long onTimeout(FXObject*,FXSelector,void*);
  long onFocusRight(FXObject*,FXSelector,void*);
  long onFocusLeft(FXObject*,FXSelector,void*);
  long onLeftBtnPress(FXObject*,FXSelector,void*);
  long onLeftBtnRelease(FXObject*,FXSelector,void*);
  long onKeyPress(FXObject*,FXSelector,void*);
  long onKeyRelease(FXObject*,FXSelector,void*);
  long onActivate(FXObject*,FXSelector,void*);
  long onDeactivate(FXObject*,FXSelector,void*);
public:
  FXMenuCascade(FXComposite* p,const FXString& text,FXIcon* ic=NULL,FXMenuPane* pup=NULL,FXuint opts=0);
  virtual void create();
  virtual void destroy();
  virtual FXbool canFocus() const;
  virtual void setFocus();
  virtual void killFocus();
  void setPopup(FXMenuPane *pup){ pane = pup; }
  FXMenuPane* getPopup() const { return pane; }
  virtual FXbool contains(FXint parentx,FXint parenty) const;
  virtual void save(FXStream& store) const;
  virtual void load(FXStream& store);
  virtual ~FXMenuCascade();
  };


// Menu title button
class FXMenuTitle : public FXMenuCaption {
  FXDECLARE(FXMenuTitle)
protected:
  FXMenuPane *pane;
protected:
  FXMenuTitle(){}
  FXMenuTitle(const FXMenuTitle&){}
public:
  long onPaint(FXObject*,FXSelector,void*);
  long onEnter(FXObject*,FXSelector,void*);
  long onLeave(FXObject*,FXSelector,void*);
  long onLeftBtnPress(FXObject*,FXSelector,void*);
  long onLeftBtnRelease(FXObject*,FXSelector,void*);
  long onKeyPress(FXObject*,FXSelector,void*);
  long onKeyRelease(FXObject*,FXSelector,void*);
  long onActivate(FXObject*,FXSelector,void*);
  long onDeactivate(FXObject*,FXSelector,void*);
  long onFocusUp(FXObject*,FXSelector,void*);
  long onFocusDown(FXObject*,FXSelector,void*);
  long onFocusIn(FXObject*,FXSelector,void*);
  long onFocusOut(FXObject*,FXSelector,void*);
  long onCmdPost(FXObject*,FXSelector,void*);
  long onCmdUnpost(FXObject*,FXSelector,void*);
public:
  FXMenuTitle(FXComposite* p,const FXString& text,FXIcon* ic=NULL,FXMenuPane* pup=NULL,FXuint opts=LAYOUT_LEFT);
  virtual void create();
  virtual FXbool canFocus() const;
  virtual void setFocus();
  virtual void killFocus();
  void setPopup(FXMenuPane *pup){ pane = pup; }
  FXMenuPane* getPopup() const { return pane; }
  virtual FXint getDefaultWidth();
  virtual FXint getDefaultHeight();
  virtual FXbool contains(FXint parentx,FXint parenty) const;
  virtual void save(FXStream& store) const;
  virtual void load(FXStream& store);
  virtual ~FXMenuTitle();
  };


// Popup menu pane
class FXMenuPane : public FXPopup {
  FXDECLARE(FXMenuPane)
protected:
  FXMenuPane(){}
  FXMenuPane(const FXMenuPane&){}
public:
  FXMenuPane(FXApp* a,FXuint opts=0);
  virtual FXbool contains(FXint parentx,FXint parenty) const;
  };


#endif
