/********************************************************************************
*                                                                               *
*                           R e g i s t r y   C l a s s                         *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXRegistry.h,v 1.15 1999/01/07 07:15:21 jeroen Exp $                     *
********************************************************************************/
#ifndef FXREGISTRY_H
#define FXREGISTRY_H


class FXRegistry {
  friend class FXApp;

protected:
  FXSectionDict  sections;        // Dictionary of sections
  const FXchar  *applicationkey;  // Application key
  const FXchar  *vendorkey;       // Vendor key
  FXbool         modified;        // Settings changed
  FXbool	 ascii;		  // ASCII file-based registry
  
protected:
  FXbool readFromDir(const FXchar* dirname,FXbool mark);
#ifdef WIN32
  FXbool readFromRegistry(HKEY hRootKey,FXbool mark);
  FXbool readRegistrySection(HKEY hAppKey,FXStringDict*& group,const char *groupname,FXbool mark);
  FXbool writeToRegistry(HKEY hRootKey);
#endif
  FXbool parseFile(const FXchar* filename,FXbool mark);
  FXbool parseBuffer(FXStringDict*& group,const FXchar* buffer,const FXchar *filename,FXint lineno,FXbool mark);
  FXbool parseValue(FXchar* value,const FXchar* buffer,const FXchar *filename,FXint lineno);
  FXbool unparseValue(FXchar* buffer,const FXchar* value);
  FXbool unparseFile(const FXchar* filename);
  
protected:
  FXRegistry(){}
  FXRegistry(const FXRegistry&){}
public:
  
  // Construct registry object; akey and vkey must be string constants.
  // Regular applications SHOULD set a vendor key!
  FXRegistry(const FXchar* akey,const FXchar* vkey=NULL);

  // Read registry
  FXbool read();
  
  // Write registry
  FXbool write();
  
  // Read a registry entry
  const FXchar *readEntry(const FXchar *section,const FXchar *key,const FXchar *def=NULL);
  FXuint readEntry(const FXchar *section,const FXchar *key,FXuint def=0);
  FXint readEntry(const FXchar *section,const FXchar *key,FXint def=0);
  FXdouble readEntry(const FXchar *section,const FXchar *key,FXdouble def=0.0);
  
  // Write a registry entry
  FXbool writeEntry(const FXchar *section,const FXchar *key,const FXchar *val);
  FXbool writeEntry(const FXchar *section,const FXchar *key,FXint val);
  FXbool writeEntry(const FXchar *section,const FXchar *key,FXuint val);
  FXbool writeEntry(const FXchar *section,const FXchar *key,FXdouble val);
  
  // Delete a registry entry
  FXbool deleteEntry(const FXchar *section,const FXchar *key);
  
  // Delete section
  FXbool deleteSection(const FXchar *section);
  
  // Mark as changed
  void setModified(FXbool mdfy=TRUE){ modified=mdfy; }

  // Is it modified
  FXbool isModified() const { return modified; }

  // Set ASCII mode
  void setAsciiMode(FXbool asciiMode);

 ~FXRegistry();
  };
  
  
#endif
