/********************************************************************************
*                                                                               *
*                     A p p l i c a t i o n   O b j e c t                       *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* Major Contributions for Windows NT by Lyle Johnson                            *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXApp.cpp,v 1.175 1999/01/15 04:54:12 jeroen Exp $                       *
********************************************************************************/
#include "xincs.h"
#include "fxdefs.h"
#include "fxkeys.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDC.h"
#include "FXWDC.h"
#include "FXCursor.h"
#include "FXFont.h"
#include "FXDrawable.h"
#include "FXBitmap.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXGIFIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXString.h"
#include "FXLabel.h"
#include "FXButton.h"
#include "FXComposite.h"
#include "FXRootWindow.h"
#include "FXShell.h"
#include "FXTopWindow.h"
#include "FXDialogBox.h"
#include "FXMessageBox.h"

/*
  Notes:
  
  - The alloc color stuff should perhaps go elsewhere.
  - Colors should be obtained in the window, as they are colormap/visual
    dependent, and different windows can have different maps.
  - Fix gray bitmap stuff
  - Do we still need both gray pixmap and bitmap?
  - Move all compiled-in resources to file FXResources.cc
  - We need a ``device context'' concept, which embodies colormaps, palettes, GC, etc.
  - Should not be able to DND drop into some windows while modal window is up
  - Perhaps combine delayed paint, scroll, and layout into one mechanism...
  - refresh() now forces total GUI updating; flush() updates display only.
  - Need way to inquire about DND type name
  - Need some sort of journal facility.
  - Should not always combine rectangles; it might get too big...
  - Sender is the FXApp, but in case of DND ops, we might have a special stand-in object
    as the sender (e.g. so we can send it messages).
  - Need two constructors for FXApp: one for deserialize, one normal...
  - Application should have a target to be notified of certain app-wide events,
    such as activation of [a window of] the application etc.
*/

// Just in case
#define bzero(ptr,size) memset(ptr,0,size)

#ifdef _HPUX_SOURCE
#define SELECT(n,r,w,e,t)  select(n,(int*)(r),(int*)(w),(int*)(e),t)
#else
#if defined(_POWER) || defined(_IBMR2)
#define SELECT(n,r,w,e,t)  select(n,(void*)(r),(void*)(w),(void*)(e),t)
extern "C" int gettimeofday(struct timeval *tv, struct timezone *tz);
#else
#define SELECT(n,r,w,e,t)  select(n,r,w,e,t)
#endif
#endif


/*******************************************************************************/

// A Timer object
struct FXTimer {
  FXTimer       *next;              // Next timeout in list
  FXObject      *target;            // Receiver object
  FXSelector     message;           // Message sent to receiver
#ifndef FX_NATIVE_WIN32
  struct timeval due;               // When time is due
#else
  UINT           id;                // Timer id
#endif
  };

  
// An Idle object
struct FXChore {
  FXChore       *next;              // Next chore in list
  FXObject      *target;            // Receiver object
  FXSelector     message;           // Message sent to receiver
  };


// A repaint event record
struct FXRepaint {
  FXRepaint     *next;              // Next repaint in list
  FXID           window;            // Window ID of the dirty window
  FXRectangle    rect;              // Dirty rectangle
  FXbool         synth;             // Synthetic expose event or real one?
  };

  
// IO Record
struct FXCBSpec {
  FXObject      *target;            // Receiver object
  FXSelector     message;           // Message sent to receiver
  };

  
// Input record
struct FXInput {
  FXCBSpec read;                    // Callback spec for read
  FXCBSpec write;                   // Callback spec for write
  FXCBSpec except;                  // Callback spec for except
  };

   
/*******************************************************************************/


// Horizontal splitter cursor
#include "hsplit.xbm"
#include "hsplit_mask.xbm"

// Vertical splitter cursor
#include "vsplit.xbm"
#include "vsplit_mask.xbm"
   
// Corner resize handle cursor
#include "resize.xbm"
#include "resize_mask.xbm"

// Color swatch drag-and-drop cursor
#include "swatch.xbm"
#include "swatch_mask.xbm"

// NO DROP drag-and-drop cursor
#include "dontdrop.xbm"
#include "dontdrop_mask.xbm"

// Upper or lower side MDI resize cursor
#include "resizetop.xbm"
#include "resizetop_mask.xbm"
   
// Right MDI resize cursor
#include "resizetopright.xbm"
#include "resizetopright_mask.xbm"

// Left MDI resize cursor
#include "resizetopleft.xbm"
#include "resizetopleft_mask.xbm"

// Left or right side MDI resize cursor
#include "resizeleft.xbm"
#include "resizeleft_mask.xbm"

// Move cursor
#include "move.xbm"
#include "move_mask.xbm"

// Drag and drop COPY
#include "dndcopy.xbm"
#include "dndcopy_mask.xbm"
   
// Drag and drop LINK
#include "dndlink.xbm"
#include "dndlink_mask.xbm"
   
// Drag and drop MOVE
#include "dndmove.xbm"
#include "dndmove_mask.xbm"

// Gray bitmap 32x2
#include "gray.xbm"

  
/*******************************************************************************/

   
// Default palette
static const FXuchar default_palette[][3]={
  {0x00,0x00,0x00},
  {0xff,0xff,0xff},
  {0xdd,0xdd,0xdd},
  {0xbb,0xbb,0xbb},
  {0x99,0x99,0x99},
  {0x77,0x77,0x77},
  {0x55,0x55,0x55},
  {0x33,0x33,0x33},
  {0x88,0x00,0x00},
  {0xcc,0x00,0x00},
  {0xff,0x00,0x00},
  {0xff,0x44,0x00},
  {0xff,0x88,0x00},
  {0xff,0xcc,0x00},
  {0xff,0xff,0x00},
  {0xcc,0xcc,0x00},
  {0x88,0x88,0x00},
  {0x44,0x44,0x00},
  {0x00,0x44,0x00},
  {0x00,0x88,0x00},
  {0x00,0xcc,0x00},
  {0x00,0xff,0x00},
  {0x00,0x44,0x44},
  {0x00,0x88,0x88},
  {0x00,0xff,0xff},
  {0x00,0x00,0x44},
  {0x00,0x00,0x88},
  {0x00,0x00,0xff},
  {0x88,0x00,0x88},
  {0xff,0xcc,0x99},
  {0xcc,0xaa,0x77},
  {0xaa,0x88,0x55},
  {0x88,0x66,0x33},
  {0x66,0x44,0x11},
  {0x44,0x22,0x00},
  {0x22,0x00,0x00},
  {0x00,0x44,0x88},
  {0x44,0x88,0xcc},
  {0x88,0xcc,0xff},
  {0x00,0xcc,0x44},
  {0x44,0x88,0x44},
  {0x88,0xcc,0x00},
  {0x22,0x44,0x11},
  {0x33,0x66,0x22},
  {0x44,0x55,0x33},
  {0x66,0x88,0x44},
  {0x33,0x66,0x22},
  {0x22,0x44,0x11},
  };


// Version number
const FXuchar FXApp::version[3]={FOX_MAJOR,FOX_MINOR,FOX_LEVEL};


/*******************************************************************************/

// Map
FXDEFMAP(FXApp) FXAppMap[]={
  FXMAPFUNC(SEL_COMMAND,FXApp::ID_QUIT,FXApp::onCmdQuit),
  };


// Implementation
FXIMPLEMENT(FXApp,FXObject,FXAppMap,ARRAYNUMBER(FXAppMap))


// Application object
FXApp* FXApp::app=NULL;


// Windows application instance handle
#ifdef FX_NATIVE_WIN32
HINSTANCE FXApp::hInstance=0;
#endif



// Initialize application object
FXApp::FXApp(const FXchar *name,const FXchar *vendor) : appname(name),registry(name,vendor){
  
  // Test if application object already exists
  if(app){ fxwarning("Warning: Trying to construct multiple application objects.\n"); }
  
  // Initialize private platform independent data
  windowdc=new FXWDC(this);           // Window device context
  bitmapdc=new FXWDC(this);           // Null device context
  nulldc=new FXDC(this);              // Null device context
  mainWindow=NULL;                    // Main window, or window group leader
  focusWindow=NULL;                   // Window which has focus
  cursorWindow=NULL;                  // Window under the cursor
  grabWindow=NULL;                    // Window that is grabbed
  refresher=NULL;                     // Window being refreshed
  root=new FXRootWindow(this);
  timers=NULL;
  chores=NULL;
  repaints=NULL;
  timerrecs=NULL;
  chorerecs=NULL;
  repaintrecs=NULL;
  invocation=NULL;                    // Modal loop invocation
  FXCALLOC(&inputs,FXInput,8);        // Input file descriptors
  ninputs=8;                          // Number of these
  maxinput=0;                         // Maximum input number
  FXCALLOC(&r_fds,fd_set,1);          // Read File Descriptor set
  FXCALLOC(&w_fds,fd_set,1);          // Write File Descriptor set
  FXCALLOC(&e_fds,fd_set,1);          // Except File Descriptor set
  done=0;                             // True if application is done
  exitcode=0;
  
  // Clear event structure
  event.type=0;
  event.window=0;
  event.time=0;
  event.win_x=0;
  event.win_y=0;
  event.root_x=0;
  event.root_y=0;
  event.state=0;
  event.code=0;
  event.last_x=0;
  event.last_y=0;
  event.click_x=0;
  event.click_y=0;
  event.click_time=0;
  event.click_window=0;
  event.click_button=0;
  event.click_count=0;
  event.moved=0;
  event.rect.x=0;
  event.rect.y=0;
  event.rect.w=0;
  event.rect.h=0;
  event.synthetic=0;
  event.requestor=0;
  event.property=0;
  event.target=0;
  event.origin=FROM_UNKNOWN;
  
  
  // X Window specific inits
#ifndef FX_NATIVE_WIN32
  display=NULL;
  connection=-1;
  wcontext=0;
  visual=NULL;
  depth=0;
  colormap=0;
  wmDeleteWindow=0;
  wmQuitApp=0;
  wmProtocols=0;
  wmMotifHints=0;
  wmTakeFocus=0;
  
  // CLIPBOARD
  cbSelection=0;
  
  // XDND
  xdndAware=0;                        // XDND awareness atom    
  xdndEnter=0;                        // XDND Message types
  xdndLeave=0;       
  xdndPosition=0;    
  xdndStatus=0;      
  xdndDrop=0;   
  xdndFinished=0;     
  xdndSelection=0;
  xdndActionMove=0;                   // XDND Copy action
  xdndActionCopy=0;                   // XDND Move action
  xdndActionLink=0;                   // XDND Link action
  xdndTypes=0;
  xdndSource=0;                       // XDND drag source window
  xdndTarget=0;                       // XDND drop target window
  xdndAction=0;                       // XDND action being performed
  xdndAccepts=FALSE;                  // XDND true if target accepts
  xdndSendPosition=FALSE;             // XDND send position update when status comes in
  xdndStatusPending=FALSE;            // XDND waiting for status feedback
  xdndFinishPending=FALSE;            // XDND waiting for drop-confirmation
  xdndStatusReceived=FALSE;           // XDND received at least one status
  xdndWantUpdates=TRUE;               // XDND target always wants new positions
  xdndVersion=XDND_PROTOCOL_VERSION;  // Version for this transaction
  xdndTypeList=NULL;                  // XDND types list
  xdndNumTypes=0;
  xdndRect.x=0;                       // XDND motion rectangle
  xdndRect.y=0;
  xdndRect.w=0;
  xdndRect.h=0;

  // Miscellaneous stuff
  shmi=TRUE;
  shmp=TRUE;
  synchronize=FALSE;
  palette=NULL;
  ncolors=0;
#endif

  // DDE
  ddeAtom=0;                          // Data exchange atom
  ddeDelete=0;                        // Data exchange delete request
  ddeData=NULL;                       // Data exchange array
  ddeSize=0;                          // Data exchange array size
    
  // Initialize public data
  
  // Make font
  normalFont=new FXFont(this,"-adobe-helvetica-bold-r-*-*-12-*-*-*-*-*-iso8859-1");
  
  // Make some cursors
  arrowCursor=new FXCursor(this,CURSOR_ARROW);
  rarrowCursor=new FXCursor(this,CURSOR_RARROW);
  textCursor=new FXCursor(this,CURSOR_IBEAM);

  // Cursors for splitter
  hsplitCursor=new FXCursor(this,hsplit_bits,hsplit_mask_bits,hsplit_width,hsplit_height,hsplit_x_hot,hsplit_y_hot);
  vsplitCursor=new FXCursor(this,vsplit_bits,vsplit_mask_bits,vsplit_width,vsplit_height,vsplit_x_hot,vsplit_y_hot);
  
  // Resize corner
  resizeCursor=new FXCursor(this,resize_bits,resize_mask_bits,resize_width,resize_height,resize_x_hot,resize_y_hot);
  
  // Color swatch
  swatchCursor=new FXCursor(this,swatch_bits,swatch_mask_bits,swatch_width,swatch_height,swatch_x_hot,swatch_y_hot);
  
  // DND no drop
  dontdropCursor=new FXCursor(this,dontdrop_bits,dontdrop_mask_bits,dontdrop_width,dontdrop_height,dontdrop_x_hot,dontdrop_y_hot);

  // Move  
  moveCursor=new FXCursor(this,move_bits,move_mask_bits,move_width,move_height,move_x_hot,move_y_hot);

  // Dragging edges/corners  
  dragHCursor=new FXCursor(this,resizetop_bits,resizetop_mask_bits,resizetop_width,resizetop_height,resizetop_x_hot,resizetop_y_hot);
  dragVCursor=new FXCursor(this,resizeleft_bits,resizeleft_mask_bits,resizeleft_width,resizeleft_height,resizeleft_x_hot,resizeleft_y_hot);
  dragTRCursor=new FXCursor(this,resizetopright_bits,resizetopright_mask_bits,resizetopright_width,resizetopright_height,resizetopright_x_hot,resizetopright_y_hot);
  dragTLCursor=new FXCursor(this,resizetopleft_bits,resizetopleft_mask_bits,resizetopleft_width,resizetopleft_height,resizetopleft_x_hot,resizetopleft_y_hot);

  // DND actions  
  dndCopyCursor=new FXCursor(this,dndcopy_bits,dndcopy_mask_bits,dndcopy_width,dndcopy_height,dndcopy_x_hot,dndcopy_y_hot);
  dndMoveCursor=new FXCursor(this,dndmove_bits,dndmove_mask_bits,dndmove_width,dndmove_height,dndmove_x_hot,dndmove_y_hot);
  dndLinkCursor=new FXCursor(this,dndlink_bits,dndlink_mask_bits,dndlink_width,dndlink_height,dndlink_x_hot,dndlink_y_hot);
  
  // Bitmaps
  grayBitmap=new FXBitmap(this,gray_bits,0,gray_width,gray_height);
  
  // Other settings
  clickSpeed=400;
  scrollSpeed=100;
  blinkSpeed=500;
  menuPause=400;
  tooltipPause=800;
  tooltipTime=3000;
  scrollbarWidth=15;
  dragDelta=6;
  
  // Init colors
  backColor=FXRGB(192,192,192);   
  foreColor=FXRGB(0,0,0);   
  hiliteColor=makeHiliteColor(backColor); 
  shadowColor=makeShadowColor(backColor); 
  borderColor=FXRGB(0,0,0);
  selforeColor=FXRGB(255,255,255);
  selbackColor=FXRGB(0,0,128);
  
  // Pointer to FXApp
  app=this;
  }


/*******************************************************************************/

#ifndef FX_NATIVE_WIN32

// Create palette
void FXApp::makePalette(){
  XColor table[256],color;
  FXuint i,gottable,alloced;

  // Test just to make sure
  if(visual->map_entries>256){ fxerror("%s::makeColor: bigger colormap than expected.\n",getClassName()); }

  FXTRACE((100,"visual->map_entries: %d\n",visual->map_entries));

  // Make palette
  ncolors=ARRAYNUMBER(default_palette);
  palette=new FXPalEntry[ncolors];
  
  gottable=0;
  
  // Allocate colors
  for(i=0; i<ncolors; i++){
    
    color.red=default_palette[i][0]*257;
    color.green=default_palette[i][1]*257;
    color.blue=default_palette[i][2]*257;
    color.flags=DoRed|DoGreen|DoBlue;

    // First try just using XAllocColor
    alloced=XAllocColor(display,colormap,&color);
    if(alloced==0){
      FXuint dist,mindist,bestmatch;
      FXint j,dr,dg,db;


      // Get colors in the map
      if(!gottable){
        for(j=0; j<visual->map_entries; j++) table[j].pixel=j;
        XQueryColors(display,colormap,table,visual->map_entries);
        gottable=1;
        }

      // Find best (weighted) match; also gives good colors for grey-ramps
      for(mindist=4294967295U,bestmatch=0,j=0; j<visual->map_entries; j++){
        dr=19660*(color.red-table[j].red);
        dg=38666*(color.green-table[j].green);
        db=7209*(color.blue-table[j].blue);
        dist=(FXuint)FXABS(dr) + (FXuint)FXABS(dg) + (FXuint)FXABS(db);
        if(dist<mindist){
          bestmatch=j;
          mindist=dist;
          }
        }

      // Return result
      color.red=table[bestmatch].red;
      color.green=table[bestmatch].green;
      color.blue=table[bestmatch].blue;

      // Try to allocate the closest match color.  This should only
      // fail if the cell is read/write.  Otherwise, we're incrementing
      // the cell's reference count.
      alloced=XAllocColor(display,colormap,&color);
      if(!alloced){
        color.pixel=bestmatch;
        color.red=table[bestmatch].red;
        color.green=table[bestmatch].green;
        color.blue=table[bestmatch].blue;
        color.flags=DoRed|DoGreen|DoBlue;
        }
      }
    
    // Fill in our own data structure
    palette[i].pixel=color.pixel;
    palette[i].red=color.red/257;
    palette[i].green=color.green/257;
    palette[i].blue=color.blue/257;
    palette[i].alloced=alloced;
    }
  }


// Get palette
void FXApp::getPalette(){
  XColor table[256];
  FXuint i;

  // Test just to make sure
  if(visual->map_entries>256){ fxerror("%s::getPalette: bigger colormap than expected.\n",getClassName()); }

  FXTRACE((100,"visual->map_entries: %d\n",visual->map_entries));

  // Make palette
  ncolors=visual->map_entries;
  palette=new FXPalEntry[ncolors];
  
  for(i=0; i<ncolors; i++) table[i].pixel=i;
  XQueryColors(display,colormap,table,visual->map_entries);
  
  // Fill in our own data structure
  for(i=0; i<ncolors; i++){
    palette[i].pixel=table[i].pixel;
    palette[i].red=table[i].red/257;
    palette[i].green=table[i].green/257;
    palette[i].blue=table[i].blue/257;
    palette[i].alloced=0;
    }
  }


// Free palette
void FXApp::freePalette(){
  FXPixel pixels[256];
  FXuint i,n;
  for(i=n=0; i<ncolors; i++){if(palette[i].alloced) pixels[n++]=palette[i].pixel;}
  if(n!=0){XFreeColors(display,colormap,pixels,n,0);}
  delete palette;
  ncolors=0;
  }

#endif


// Get a color
FXPixel FXApp::matchColor(FXuint r,FXuint g,FXuint b){
#ifndef FX_NATIVE_WIN32
  register FXuint bestmatch,j,dist,mindist;
  register FXint dr,dg,db;
  for(mindist=4294967295U,bestmatch=0,j=0; j<ncolors; j++){
    dr=(r-palette[j].red);
    dg=(g-palette[j].green);
    db=(b-palette[j].blue);
    dist=(FXuint)FXABS(dr) + (FXuint)FXABS(dg) + (FXuint)FXABS(db);
    if(dist<mindist){
      bestmatch=j;
      mindist=dist;
      }
    }
  FXASSERT(bestmatch<ncolors);
  return palette[bestmatch].pixel;
#else
  return PALETTERGB(r,g,b);
#endif
  }


/*******************************************************************************/


// Perhaps should do something else...
#ifndef FX_NATIVE_WIN32
static int xerrorhandler(Display* dpy,XErrorEvent* eev){
  char buf[1000];
  XGetErrorText(dpy,eev->error_code,buf,sizeof(buf));
  fxwarning("X Error: code %d major %d minor %d: %s.\n",eev->error_code,eev->request_code,eev->minor_code,buf);
  return 1;
  }
#endif


// Fatal error (e.g. lost connection)
#ifndef FX_NATIVE_WIN32
static int xfatalerrorhandler(Display*){
  fxerror("X Fatal error.\n");
  return 1;
  }
#endif

/*******************************************************************************/


// Open the display
#ifndef FX_NATIVE_WIN32
int FXApp::openDisplay(const char* dpyname){
  const unsigned char black_bits[]={0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff};
  unsigned long gmask;
  const char* dpy;
  XGCValues gval;
 
  
  // Set error handler
  XSetErrorHandler(xerrorhandler);
  
  // Set fatal handler; it should save the app data & exit the program
  XSetIOErrorHandler(xfatalerrorhandler);
  
  // Try locate display
  dpy=dpyname;
  if(!dpy){
    dpy=getenv("DISPLAY");
    if(!dpy) dpy=":0.0";
    }

  // What's going on
  FXTRACE((100,"%s::openDisplay: opening display: %s.\n",getClassName(),dpy));
  
  // Open display
  display=XOpenDisplay(dpy);
  if(!display){ fxerror("%s::openDisplay: unable to open display %s\n",getClassName(),dpy); }

  // For debugging
  if(synchronize) XSynchronize(display,TRUE);

  // Get X connection number
  connection=ConnectionNumber(display);

  // Make hash context for window mapping
  wcontext=XUniqueContext();
  
  // Get default colormap
  colormap=XDefaultColormap(display,XDefaultScreen(display));

  // Get default visual
  visual=XDefaultVisual(display,XDefaultScreen(display));
  
  // If PseudoColor, claim colormap
  if(visual->c_class==PseudoColor || visual->c_class==GrayScale){
    FXTRACE((100,"Pseudo color display.\n"));
    makePalette();
    }
  
  // If StaticColor, just grab colormap
  else if(visual->c_class==StaticColor || visual->c_class==StaticGray){
    FXTRACE((100,"Static color display.\n"));
    getPalette();
    }
  
  // True color
  else{
    FXTRACE((100,"True color display.\n"));
    }
  
  // See if we wanted/have XSHM
#ifdef HAVE_XSHM
  FXTRACE((100,"Checking for shared memory.\n"));
  if(shmi || shmp){
    int maj,min,dum; Bool pm;
    shmi=FALSE;
    shmp=FALSE;
    if(XQueryExtension(display,"MIT-SHM",&dum,&dum,&dum)){
      if(XShmQueryVersion(display,&maj,&min,&pm)){
        shmp=pm && (XShmPixmapFormat(display)==ZPixmap);
        shmi=TRUE;
        }
      }
    }
  FXTRACE((100,"Shared Images =%d\n",shmi));
  FXTRACE((100,"Shared Pixmaps=%d\n",shmp));
   
  // Don't have it!
#else
  FXTRACE((100,"Shared memory not available.\n"));
  shmi=FALSE;
  shmp=FALSE;
#endif

  // Window manager communication
  wmDeleteWindow=XInternAtom(display,"WM_DELETE_WINDOW",0);
  wmQuitApp=XInternAtom(display,"_WM_QUIT_APP",0);
  wmProtocols=XInternAtom(display,"WM_PROTOCOLS",0);
  wmMotifHints=XInternAtom(display,"_MOTIF_WM_HINTS",0);
  wmTakeFocus=XInternAtom(display,"WM_TAKE_FOCUS",0);
  
  // Clipboard
  cbSelection=XInternAtom(display,"CLIPBOARD",0);

  // XDND protocol awareness
  xdndAware=XInternAtom(display,"XdndAware",0);
  
  // XDND Messages
  xdndEnter=XInternAtom(display,"XdndEnter",0);
  xdndLeave=XInternAtom(display,"XdndLeave",0);
  xdndPosition=XInternAtom(display,"XdndPosition",0);
  xdndStatus=XInternAtom(display,"XdndStatus",0);
  xdndDrop=XInternAtom(display,"XdndDrop",0);
  xdndFinished=XInternAtom(display,"XdndFinished",0);
  
  // XDND Selection atom
  xdndSelection=XInternAtom(display,"XdndSelection",0);

  // XDND Actions
  xdndActionCopy=XInternAtom(display,"XdndActionCopy",0);
  xdndActionMove=XInternAtom(display,"XdndActionMove",0);
  xdndActionLink=XInternAtom(display,"XdndActionLink",0);
  
  // XDND Types list
  xdndTypes=XInternAtom(display,"XdndTypeList",0);
  
  // DDE property
  ddeAtom=XInternAtom(display,"_FOX_DDE",0);
  ddeDelete=XInternAtom(display,"DELETE",0);
  
  // Init DDE stuff
  ddeData=NULL;
  ddeSize=0;
  
  // Black stipple
  blackstipple=XCreateBitmapFromData(display,XDefaultRootWindow(display),(char*)black_bits,8,8);

  // Create GC for windows
  gmask=GCFillStyle|GCGraphicsExposures;
  gval.fill_style=FillSolid;
  gval.graphics_exposures=False;
  ((FXWDC*)windowdc)->dc=XCreateGC(display,XDefaultRootWindow(display),gmask,&gval);

  // Create GC for bitmaps
  gmask=GCFillStyle|GCGraphicsExposures;
  gval.fill_style=FillSolid;
  gval.graphics_exposures=False;
  ((FXWDC*)bitmapdc)->dc=XCreateGC(display,blackstipple,gmask,&gval);
  
  return 1;
  }
#endif
  
  
// Close display
#ifndef FX_NATIVE_WIN32
void FXApp::closeDisplay(){
  if(!display) return;
  XFreeGC(display,((FXWDC*)windowdc)->dc);
  XFreeGC(display,((FXWDC*)bitmapdc)->dc);
  XCloseDisplay(display);
  }
#endif



// Register the window classes
// We may need several classes: for popups, dialogs, controls, and toplevel windows
#ifdef FX_NATIVE_WIN32

// DDE callback function
HDDEDATA CALLBACK FXApp::DdeCallback(UINT iType,UINT iFmt,HCONV hConv,HSZ hsz1,HSZ hsz2,HDDEDATA hData,DWORD dwData1,DWORD dwData2){
  return app->ddeCallback(iType,iFmt,hConv,hsz1,hsz2,hData,dwData1,dwData2);
  }


void FXApp::registerWindowClasses(){
  WNDCLASSEX wndclass;
  
  // Register our sole window class
  wndclass.cbSize=sizeof(WNDCLASSEX);
  wndclass.style=CS_HREDRAW|CS_VREDRAW;
  wndclass.lpfnWndProc=FXApp::wndproc;
  wndclass.cbClsExtra=0;
  wndclass.cbWndExtra=sizeof(FXWindow*);
  wndclass.hInstance=FXApp::hInstance;
  wndclass.hIcon=LoadIcon(NULL,IDI_APPLICATION);
  wndclass.hCursor=LoadCursor(NULL,IDC_ARROW);
  wndclass.hbrBackground=(HBRUSH) GetStockObject(NULL_BRUSH);
  wndclass.lpszMenuName=NULL;
  wndclass.lpszClassName="FXWindow";
  wndclass.hIconSm=LoadIcon(NULL,IDI_APPLICATION);
  RegisterClassEx(&wndclass);

  // Initialize DDE library
  idInst=0;
  if(DdeInitialize(&idInst,DdeCallback,CBF_FAIL_EXECUTES|CBF_FAIL_POKES|CBF_SKIP_REGISTRATIONS|CBF_SKIP_UNREGISTRATIONS,0)==DMLERR_NO_ERROR){
    HSZ hszService=DdeCreateStringHandle(idInst,appname,0);
    DdeNameService(idInst,hszService,NULL,DNS_REGISTER);
    }

  ddeDelete=RegisterClipboardFormat("DELETE");

  // XDND messages
  xdndEnter=RegisterWindowMessage("XdndEnter");
  xdndLeave=RegisterWindowMessage("XdndLeave");
  xdndPosition=RegisterWindowMessage("XdndPosition");
  xdndStatus=RegisterWindowMessage("XdndStatus");
  xdndDrop=RegisterWindowMessage("XdndDrop");
  xdndFinished=RegisterWindowMessage("XdndFinished");

  // XDND Actions
  xdndActionCopy=GlobalAddAtom("XdndActionCopy");
  xdndActionMove=GlobalAddAtom("XdndActionMove");
  xdndActionLink=GlobalAddAtom("XdndActionLink");
  }

#endif


/*******************************************************************************/

// Compare times
static inline int operator<(const struct timeval& a,const struct timeval& b){
  return (a.tv_sec<b.tv_sec) || (a.tv_sec==b.tv_sec && a.tv_usec<b.tv_usec);
  }

#ifdef FX_NATIVE_WIN32

VOID CALLBACK FXApp::TimerProc(HWND hwnd,UINT uMsg,UINT idEvent,DWORD dwTime){
  app->timerProc(idEvent,dwTime);
  }

// Fired timer, unlink from list
// Note the callback may do anything, including setting another timer;
// So best is to unlink before calling it, and add it to the free list
// after calling it.
void FXApp::timerProc(UINT id,DWORD dwTime){
  FXTimer *h,**hh;
  for(h=timers,hh=&timers; h && h->id!=id; hh=&h->next,h=h->next);
  if(h==NULL) return;
  FXASSERT(h->id==id);
  *hh=h->next;
  if(h->target && h->target->handle(this,MKUINT(h->message,SEL_TIMEOUT),&event)) refresher=root;
  h->next=timerrecs;
  timerrecs=h;
  }

#endif


// Add timeout, sorted by time
FXTimer* FXApp::addTimeout(FXint ms,FXObject* tgt,FXSelector sel){
  if(ms<1){ fxerror("%s::addTimeout: wait time should be greater than 0\n",getClassName()); }
  FXTimer *t;
  if(timerrecs){
    t=timerrecs;
    timerrecs=t->next;
    }
  else{
    t=new FXTimer;
    }
  t->target=tgt;
  t->message=sel;
#ifndef FX_NATIVE_WIN32
  gettimeofday(&t->due,NULL);
  t->due.tv_sec+=ms/1000;
  t->due.tv_usec+=(ms%1000)*1000;
  if(t->due.tv_usec>=1000000){
    t->due.tv_usec-=1000000;
    t->due.tv_sec+=1;
    }
  FXTimer **hh=&timers;
  FXTimer *h=*hh;
  while(h && (h->due < t->due)){
    hh=&h->next;
    h=*hh;
    }
  t->next=h;
  *hh=t;
#else
  t->id=SetTimer(NULL,NULL,ms,(TIMERPROC)TimerProc);    ///Use MsgWaitMultipleObjects() ???? very similar to select()
  t->next=timers;
  timers=t;
#endif
  return t;
  }


// Remove timeout from the list
FXTimer* FXApp::removeTimeout(FXTimer *t){
  if(t){
    FXTimer *h,**hh;
    for(h=timers,hh=&timers; h!=t; hh=&h->next,h=h->next){
      if(h==NULL) return NULL;
      }
    FXASSERT(h==t);
    *hh=t->next;
    t->next=timerrecs;
    timerrecs=t;
#ifdef FX_NATIVE_WIN32
    KillTimer(NULL,t->id);
#endif
    }
  return NULL;
  }


/*******************************************************************************/


// Add chore to list
FXChore* FXApp::addChore(FXObject* tgt,FXSelector sel){
  FXChore *c;
  if(chorerecs){
    c=chorerecs;
    chorerecs=c->next;
    }
  else{
    c=new FXChore;
    }
  c->target=tgt;
  c->message=sel;
  c->next=chores;
  chores=c;
  return c;
  }


// Remove chore from the list
FXChore* FXApp::removeChore(FXChore *c){
  if(c){
    FXChore *h,**hh;
    for(h=chores,hh=&chores; h!=c; hh=&h->next,h=h->next){
      if(h==NULL) return NULL;
      }
    FXASSERT(h==c);
    *hh=c->next;
    c->next=chorerecs;
    chorerecs=c;
    }
  return NULL;
  }


/*******************************************************************************/


// Add input
void FXApp::addInput(FXint fd,FXuint mode,FXObject *tgt,FXSelector sel){
  if(fd<0 || mode==INPUT_NONE) return;
  if(fd>=FD_SETSIZE){ fxerror("%s::addInput: illegal file descriptor.\n",getClassName()); }
  if(fd>=ninputs){
    FXRESIZE(&inputs,FXInput,fd+1);
    memset(&inputs[ninputs],0,sizeof(FXInput)*(fd+1-ninputs));
    ninputs=fd+1;
    }
  FXASSERT(inputs);
  FXASSERT(fd<ninputs);
  if(mode&INPUT_READ){
    inputs[fd].read.target=tgt;
    inputs[fd].read.message=sel;
    FD_SET(fd,(fd_set*)r_fds);
    }
  if(mode&INPUT_WRITE){
    inputs[fd].write.target=tgt;
    inputs[fd].write.message=sel;
    FD_SET(fd,(fd_set*)w_fds);
    }
  if(mode&INPUT_EXCEPT){
    inputs[fd].except.target=tgt;
    inputs[fd].except.message=sel;
    FD_SET(fd,(fd_set*)e_fds);
    }
  if(fd>maxinput) maxinput=fd;
  }


// Remove input
void FXApp::removeInput(FXint fd,FXuint mode){
  if(fd<0 || mode==INPUT_NONE) return;
  if(fd>maxinput){ fxerror("%s::removeInput: illegal file descriptor.\n",getClassName()); }
  if(mode&INPUT_READ){
    inputs[fd].read.target=NULL;
    inputs[fd].read.message=0;
    FD_CLR(fd,(fd_set*)r_fds);
    }
  if(mode&INPUT_WRITE){
    inputs[fd].write.target=NULL;
    inputs[fd].write.message=0;
    FD_CLR(fd,(fd_set*)w_fds);
    }
  if(mode&INPUT_EXCEPT){
    inputs[fd].except.target=NULL;
    inputs[fd].except.message=0;
    FD_CLR(fd,(fd_set*)e_fds);
    }
  while(0<=maxinput){
    if(inputs[maxinput].read.target || inputs[maxinput].write.target || inputs[maxinput].except.target) break;
    maxinput--;
    }
  }


/*******************************************************************************/


// Add a repaint or scroll rectangle
#ifndef FX_NATIVE_WIN32
void FXApp::addRepaint(FXID win,FXint x,FXint y,FXint w,FXint h,FXbool synth){
  register FXRepaint *r,**pr;
  w+=x;
  h+=y;
  for(r=repaints,pr=&repaints; r; pr=&r->next,r=r->next){
    if(r->window==win){
      
      // Composite rectangles
      r->rect.x=FXMIN(x,r->rect.x);
      r->rect.y=FXMIN(y,r->rect.y);
      r->rect.w=FXMAX(w,r->rect.w);
      r->rect.h=FXMAX(h,r->rect.h);
      
      // Will become a synthetic expose event!
      r->synth|=synth;

      return;
      }
    }
    
  // Nope, get new rectangle
  if(repaintrecs){
    r=repaintrecs;
    repaintrecs=r->next;
    }
  else{
    r=new FXRepaint;
    }
  
  // And add at the end
  r->window=win;
  r->rect.x=x;
  r->rect.y=y;
  r->rect.w=w;
  r->rect.h=h;
  r->synth=synth;
  r->next=NULL;
  *pr=r;
  }
#endif


/*******************************************************************************/


#ifndef FX_NATIVE_WIN32

// Get an event
void FXApp::getNextEvent(XEvent& ev){
  struct timeval now,delta;
  fd_set readfds;
  fd_set writefds;
  fd_set exceptfds;
  int maxfds;
  int nfds;
  int fff;
  
  while(1){
    while(1){

      // Handle all currently outstanding timers.
      // This can NOT loop forever, as newly scheduled timers
      // will be scheduled LATER than the current time!
      gettimeofday(&now,NULL);
      while(timers){
        FXTimer* t=timers;
        if(now < t->due) break;
        timers=t->next;
        
        // Call the timeout handler
        if(t->target && t->target->handle(this,MKUINT(t->message,SEL_TIMEOUT),&event)) refresh();
        
        // Add record to list of recycled timer records
        t->next=timerrecs;
        timerrecs=t;
        }

      // Any events already received?
      if(XEventsQueued(display,QueuedAfterFlush)) break;
#ifdef HUMMINGBIRD
      if(XEventsQueued(display,QueuedAfterReading)) break;
#endif

      // Prepare fd's to watch
#ifndef HUMMINGBIRD
      FD_ZERO(&readfds);
      FD_ZERO(&writefds);
      FD_ZERO(&exceptfds);
      FD_SET(connection,&readfds);

      // Add other inputs...
      maxfds=connection;

      delta.tv_usec=0;
      delta.tv_sec=0;

      // Do a quick poll for any ready events (and ONLY event!)
      nfds=SELECT(maxfds+1,&readfds,&writefds,&exceptfds,&delta);
#else
      nfds=0;
#endif

      // No events, do idle processing
      if(nfds==0){

        // Release the expose events...
        if(repaints){
          FXRepaint *r=repaints;
          ev.xany.type=Expose;
          ev.xexpose.window=r->window;
          ev.xexpose.send_event=r->synth;
          ev.xexpose.x=r->rect.x;
          ev.xexpose.y=r->rect.y;
          ev.xexpose.width=r->rect.w-r->rect.x;
          ev.xexpose.height=r->rect.h-r->rect.y;
          repaints=r->next;
          r->next=repaintrecs;
          repaintrecs=r;
          return;
          }
        
        // GUI updating:- walk the whole widget tree.
        if(refresher){
          
          // Call the refresher handler
          refresher->handle(this,MKUINT(0,SEL_UPDATE),&event);
          
          // Next widget to refresh
          if(refresher->getFirst()){
            refresher=refresher->getFirst();
            continue;
            }
          while(!refresher->getNext() && refresher->getParent()!=root){
            refresher=refresher->getParent();
            }
          refresher=refresher->getNext();
          continue;
          }
        
        // Do our chores :-)
        if(chores){
          FXChore* c=chores;
          chores=c->next;
          
          // Call the chore handler
          if(c->target && c->target->handle(this,MKUINT(c->message,SEL_CHORE),&event)) refresh();
          
          // Add record to list of recycled chore records
          c->next=chorerecs;
          chorerecs=c;
          continue;
          }

        // Now, block till timeout, i/o, or event
#ifndef HUMMINGBIRD
        readfds=*((fd_set*)r_fds);
        writefds=*((fd_set*)w_fds);
        exceptfds=*((fd_set*)e_fds);
        FD_SET(connection,&readfds);

        // Add other inputs...
        maxfds=FXMAX(maxinput,connection);

        // We're about to block now; if there are
        // timers, we block only for a little while.
        if(timers){

          // All that testing above may have taken some time...
          gettimeofday(&now,NULL);

          // Compute how long to wait
          delta.tv_usec=timers->due.tv_usec-now.tv_usec;
          delta.tv_sec=timers->due.tv_sec-now.tv_sec;
          while(delta.tv_usec<0){
            delta.tv_usec+=1000000;
            delta.tv_sec-=1;
            }
          FXASSERT(0<=delta.tv_usec && delta.tv_usec<1000000);

          // Some timers are already due; do them right away!
          if(delta.tv_sec<0 || (delta.tv_sec==0 && delta.tv_usec==0)) continue;

          // Block till timer or event or interrupt
          nfds=SELECT(maxfds+1,&readfds,&writefds,&exceptfds,&delta);

          // Timed out, so handle those timeouts
          if(nfds==0) continue;
          }

        // If no timers, we block till event or interrupt
        else{
          nfds=SELECT(maxfds+1,&readfds,&writefds,&exceptfds,NULL);
          }
#endif
        }

      // Interrupt
#ifndef HUMMINGBIRD
      if(nfds<0){
        if(errno==EAGAIN || errno==EINTR){ errno=0; continue; }
        fxerror("Application terminated: interrupt or lost connection errno=%d\n",errno);
        }

      // Try I/O channels first, if any
      if(nfds>1 || !FD_ISSET(connection,&readfds)){
        FXTRACE((100,"Other active FDs: "));
        for(fff=0; fff<=maxinput; fff++){
          if(fff==connection) continue;
          if(FD_ISSET(fff,&readfds)){
            FXTRACE((100,"R%d ",fff));
            if(inputs[fff].read.target && inputs[fff].read.target->handle(this,MKUINT(inputs[fff].read.message,SEL_IO_READ),(void*)fff)){
              refresh();
              }
            }
          if(FD_ISSET(fff,&writefds)){
            FXTRACE((100,"W%d ",fff));
            if(inputs[fff].write.target && inputs[fff].write.target->handle(this,MKUINT(inputs[fff].write.message,SEL_IO_WRITE),(void*)fff)){
              refresh();
              }
            }
          if(FD_ISSET(fff,&exceptfds)){
            FXTRACE((100,"E%d ",fff));
            if(inputs[fff].except.target && inputs[fff].except.target->handle(this,MKUINT(inputs[fff].read.message,SEL_IO_EXCEPT),(void*)fff)){
              refresh();
              }
            }
          }
        FXTRACE((100,"\n"));
        }

      // Got event
      if(FD_ISSET(connection,&readfds)){
        if(XEventsQueued(display,QueuedAfterReading)) break;
        }
#else
      break;
#endif
      }
    
    // Get an event
    XNextEvent(display,&ev);

    // Compress motion events
    if(ev.xany.type==MotionNotify){
      while(XPending(display)){
        XEvent e;
        XPeekEvent(display,&e);
        if((e.xany.type!=MotionNotify) || (ev.xmotion.window!=e.xmotion.window) || (ev.xmotion.subwindow!=e.xmotion.subwindow)) break;
        XNextEvent(display,&ev);
        }
      }

    // Save expose events for later...
#ifndef HUMMINGBIRD
    if(ev.xany.type==Expose || ev.xany.type==GraphicsExpose){
      addRepaint(ev.xexpose.window,ev.xexpose.x,ev.xexpose.y,ev.xexpose.width,ev.xexpose.height,0);
      continue;
      }
#endif
    
    // Regular event
    return;
    }
  }


// Peek for event
FXbool FXApp::peekEvent(){
  struct timeval delta;
  fd_set readfds;
  fd_set writefds;
  fd_set exceptfds;
  int maxfds;
  int nfds;
  
  // Events queued up in client already
  if(XEventsQueued(display,QueuedAfterFlush)) return TRUE;

  // Prepare fd's to watch
  FD_ZERO(&readfds);
  FD_ZERO(&writefds);
  FD_ZERO(&exceptfds);
  FD_SET(connection,&readfds);

  // Add other inputs...
  maxfds=connection;

  delta.tv_usec=0;
  delta.tv_sec=0;

  // Do a quick poll for any ready events
  nfds=SELECT(maxfds+1,&readfds,&writefds,&exceptfds,&delta);
  
  // Interrupt
#ifndef HUMMINGBIRD
  if(nfds<0){
    if(errno==EAGAIN || errno==EINTR){ errno=0; return FALSE; }
    fxerror("Application terminated: interrupt or lost connection errno=%d\n",errno);
    }
  
  // I/O activity
  if(nfds>0){
    if(FD_ISSET(connection,&readfds)){
      if(XEventsQueued(display,QueuedAfterReading)) return TRUE;
      }
    }
#endif
  
  // No events
  return FALSE;
  }


// Dispatch event to widget
void FXApp::dispatchEvent(XEvent& ev){
  FXWindow *window,*shell;
  FXuint    ks1,ks2;
  Window    tmp;
  XEvent    se;
  if(XFindContext(display,ev.xany.window,wcontext,(XPointer*)&window)==0){
    FXASSERT(window);
    shell=window->getShell();
    FXASSERT(shell);
    
    // Translate it
    switch(ev.xany.type){

      // Composite expose events into rectangle
      case Expose:
      case GraphicsExpose:
        event.type=SEL_PAINT;
        event.window=ev.xexpose.window;
        event.rect.x=ev.xexpose.x;
        event.rect.y=ev.xexpose.y;
        event.rect.w=ev.xexpose.width;
        event.rect.h=ev.xexpose.height;
        event.synthetic=ev.xexpose.send_event;
        //window->setClipRectangle(event.x,event.y,event.w,event.h);
        window->handle(this,MKUINT(0,SEL_PAINT),&event);
        //window->clearClipRectangle();
        break;

      // Keyboard
      case KeyPress:
      case KeyRelease:
        event.type=SEL_KEYPRESS+ev.xkey.type-KeyPress;
        event.window=ev.xkey.window;
        event.time=ev.xkey.time;
        event.win_x=ev.xkey.x;
        event.win_y=ev.xkey.y;
        event.root_x=ev.xkey.x_root;
        event.root_y=ev.xkey.y_root;
        event.state=ev.xkey.state;
        ks1=XLookupKeysym(&ev.xkey,0);
        ks2=XLookupKeysym(&ev.xkey,1);
        if((event.state&NUMLOCKMASK) && (KEY_KP_Space<=ks2) && (ks2<=KEY_KP_Equal)){  // This is still doubtfull!
          if(event.state&SHIFTMASK) event.code=ks1; else event.code=ks2;
          }
        else if((KEY_a<=ks1) && (ks1<=KEY_z)){            // Alpha shifted and/or caplocked
          if(((event.state&CAPSLOCKMASK)!=0) ^ ((event.state&SHIFTMASK)!=0)) event.code=ks2; else event.code=ks1;
          }
        else if((event.state&SHIFTMASK) && (ks2!=0)){     // Simple shifted
          event.code=ks2;
          }
        else{                                             // Unshifted
          event.code=ks1;
          }
// {char buf[20]; KeySym sym; static XComposeStatus cs;
// XLookupString(&ev.xkey,buf,20,&sym,&cs);
// FXTRACE((10,"KEY: keysym=0x%04x state=%04x ks1=%x ks2=%x sym=%x\n",event.code,event.state,ks1,ks2,sym));
// }
        if(event.type==KeyPress){
          if(!event.moved && (event.time-event.click_time<clickSpeed) && (event.window==event.click_window) && (event.code==event.click_button)){
            event.click_count++;
            event.click_time=event.time;
            }
          else{
            event.click_count=1;
            event.click_x=event.win_x;
            event.click_y=event.win_y;
            event.click_window=event.window;
            event.click_button=event.code;
            event.click_time=event.time;
            }
          if(!(event.state&(LEFTBUTTONMASK|MIDDLEBUTTONMASK|RIGHTBUTTONMASK))) event.moved=0;
          }
        if(shell->passModalEvents()){
          if(shell->handle(this,MKUINT(0,event.type),&event)) refresh();
          }
        else if(ev.xany.type==KeyPress){
          beep();
          }
        event.last_x=event.win_x;
        event.last_y=event.win_y;
        break;
        
      // Motion
      case MotionNotify:
        event.type=SEL_MOTION;
        event.window=ev.xmotion.window;
        event.time=ev.xmotion.time;
        event.win_x=ev.xmotion.x;
        event.win_y=ev.xmotion.y;
        event.root_x=ev.xmotion.x_root;
        event.root_y=ev.xmotion.y_root;
        event.code=0;
        event.state=ev.xmotion.state;
        if(!event.moved && (FXABS(ev.xmotion.x-event.click_x)>=dragDelta || FXABS(ev.xmotion.y-event.click_y)>=dragDelta)){
          event.moved=1;
          }
        if(grabWindow){
          if(grabWindow->handle(this,MKUINT(0,event.type),&event)) refresh();
          }
        else if(window->passModalEvents()){
          if(window->handle(this,MKUINT(0,SEL_MOTION),&event)) refresh();
          }
        event.last_x=event.win_x;
        event.last_y=event.win_y;
        break;
        
      // Button
      case ButtonPress:
      case ButtonRelease:
        event.window=ev.xbutton.window;
        event.time=ev.xbutton.time;
        event.win_x=ev.xbutton.x;
        event.win_y=ev.xbutton.y;
        event.root_x=ev.xbutton.x_root;
        event.root_y=ev.xbutton.y_root;
        event.code=ev.xbutton.button;
        event.state=ev.xbutton.state;
        if(ev.xbutton.type==ButtonPress){
          event.type=SEL_LEFTBUTTONPRESS+((ev.xbutton.button-1)<<1);
          if(!event.moved && (event.time-event.click_time<clickSpeed) && (event.window==event.click_window) && (event.code==event.click_button)){
            event.click_count++;
            event.click_time=event.time;
            }
          else{
            event.click_count=1;
            event.click_x=event.win_x;
            event.click_y=event.win_y;
            event.click_window=event.window;
            event.click_button=event.code;
            event.click_time=event.time;
            }
          if(!(event.state&(LEFTBUTTONMASK|MIDDLEBUTTONMASK|RIGHTBUTTONMASK))) event.moved=0;
          }
        else if(ev.xbutton.type==ButtonRelease){
          event.type=SEL_LEFTBUTTONRELEASE+((ev.xbutton.button-1)<<1);
          }
        if(grabWindow){
          if(grabWindow->handle(this,MKUINT(0,event.type),&event)) refresh();
          }
        else if(window->passModalEvents()){
          if(window->handle(this,MKUINT(0,event.type),&event)) refresh();
          }
        else if(ev.xany.type==ButtonPress){
          beep();
          }
        event.last_x=event.win_x;
        event.last_y=event.win_y;
        break;

      // Crossing
      case EnterNotify:
      case LeaveNotify:
        if(ev.xcrossing.mode==NotifyGrab || ev.xcrossing.mode==NotifyUngrab || (ev.xcrossing.mode==NotifyNormal && ev.xcrossing.detail!=NotifyInferior)){
          event.type=SEL_ENTER+ev.xany.type-EnterNotify;
          event.time=ev.xcrossing.time;
          event.win_x=ev.xcrossing.x;
          event.win_y=ev.xcrossing.y;
          event.root_x=ev.xcrossing.x_root;
          event.root_y=ev.xcrossing.y_root;
          event.window=ev.xcrossing.window;
          event.code=ev.xcrossing.mode;
          if(window->handle(this,MKUINT(0,event.type),&event)) refresh();
          event.last_x=event.win_x;
          event.last_y=event.win_y;
          }
        break;

      // Focus
      case FocusIn:
      case FocusOut:
        event.type=SEL_FOCUSIN+ev.xfocus.type-FocusIn;
        event.window=ev.xfocus.window;
        event.code=ev.xfocus.mode;
        if(shell->handle(this,MKUINT(0,event.type),&event)) refresh();
        break;

      // Map
      case MapNotify:
        event.type=SEL_MAP;
        event.window=ev.xmap.window;
        window->handle(this,MKUINT(0,SEL_MAP),&event);
        break;

      // Unmap
      case UnmapNotify:
        event.type=SEL_UNMAP;
        event.window=ev.xunmap.window;
        window->handle(this,MKUINT(0,SEL_UNMAP),&event);
        break;

      // Create
      case CreateNotify:
        event.type=SEL_CREATE;
        event.window=ev.xcreatewindow.window;
        event.rect.x=ev.xcreatewindow.x;
        event.rect.y=ev.xcreatewindow.y;
        event.rect.w=ev.xcreatewindow.width;
        event.rect.h=ev.xcreatewindow.height;
        if(window->handle(this,MKUINT(0,SEL_CREATE),&event)) refresh();
        break;

      // Destroy
      case DestroyNotify:
        event.type=SEL_DESTROY;
        event.window=ev.xdestroywindow.window;
        if(window->handle(this,MKUINT(0,SEL_DESTROY),&event)) refresh();
        break;

      // Configure
      case ConfigureNotify:
        event.type=SEL_CONFIGURE;
        event.window=ev.xconfigure.window;
        event.rect.x=ev.xconfigure.x;
        event.rect.y=ev.xconfigure.y;
        event.rect.w=ev.xconfigure.width;
        event.rect.h=ev.xconfigure.height;
        if(window->handle(this,MKUINT(0,SEL_CONFIGURE),&event)) refresh();
        break;

      // Selection Clear
      case SelectionClear:
        
        // Lost primary selection:- unhighlight
        if(ev.xselectionclear.selection==XA_PRIMARY){
          event.type=SEL_SELECTION_LOST;
          event.window=ev.xselectionclear.window;
          event.time=ev.xselectionclear.time;
          if(window->handle(this,MKUINT(0,SEL_SELECTION_LOST),&event)) refresh();
          }
        
        // Lost clipboard; widget should delete clipped data
        else if(ev.xselectionclear.selection==cbSelection){
          event.type=SEL_SELECTION_REQUEST;
          event.window=ev.xselectionclear.window;
          event.time=ev.xselectionclear.time;
          event.requestor=None;
          event.property=ddeAtom;
          event.target=ddeDelete;                   // Target is DELETE
          event.origin=FROM_CLIPBOARD;
          if(window->handle(this,MKUINT(0,SEL_SELECTION_REQUEST),&event)) refresh();
          }
        
        break;

      // Selection Request
      case SelectionRequest:
        event.type=SEL_SELECTION_REQUEST;
        event.window=ev.xselectionrequest.owner;
        event.time=ev.xselectionrequest.time;
        event.requestor=ev.xselectionrequest.requestor;
        event.property=ev.xselectionrequest.property;
        event.target=ev.xselectionrequest.target;
        event.origin=FROM_UNKNOWN;
        if(ev.xselectionrequest.selection==XA_PRIMARY) event.origin=FROM_SELECTION;
        if(ev.xselectionrequest.selection==cbSelection) event.origin=FROM_CLIPBOARD;
        if(ev.xselectionrequest.selection==xdndSelection) event.origin=FROM_DRAGNDROP;
        if(window->handle(this,MKUINT(0,SEL_SELECTION_REQUEST),&event)) refresh();
        se.xselection.type=SelectionNotify;
        se.xselection.property=event.property; // May be None
        se.xselection.send_event=TRUE;
        se.xselection.display=display;
        se.xselection.requestor=ev.xselectionrequest.requestor;
        se.xselection.selection=ev.xselectionrequest.selection;
        se.xselection.target=ev.xselectionrequest.target;
        se.xselection.time=event.time;
        XSendEvent(display,ev.xselectionrequest.requestor,True,NoEventMask,&se);
        break;

      // Circulate 
      case CirculateNotify:
        event.type=SEL_RAISED+(ev.xcirculate.place&1);
        event.window=ev.xcirculate.window;
        window->handle(this,MKUINT(0,event.type),&event);
        break;

      // Visibility
      case VisibilityNotify:
        event.type=SEL_VISIBILITY;
        event.window=ev.xvisibility.window;
        event.code=ev.xvisibility.state;
        window->handle(this,MKUINT(0,SEL_VISIBILITY),&event);
        break;

      // Keyboard mapping
      case MappingNotify:
        if(ev.xmapping.request!=MappingPointer) XRefreshKeyboardMapping(&ev.xmapping);
        break;

      // Client message
      case ClientMessage:
        if(ev.xclient.format==32){
        
          // WM_PROTOCOLS
          if(ev.xclient.message_type==wmProtocols){
            if((FXID)ev.xclient.data.l[0]==wmDeleteWindow){
              event.type=SEL_CLOSE;
              event.window=ev.xclient.window;
              window->handle(this,MKUINT(0,SEL_CLOSE),&event);
              }
            else if((FXID)ev.xclient.data.l[0]==wmQuitApp){
              event.type=SEL_CLOSE;
              event.window=ev.xclient.window;
              window->handle(this,MKUINT(0,SEL_CLOSE),&event);
              }
            }
          
          // XDND Enter from source
          else if(ev.xclient.message_type==xdndEnter){
            FXTRACE((100,"DNDEnter from remote\n"));
            event.type=SEL_DND_ENTER;
            event.window=ev.xclient.window;
            xdndTarget=ev.xclient.window;                             // Target of the message
            xdndSource=ev.xclient.data.l[0];                          // Source window
            xdndVersion=((FXuint)ev.xclient.data.l[1])>>24;           // Protocol version
            if(ev.xclient.data.l[1]&1){                               // More than 3 types
              unsigned long ni,b; int fmt; Atom typ;
              unsigned char* data=NULL;
              if(XGetWindowProperty(display,xdndSource,xdndTypes,0,1024,False,XA_ATOM,&typ,&fmt,&ni,&b,&data)!=Success) return;
              if(!data) return;
              if(typ!=XA_ATOM || fmt!=32 || ni==0){ XFree(data); return; }
              xdndTypeList=(FXID*)data;
              xdndNumTypes=ni;
              }
            else{                                                     // If types if less than 3
              xdndTypeList=(FXID*)malloc(3*sizeof(FXID));
              xdndNumTypes=0;
              if(ev.xclient.data.l[2]){xdndTypeList[0]=ev.xclient.data.l[2];xdndNumTypes++;}
              if(ev.xclient.data.l[3]){xdndTypeList[1]=ev.xclient.data.l[3];xdndNumTypes++;}
              if(ev.xclient.data.l[4]){xdndTypeList[2]=ev.xclient.data.l[4];xdndNumTypes++;}
              }
            if(window->handle(this,MKUINT(0,SEL_DND_ENTER),&event)) refresh();
            }
          
          // XDND Leave from source
          else if(ev.xclient.message_type==xdndLeave){
            FXTRACE((100,"DNDLeave from remote\n"));
            event.type=SEL_DND_LEAVE;
            event.window=ev.xclient.window;
            xdndTarget=ev.xclient.window;
            xdndSource=ev.xclient.data.l[0];                          // Source window
            if(window->handle(this,MKUINT(0,SEL_DND_LEAVE),&event)) refresh();
            if(xdndTypeList) XFree(xdndTypeList);
            xdndTypeList=NULL;
            xdndNumTypes=0;
            xdndVersion=XDND_PROTOCOL_VERSION;                        // Done, restore stuff
            xdndTarget=0;
            xdndSource=0;
            }
          
          // XDND Drop from source
          else if(ev.xclient.message_type==xdndDrop){
            FXTRACE((100,"DNDDrop from remote\n"));
            event.type=SEL_DND_DROP;
            event.window=ev.xclient.window;
            xdndTarget=ev.xclient.window;
            xdndSource=ev.xclient.data.l[0];                          // Source window
            event.time=ev.xclient.data.l[2];
            if(window->handle(this,MKUINT(0,SEL_DND_DROP),&event)) refresh();
            se.xclient.type=ClientMessage;                            // Drop window has finished processing of the drop
            se.xclient.display=display;
            se.xclient.message_type=xdndFinished;
            se.xclient.format=32;
            se.xclient.window=xdndSource;
            se.xclient.data.l[0]=xdndTarget;                          // Target window, i.e. me!
            se.xclient.data.l[1]=0;
            se.xclient.data.l[2]=0;
            se.xclient.data.l[3]=0;
            se.xclient.data.l[4]=0;
            XSendEvent(display,xdndSource,True,NoEventMask,&se);
            if(xdndTypeList) XFree(xdndTypeList);
            xdndTypeList=NULL;
            xdndNumTypes=0;
            xdndVersion=XDND_PROTOCOL_VERSION;
            xdndTarget=0;
            xdndSource=0;
            }
          
          // XDND Position from source
          else if(ev.xclient.message_type==xdndPosition){
            FXTRACE((100,"DNDPosition from remote\n"));
            event.type=SEL_DND_MOTION;
            event.window=ev.xclient.window;
            xdndTarget=ev.xclient.window;
            xdndSource=ev.xclient.data.l[0];                          // Source window
            event.root_x=((FXuint)ev.xclient.data.l[2])>>16;
            event.root_y=((FXuint)ev.xclient.data.l[2])&0xffff;
            event.time=ev.xclient.data.l[3];
            xdndAction=ev.xclient.data.l[4];                          // Drag and Drop Action
            xdndAccepts=FALSE;                                        // If message is not handled, we assume no drop!
            XTranslateCoordinates(display,XDefaultRootWindow(display),event.window,event.root_x,event.root_y,&event.win_x,&event.win_y,&tmp);
            if(window->handle(this,MKUINT(0,SEL_DND_MOTION),&event)) refresh();
            event.last_x=event.win_x;
            event.last_y=event.win_y;
            se.xclient.type=ClientMessage;
            se.xclient.display=display;
            se.xclient.message_type=xdndStatus;
            se.xclient.format=32;
            se.xclient.window=xdndSource;
            se.xclient.data.l[0]=xdndTarget;                          // Target window, i.e. me!
            se.xclient.data.l[1]=0;
            if(xdndAccepts) se.xclient.data.l[1]|=1;                  // Target accepted
            if(xdndWantUpdates) se.xclient.data.l[1]|=2;              // Target wants continuous position updates
            se.xclient.data.l[2]=MKUINT(xdndRect.y,xdndRect.x);
            se.xclient.data.l[3]=MKUINT(xdndRect.h,xdndRect.w);
            se.xclient.data.l[4]=xdndAction;                          // Type of Drag and Drop Action accepted
            XSendEvent(display,xdndSource,True,NoEventMask,&se);
            }
          
          // XDND Status from target
          else if(ev.xclient.message_type==xdndStatus){
            FXTRACE((100,"DNDStatus from %s remote\n",xdndTarget==ev.xclient.data.l[0] ? "active" : "old"));

            // Ignore status messages which are (no longer) applicable
            if(xdndTarget==(FXID)ev.xclient.data.l[0]){
              xdndAccepts=ev.xclient.data.l[1]&1;
              xdndAction=ev.xclient.data.l[4];                        // Drag and Drop Action
              xdndWantUpdates=(ev.xclient.data.l[1]&2)!=0;
              xdndRect.x=((FXuint)ev.xclient.data.l[2])>>16;
              xdndRect.y=((FXuint)ev.xclient.data.l[2])&0xffff;
              xdndRect.w=((FXuint)ev.xclient.data.l[3])>>16;
              xdndRect.h=((FXuint)ev.xclient.data.l[3])&0xffff;
              xdndStatusReceived=TRUE;
              xdndStatusPending=FALSE;
              if(xdndSendPosition){                                   // Long overdue position update must now be sent
                se.xclient.type=ClientMessage;
                se.xclient.display=display;
                se.xclient.message_type=xdndPosition;
                se.xclient.format=32;
                se.xclient.window=xdndTarget;
                se.xclient.data.l[0]=xdndSource;                      // Source window
                se.xclient.data.l[1]=0;
                se.xclient.data.l[2]=MKUINT(xdndYPos,xdndXPos);       // Coordinates
                se.xclient.data.l[3]=event.time;                      // Time stamp
                se.xclient.data.l[4]=xdndAction;                      // Drag and Drop action
                XSendEvent(display,xdndTarget,True,NoEventMask,&se);
                xdndSendPosition=FALSE;
                xdndStatusPending=TRUE;                               // We're now waiting for the next status message!
                }
              }
            }
          }
        break;

      // Others
      default:
        break;
      }
    }
  }

#else

void FXApp::getNextEvent(MSG& msg){
  while(TRUE){
    while(!PeekMessage(&msg,NULL,0,0,PM_NOREMOVE)){
    
      // GUI updating
      if(refresher){
      
	// Call the refresher handler
	refresher->handle(this,MKUINT(0,SEL_UPDATE),&event);
      
	// Next widget to refresh
	if(refresher->getFirst()){
	  refresher=refresher->getFirst();
	  continue;
	  }
	while(!refresher->getNext() && refresher->getParent()!=root){
	  refresher=refresher->getParent();
	  }
	refresher=refresher->getNext();
	continue;
	}
    
      // Chores
      if(chores){
	FXChore* c=chores;
	chores=c->next;
      
	// Call the chore handler
	if(c->target && c->target->handle(this,MKUINT(c->message,SEL_CHORE),&event)) refresh();
      
	// Add record to list of recycled chore records
	c->next=chorerecs;
	chorerecs=c;
	continue;
	}

      // No updates or chores pending, so block
      break;
      }
      // Get an event
      GetMessage(&msg,NULL,0,0);

      // Compress motion events
      if(msg.message==WM_MOUSEMOVE){
	MSG m;
	while(PeekMessage(&m,NULL,0,0,PM_NOREMOVE)){
	  if(m.message!=WM_MOUSEMOVE || m.hwnd!=msg.hwnd) break;
	  GetMessage(&msg,NULL,0,0);
	  }
	}

      // Regular event
      return;
    }
  }

void FXApp::dispatchEvent(MSG& msg){
  TranslateMessage(&msg);
  DispatchMessage(&msg);
  }

#endif

/*******************************************************************************/



// Flush pending repaints
void FXApp::flush(){
#ifndef FX_NATIVE_WIN32
  FXRepaint *r;
  XEvent ev;
   
  // First compound extant expose events
  while(XCheckMaskEvent(display,ExposureMask,&ev)){
    addRepaint(ev.xexpose.window,ev.xexpose.x,ev.xexpose.y,ev.xexpose.width,ev.xexpose.height,0);
    }
  
  // Then process them
  while(repaints){
    r=repaints;
    ev.xany.type=Expose;
    ev.xexpose.window=r->window;
    ev.xexpose.x=r->rect.x;
    ev.xexpose.y=r->rect.y;
    ev.xexpose.width=r->rect.w-r->rect.x;
    ev.xexpose.height=r->rect.h-r->rect.y;
    repaints=r->next;
    r->next=repaintrecs;
    repaintrecs=r;
    dispatchEvent(ev);
    }
  
  // Flush the buffers to the server
  XFlush(display);
#else
  FXASSERT(FALSE);
#endif
  }


// Force GUI refresh
void FXApp::forceRefresh(){
  refresher=root;
  while(refresher){
    refresher->handle(this,MKUINT(0,SEL_UPDATE),&event);
    if(refresher->getFirst()){
      refresher=refresher->getFirst();
      continue;
      }
    while(!refresher->getNext() && refresher->getParent()!=root){
      refresher=refresher->getParent();
      }
    refresher=refresher->getNext();
    }
  }


// Perform refresh some time
void FXApp::refresh(){
  refresher=root;
  }


// Run application
FXint FXApp::run(){
  runUntil(done);
  return exitcode;
  }


// Break out of topmost event loop
void FXApp::stop(FXint code){
  exitcode=code;
  done=1;
  }
  

// Run till some flag becomes non-zero
void FXApp::runUntil(FXuint& condition){
#ifndef FX_NATIVE_WIN32
  XEvent ev;
  while(condition==0){
    getNextEvent(ev);
    dispatchEvent(ev);
    }
#else
  MSG msg;
  while(condition==0){
    getNextEvent(msg);
    dispatchEvent(msg);
    }
#endif
  }


// Run modal for window
FXint FXApp::runModalFor(FXWindow* window){
  FXInvocation inv;
  if(window==NULL || window->id()==0){ fxerror("%s::runModalFor: illegal window specified.\n"); }
  inv.upper=invocation;
  inv.window=window->id();
  inv.done=FALSE;
  inv.code=0;
  invocation=&inv;
#ifndef FX_NATIVE_WIN32
  XEvent ev;
  while(!inv.done){
    getNextEvent(ev);
    dispatchEvent(ev);
    }
#else
  MSG msg;
  while(!inv.done){
    getNextEvent(msg);
    dispatchEvent(msg);
    }
#endif
  invocation=inv.upper;
  return inv.code;
  }


// Break out of current modal loop
void FXApp::stopModal(FXint value){
  if(invocation==NULL){ fxerror("%s::stopModal: not in a modal loop.\n",getClassName()); }
  invocation->done=TRUE;
  invocation->code=value;
  }


// Run modal while window is shown, or until stopModal is called
FXint FXApp::runModalWhileShown(FXWindow* window){
  FXInvocation inv;
  if(window==NULL || window->id()==0){ fxerror("%s::runModalWhileShown: illegal window specified.\n"); }
  inv.upper=invocation;
  inv.window=window->id();
  inv.done=FALSE;
  inv.code=0;
  invocation=&inv;
#ifndef FX_NATIVE_WIN32
  XEvent ev;
  while(!inv.done && window->shown()){
    getNextEvent(ev);
    dispatchEvent(ev);
    }
#else
  MSG msg;
  while(!inv.done && window->shown()){
    getNextEvent(msg);
    dispatchEvent(msg);
    }
#endif
  invocation=inv.upper;
  return inv.code;
  }


// Test if the window is involved in the innermost modal invocation
FXbool FXApp::isModal(FXWindow *window) const {
  if(window==NULL || window->id()==0){ fxerror("%s::isModal: illegal window specified.\n"); }
  return invocation && invocation->window==window->id();
  }

  
// Initialize application
void FXApp::init(int& argc,char** argv){
#ifndef FX_NATIVE_WIN32
  char *dpy=NULL;
#endif
  int i,j;

  // Verify implementation invariants
  FXASSERT(sizeof(FXuchar)==1);
  FXASSERT(sizeof(FXbool)==1);
  FXASSERT(sizeof(FXchar)==1);
  FXASSERT(sizeof(FXushort)==2);
  FXASSERT(sizeof(FXshort)==2);
  FXASSERT(sizeof(FXuint)==4);
  FXASSERT(sizeof(FXint)==4);
  FXASSERT(sizeof(FXfloat)==4);
  FXASSERT(sizeof(FXdouble)==8);
  
  FXASSERT(sizeof(FXString)==sizeof(FXchar*));
  
  // Long is not always available on all implementations
#ifdef FX_LONG
  FXASSERT(sizeof(FXulong)==8);
  FXASSERT(sizeof(FXlong)==8);
#endif

  // No shared memory with Windoze
#ifndef FX_NATIVE_WIN32
#ifdef WIN32
  shmi=FALSE;
  shmp=FALSE;
#endif
#endif
  
  // Parse out FOX args
  i=j=1;
  while(j<argc){
#ifndef FX_NATIVE_WIN32
    // Start synchronized mode
    if(strcmp(argv[j],"-sync")==0){
      synchronize=TRUE;
      j++;
      continue;
      }
    
    // Start synchronized mode
    if(strcmp(argv[j],"-noshm")==0){
      shmi=FALSE;
      shmp=FALSE;
      j++;
      continue;
      }
    
    // Alternative display
    if(strcmp(argv[j],"-display")==0){
      j++;
      if(j>=argc){ fxerror("%s:init: missing argument for -display.\n",getClassName()); }
      dpy=argv[j];
      j++;
      continue;
      }
    
#endif
    
    // Set trace level
    if(strcmp(argv[j],"-tracelevel")==0){
      j++;
      if(j>=argc){ fxerror("%s:init: missing argument for -tracelevel.\n",getClassName()); }
      if(sscanf(argv[j],"%d",&fxTraceLevel)!=1){ fxerror("%s::init: expected trace level number.\n",getClassName()); }
      j++;
      continue;
      }
    
    // Copy program arguments
    argv[i++]=argv[j++];
    }
  
  // Adjust argment count
  argv[i]=NULL;
  argc=i;
  
  // Open Display for X-Windows
#ifndef FX_NATIVE_WIN32
  openDisplay(dpy);
#endif
  
  // Read the registry
  registry.read();
  
  // Change some settings
  clickSpeed=registry.readEntry("SETTINGS","clickspeed",400);
  scrollSpeed=registry.readEntry("SETTINGS","scrollspeed",100);
  blinkSpeed=registry.readEntry("SETTINGS","blinkspeed",500);
  menuPause=registry.readEntry("SETTINGS","menupause",400);
  tooltipPause=registry.readEntry("SETTINGS","tippause",800);
  tooltipTime=registry.readEntry("SETTINGS","tiptime",3000);
  scrollbarWidth=registry.readEntry("SETTINGS","scrollbarwidth",15);
  dragDelta=registry.readEntry("SETTINGS","dragdelta",6);
  }


// Exit application
void FXApp::exit(FXint code){
  
  // Write the registry
  registry.write();
  
#ifdef FX_NATIVE_WIN32
  DdeUninitialize(idInst);
#endif

  // Exit the program
  ::exit(code);
  }
  

// Create application's windows
void FXApp::create(){ 
#ifndef FX_NATIVE_WIN32
  if(display==NULL){ fxerror("%s::create(): Should open display first\n",getClassName()); }
#endif

  // Create default font
  normalFont->create();
  
  // Create cursors
  arrowCursor->create();
  rarrowCursor->create();
  textCursor->create();
  hsplitCursor->create();
  vsplitCursor->create();
  resizeCursor->create();
  swatchCursor->create();
  dontdropCursor->create();
  moveCursor->create();
  dragHCursor->create();
  dragVCursor->create();
  dragTLCursor->create();
  dragTRCursor->create();
  dndCopyCursor->create();
  dndMoveCursor->create();
  dndLinkCursor->create();
  
  // Create stipple bitmap
  grayBitmap->create();
  
  // Register window classes MS-Windows
#ifdef FX_NATIVE_WIN32
  registerWindowClasses();
#endif

  // Create all windows
  root->create();
  }


// Destroy application's windows
void FXApp::destroy(){
  root->destroy();
  }


#ifdef FX_NATIVE_WIN32

extern int main(int argc,char **argv);

// This is the actual entry point for all FOX apps
int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,PSTR szCmdLine,INT iCmdShow){
  FXApp::hInstance=hInstance;
  return main(__argc,__argv);
  }


// Convert Win32 virtual key codes (and modifiers) to FOX keycaps
static FXuint winkey2foxkey(FXuint state,WPARAM wParam){
  FXuint i;

  // Letters
  if(wParam>=65&&wParam<=90){
    if(((state&CAPSLOCKMASK)!=0) ^ ((state&SHIFTMASK)!=0))
      return wParam+(KEY_A-65);
    else
      return wParam+(KEY_a-KEY_A);
    }

  // Numbers
  FXuint numbersmap[]={
    KEY_parenright,KEY_exclam,KEY_at,KEY_numbersign,KEY_dollar,KEY_percent,
    KEY_asciicircum,KEY_ampersand,KEY_asterisk,KEY_parenleft
    };
  if(wParam>=48&&wParam<=57){
    if(state&SHIFTMASK)
      return numbersmap[wParam-48];
    else
      return KEY_0+(wParam-48);
    }

  // Keypad numbers
  if(wParam>=VK_NUMPAD0&&wParam<=VK_NUMPAD9){
    return KEY_KP_0+(wParam-VK_NUMPAD0);
    }

  // Undocumented keycodes; how portable???
  if(state&SHIFTMASK){
    FXuint codesmap[]={
      0xC0, KEY_asciitilde, 0xBD, KEY_underscore, 0xBB, KEY_plus, 0xDB, KEY_braceleft,
      0xDD, KEY_braceright, 0xDC, KEY_bar, 0xBA, KEY_colon, 0xDE, KEY_quotedbl,
      0xBC, KEY_less, 0xBE, KEY_greater, 0xBF, KEY_question
      };
    for(i=0;i<ARRAYNUMBER(codesmap)/2;i++)
      if(codesmap[2*i]==wParam) return codesmap[2*i+1];
    }
  else{
    FXuint codesmap[]={
      0xC0, KEY_grave, 0xBD, KEY_minus, 0xBB, KEY_equal, 0xDB, KEY_bracketleft,
      0xDD, KEY_bracketright, 0xDC, KEY_backslash, 0xBA, KEY_semicolon, 0xDE, KEY_apostrophe,
      0xBC, KEY_comma, 0xBE, KEY_period, 0xBF, KEY_slash
      };
    for(i=0;i<ARRAYNUMBER(codesmap)/2;i++)
      if(codesmap[2*i]==wParam) return codesmap[2*i+1];
    }

  // Function keys
  if(wParam>=VK_F1&&wParam<=VK_F16){
    return KEY_F1+(wParam-VK_F1);
    }

  // Left or Right Shift Key
  if(wParam==VK_SHIFT){
    return (GetKeyState(VK_LSHIFT)<0) ? KEY_Shift_L : KEY_Shift_R;
    }

  // Left or Right Alt key
  if(wParam==VK_MENU){
    return (GetKeyState(VK_RMENU)<0) ? KEY_Alt_L : KEY_Alt_R;
    }

  // Left or Right Ctrl key
  if(wParam==VK_CONTROL){
    return (GetKeyState(VK_LCONTROL)<0) ? KEY_Control_L : KEY_Control_R;
    }

  // Pause/Break key
  if(wParam==VK_PAUSE){
    return (GetKeyState(VK_SHIFT)<0) ? KEY_Pause : KEY_Break;
    }

  // Miscellaneous
  FXuint miscmap[]={
    VK_SPACE, KEY_space,
    VK_BACK, KEY_BackSpace,
    VK_TAB, KEY_Tab,
    VK_RETURN, KEY_Return,
    VK_SCROLL, KEY_Scroll_Lock,
    VK_ESCAPE, KEY_Escape,
    VK_INSERT, KEY_Insert,
    VK_DELETE, KEY_Delete,
    VK_HOME, KEY_Home,
    VK_LEFT, KEY_Left,
    VK_UP, KEY_Up,
    VK_RIGHT, KEY_Right,
    VK_DOWN, KEY_Down,
    VK_PRIOR, KEY_Prior,
    VK_NEXT, KEY_Next,
    VK_END, KEY_End,
    VK_CAPITAL, KEY_Caps_Lock,
    VK_NUMLOCK, KEY_Num_Lock,
    VK_MULTIPLY, KEY_KP_Multiply,
    VK_ADD, KEY_KP_Add,
    VK_SEPARATOR, KEY_KP_Separator,
    VK_SUBTRACT, KEY_KP_Subtract,
    VK_DECIMAL, KEY_KP_Decimal,
    VK_DIVIDE, KEY_KP_Divide,
    VK_SNAPSHOT, KEY_Print,
    };
  for(i=0;i<ARRAYNUMBER(miscmap)/2;i++)
    if(miscmap[2*i]==wParam) return miscmap[2*i+1];

  fxwarning("Caught unrecognized key code [%4x]\n",wParam);

  return KEY_VoidSymbol;
  }


// DDE callback function
HDDEDATA FXApp::ddeCallback(UINT iType,UINT iFmt,HCONV hConv,HSZ hsz1,HSZ hsz2,HDDEDATA hData,DWORD dwData1,DWORD dwData2){
  char szBuffer[32];
  switch(iType){
    case XTYP_CONNECT:
      // Check application name
      DdeQueryString(idInst,hsz2,szBuffer,sizeof(szBuffer),0);
      if(strcmp(szBuffer,appname)!=0)
	return FALSE;

      // Check topic name
      DdeQueryString(idInst,hsz1,szBuffer,sizeof(szBuffer),0);
      if(strcmp(szBuffer,"Topic")!=0)
	return FALSE;

      // Connection approved
      return (HDDEDATA) TRUE;
      break;
    case XTYP_ADVSTART:
    case XTYP_REQUEST:
    case XTYP_ADVREQ:
    case XTYP_ADVSTOP:
      FXASSERT(FALSE);
      break;
    case XTYP_REGISTER:
    case XTYP_UNREGISTER:
      // We do not expect these messages since we specified the CBF_SKIP_REGISTRATIONS
      // and CBF_SKIP_UNREGISTRATIONS options in the call to DdeInitialize()
      FXASSERT(FALSE);
      break;
    default:
      FXASSERT(FALSE);
    }
  return NULL;
  }


// This window procedure is a static member function of class FXApp.
// Its sole purpose is to forward the message info on to FXApp::dispatchEvent().
LRESULT CALLBACK FXApp::wndproc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam){
  return app->dispatchEvent(hwnd,iMsg,wParam,lParam);
  }


// Message dispatching
LRESULT FXApp::dispatchEvent(HWND hwnd,UINT iMsg,WPARAM wParam,LPARAM lParam){
  HDC hdc;
  PAINTSTRUCT ps;
  LPCREATESTRUCT pCS;
  FXWindow *window,*shell;
  POINT pt;
  BOOL fShow;
  RECT rect;
  UINT iFormat;
	
  window=(FXWindow*)GetWindowLong(hwnd,0);
  if(window) shell=window->getShell();

  // Translate Win32 message to FOX message type
  switch(iMsg){
    
    // Composite expose events into rectangle
    case WM_PAINT:
      hdc=BeginPaint(hwnd,&ps);
      event.type=SEL_PAINT;
      event.window=hwnd;
      event.rect.x=(FXshort)ps.rcPaint.left;
      event.rect.y=(FXshort)ps.rcPaint.top;
      event.rect.w=(FXshort)(ps.rcPaint.right-ps.rcPaint.left+1);
      event.rect.h=(FXshort)(ps.rcPaint.bottom-ps.rcPaint.top+1);
      event.synthetic=1;		// When in doubt, assume synthetic!
      window->handle(this,MKUINT(0,SEL_PAINT),&event);
      EndPaint(hwnd, &ps);
      return 0;

    // Keyboard
    case WM_KEYDOWN:
    case WM_SYSKEYDOWN:
    case WM_KEYUP:
    case WM_SYSKEYUP:
      event.type=((iMsg==WM_KEYDOWN)||(iMsg==WM_SYSKEYDOWN))?SEL_KEYPRESS:SEL_KEYRELEASE;
      event.window=hwnd;
      event.time=GetMessageTime();
      GetCursorPos(&pt);
      event.root_x=pt.x;
      event.root_y=pt.y;
      ScreenToClient(hwnd,&pt);
      event.win_x=pt.x;
      event.win_y=pt.y;
      event.state=fxmodifierkeys();

      // Translate virtual key code
      event.code=winkey2foxkey(event.state,wParam);

      if(event.type==SEL_KEYPRESS){
        if(!event.moved && (event.time-event.click_time<clickSpeed) && (event.window==event.click_window) && (event.code==event.click_button)){
          event.click_count++;
          event.click_time=event.time;
          }
        else{
          event.click_count=1;
          event.click_x=event.win_x;
          event.click_y=event.win_y;
          event.click_window=event.window;
          event.click_button=event.code;
          event.click_time=event.time;
          }
        if(!(event.state&(LEFTBUTTONMASK|MIDDLEBUTTONMASK|RIGHTBUTTONMASK))) event.moved=0;
        }
      if(shell->passModalEvents()){
	if(shell->handle(this,MKUINT(0,event.type),&event)) refresh();
        }
      else if(iMsg==WM_KEYDOWN||iMsg==WM_SYSKEYDOWN){
	beep();
        }
      event.last_x=event.win_x;
      event.last_y=event.win_y;
      return 0;

    // Motion
    case WM_MOUSEMOVE:
      
      // Synthesize enter and leave events first
      if(window!=getCursorWindow()){
        event.type=SEL_ENTER;
        event.time=GetMessageTime();
        event.win_x=pt.x=LOWORD(lParam);
        event.win_y=pt.y=HIWORD(lParam);
        ClientToScreen(hwnd,&pt);
        event.root_x=pt.x;
        event.root_y=pt.y;
        event.window=hwnd;
        event.code=CROSSINGNORMAL;
        if(window->handle(this,MKUINT(0,SEL_ENTER),&event)) refresh();
        event.last_x=event.win_x;
        event.last_y=event.win_y;

#if _WIN32_WINNT >= 0x0400
	// Set up a mouse tracker to catch the leave (Windows NT only)
	TRACKMOUSEEVENT tme;
	tme.cbSize=sizeof(TRACKMOUSEEVENT);
	tme.dwFlags=TME_LEAVE;
	tme.hwndTrack=hwnd;
	tme.dwHoverTime=HOVER_DEFAULT; // Unused in this case
	TrackMouseEvent(&tme);
#else
	// This works as long as the newly entered window
	// still belongs to this application. Otherwise we
	// won't know that we left the last active window.
	event.type=SEL_LEAVE;
	event.time=GetMessageTime();
	event.win_x=pt.x=LOWORD(lParam);
	event.win_y=pt.y=HIWORD(lParam);
	FXASSERT(event.window);
	ClientToScreen((HWND)event.window,&pt);
	event.root_x=pt.x;
	event.root_y=pt.y;
	event.code=CROSSINGNORMAL;
	if(window->handle(this,MKUINT(0,SEL_LEAVE),&event)) refresh();
	event.last_x=event.win_x;
	event.last_y=event.win_y;
#endif
        }

      // Normal motion event
      event.type=SEL_MOTION;
      event.window=hwnd;
      event.time=GetMessageTime();
      event.win_x=pt.x=(int)((short)LOWORD(lParam)); // Position in client coordinates
      event.win_y=pt.y=(int)((short)HIWORD(lParam));
      ClientToScreen(hwnd,&pt);
      event.root_x=pt.x;
      event.root_y=pt.y;
      event.code=0;
      event.state=fxmodifierkeys();

      // Windows seems to generate spurious motion events in some cases.
      // If the mouse position is *exactly* the same as last time, don't
      // generate a new SEL_MOTION message.
      if(event.last_x==event.win_x && event.last_y==event.win_y)
	return 0;

      if(!event.moved && (FXABS(event.win_x-event.click_x)>=dragDelta || FXABS(event.win_y-event.click_y)>=dragDelta)){
        event.moved=1;
        }
      if(grabWindow){
	if(grabWindow->handle(this,MKUINT(0,SEL_MOTION),&event)) refresh();
        }
      else if(window->passModalEvents()){
	if(window->handle(this,MKUINT(0,SEL_MOTION),&event)) refresh();
        }
      event.last_x=event.win_x;
      event.last_y=event.win_y;
      return 0;

    // Button
    case WM_LBUTTONDOWN:
    case WM_MBUTTONDOWN:
    case WM_RBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_MBUTTONUP:
    case WM_RBUTTONUP:
      event.window=hwnd;
      event.time=GetMessageTime();
      event.win_x=pt.x=LOWORD(lParam);
      event.win_y=pt.y=HIWORD(lParam);
      ClientToScreen(hwnd, &pt);
      event.root_x=pt.x;
      event.root_y=pt.y;
      event.state=fxmodifierkeys();   
      switch(iMsg){
        case WM_LBUTTONDOWN:
          event.type=SEL_LEFTBUTTONPRESS;
          event.state&=~LEFTBUTTONMASK;         // Was not down till now
          event.code=1;
          break;
        case WM_MBUTTONDOWN:
          event.type=SEL_MIDDLEBUTTONPRESS;
          event.state&=~MIDDLEBUTTONMASK;       // Was not down till now
          event.code=2;
          break;
        case WM_RBUTTONDOWN:
          event.type=SEL_RIGHTBUTTONPRESS;
          event.state&=~RIGHTBUTTONMASK;        // Was not down till now
          event.code=3;
          break;
        case WM_LBUTTONUP:
          event.type=SEL_LEFTBUTTONRELEASE;
          event.code=1;
          break;
        case WM_MBUTTONUP:
          event.type=SEL_MIDDLEBUTTONRELEASE;
          event.code=2;
          break;
        case WM_RBUTTONUP:
          event.type=SEL_RIGHTBUTTONRELEASE;
          event.code=3;
          break;
        }
      if(iMsg==WM_LBUTTONDOWN || iMsg==WM_MBUTTONDOWN || iMsg==WM_RBUTTONDOWN){
        if(!event.moved && (event.time-event.click_time<clickSpeed) && (event.window==event.click_window) && (event.code==event.click_button)){
          event.click_count++;
          event.click_time=event.time;
          }
        else{
          event.click_count=1;
          event.click_x=event.win_x;
          event.click_y=event.win_y;
          event.click_window=event.window;
          event.click_button=event.code;
          event.click_time=event.time;
          }
        if(!(event.state&(LEFTBUTTONMASK|MIDDLEBUTTONMASK|RIGHTBUTTONMASK))) event.moved=0;
	}
      else if(iMsg==WM_LBUTTONUP || iMsg==WM_MBUTTONUP || iMsg==WM_RBUTTONUP){
        }
      if(grabWindow){
	if(grabWindow->handle(this,MKUINT(0,event.type),&event)) refresh();
        }
      else if(window->passModalEvents()){
	if(window->handle(this,MKUINT(0,event.type),&event)) refresh();
        }
      else if(iMsg==WM_LBUTTONDOWN || iMsg==WM_MBUTTONDOWN || iMsg==WM_RBUTTONDOWN){
	beep();
        }
      event.last_x=event.win_x;
      event.last_y=event.win_y;
      return 0;

#if _WIN32_WINNT >= 0x0400
    // Crossing
    case WM_MOUSELEAVE:
      event.type=SEL_LEAVE;
      event.time=GetMessageTime();
      event.win_x=pt.x=LOWORD(lParam);
      event.win_y=pt.y=HIWORD(lParam);
      FXASSERT(event.window);
      ClientToScreen((HWND)event.window,&pt);
      event.root_x=pt.x;
      event.root_y=pt.y;
      event.code=CROSSINGNORMAL;
      if(window->handle(this,MKUINT(0,SEL_LEAVE),&event)) refresh();
      event.last_x=event.win_x;
      event.last_y=event.win_y;
      return 0;
#endif

    // Focus
    case WM_SETFOCUS:
    case WM_KILLFOCUS:
      event.type=(iMsg==WM_SETFOCUS)?SEL_FOCUSIN:SEL_FOCUSOUT;
      event.window=hwnd;
      event.code=0;		// ?? NotifyNormal, NotifyGrab, NotifyUngrab
      if(shell->handle(this,MKUINT(0,event.type),&event)) refresh();
      return 0;

    // Map or Unmap
    case WM_SHOWWINDOW:
      fShow=(BOOL)wParam;
      event.type=fShow ? SEL_MAP : SEL_UNMAP;
      event.window=hwnd;
      window->handle(this,MKUINT(0,event.type),&event);
      return 0;

    // Create
    case WM_CREATE:
      pCS=(LPCREATESTRUCT)lParam;
      window=static_cast<FXWindow*>(pCS->lpCreateParams);
      event.type=SEL_CREATE;
      event.window=hwnd;
      event.rect.x=pCS->x;
      event.rect.y=pCS->y;
      event.rect.w=pCS->cx;
      event.rect.h=pCS->cy;
      SetWindowLong(hwnd,0,reinterpret_cast<LONG>(pCS->lpCreateParams));
      if(window->handle(this,MKUINT(0,SEL_CREATE),&event)) refresh();
      return 0;

    // Destroy
    case WM_DESTROY:
      
      ////// Semantics: SEL_CLOSE is a suggestion that the window be closed;
      ////// SEL_DESTROY is a notify that destruction has already taken place.
      ////// Thus, a toplevel window gets a close, and MAY decide not to be closed;
      ////// If it does, it gets a SEL_DESTROY also.
      if(window->isMemberOf(&FXTopWindow::metaClass)){
        event.type=SEL_CLOSE;
        event.window=hwnd;
        window->handle(this,MKUINT(0,SEL_CLOSE),&event);
        }
      else{
        event.type=SEL_DESTROY;
        event.window=hwnd;
        if(window->handle(this,MKUINT(0,SEL_DESTROY),&event)) refresh();
        }
      return 0;

    // Configure (size)
    case WM_SIZE:
      event.window=hwnd;
      event.rect.x=0;
      event.rect.y=0;
      event.rect.w=LOWORD(lParam);
      event.rect.h=HIWORD(lParam);
      if(window->handle(this,MKUINT(0,SEL_CONFIGURE),&event)) refresh();
      return 0;

    case WM_SIZING:
      event.window=hwnd;
      GetClientRect(hwnd,&rect);
      event.rect.x=0;
      event.rect.y=0;
      event.rect.w=(FXshort)(rect.right-rect.left);
      event.rect.h=(FXshort)(rect.bottom-rect.top);
      if(window->handle(this,MKUINT(0,SEL_CONFIGURE),&event)) refresh();
      return 0;

    // Configure (move)
    case WM_MOVE:
    case WM_MOVING:
      event.window=hwnd;
      event.rect.x=LOWORD(lParam);
      event.rect.y=HIWORD(lParam);
      GetClientRect(hwnd,&rect);
      event.rect.w=(FXshort)rect.right;
      event.rect.h=(FXshort)rect.bottom;
      if(window->handle(this,MKUINT(0,SEL_CONFIGURE),&event)) refresh();
      return 0;

    case WM_ERASEBKGND:
      // Do nothing in response to this message to avoid flashing during redraws...
      return 0;

    case WM_STYLECHANGING:
    case WM_STYLECHANGED:
      return DefWindowProc(hwnd,iMsg,wParam,lParam);

    // Notify the grab window, then break the capture
    case WM_CANCELMODE:
      event.window=hwnd;
      if(grabWindow){
	if(grabWindow->handle(this,MKUINT(0,SEL_UNGRABBED),&event)) refresh();
        }
      return DefWindowProc(hwnd,iMsg,wParam,lParam);

    // Lost clipboard ownership
    case WM_DESTROYCLIPBOARD:
      event.type=SEL_SELECTION_REQUEST;
      event.window=hwnd;
      event.time=GetMessageTime();
      event.requestor=NULL; // No requestor; system sends this message when window loses clipboard ownership
      event.property=NULL;
      event.target=ddeDelete;                   // Target is DELETE
      event.origin=FROM_CLIPBOARD;
      if(window->handle(this,MKUINT(0,SEL_SELECTION_REQUEST),&event)) refresh();
      return 0;
    
    // Safeguard it in the clipboard
    case WM_RENDERALLFORMATS:
      OpenClipboard(hwnd);
      EmptyClipboard();
      iFormat=0;
      while(iFormat=EnumClipboardFormats(iFormat)){
	event.type=SEL_SELECTION_REQUEST;
	event.window=hwnd;
	event.time=GetMessageTime();
	event.requestor=NULL; // No requestor; system sends this message when window is destroyed
	event.property=NULL;
	event.target=iFormat;
	event.origin=FROM_CLIPBOARD;
	if(window->handle(this,MKUINT(0,SEL_SELECTION_REQUEST),&event)) refresh();
	}
      event.origin=FROM_UNKNOWN;
      CloseClipboard();
      return 0;
      
    // We're asked to provide certain format to the clipboard
    case WM_RENDERFORMAT:
      event.window=hwnd;
      event.type=SEL_SELECTION_REQUEST;
      event.window=hwnd;
      event.time=GetMessageTime();
      event.requestor=NULL;
      event.property=NULL;
      event.target=wParam; // Desired format
      event.origin=FROM_CLIPBOARD;
      if(window->handle(this,MKUINT(0,SEL_SELECTION_REQUEST),&event)) refresh();
      event.origin=FROM_UNKNOWN;
      return 0;
    
    case WM_SETCURSOR:
      if(window->getDefaultCursor()&&LOWORD(lParam)==HTCLIENT){
	SetCursor((HCURSOR)window->getDefaultCursor()->id());
	return 0;
	}
      return DefWindowProc(hwnd,iMsg,wParam,lParam);

    // Selection Clear
    // Selection Request
    // Circulate
    // Visibility
    // Keyboard Mapping
    // Client message
    case WM_NCMOUSEMOVE:
    case WM_NCCREATE:
    case WM_NCHITTEST:
    case WM_NCCALCSIZE:
    case WM_NCPAINT:
    case WM_NCLBUTTONDOWN:
    case WM_NCACTIVATE:
    case WM_NCDESTROY:
      return DefWindowProc(hwnd,iMsg,wParam,lParam);
    case WM_ENTERSIZEMOVE:
    case WM_EXITSIZEMOVE:
      return DefWindowProc(hwnd,iMsg,wParam,lParam);
    case WM_CHAR:
    case WM_PARENTNOTIFY:
    case WM_GETMINMAXINFO:
    case WM_SETTEXT:
    case WM_GETTEXT:
    case WM_SYSCOMMAND:
    case WM_QUERYOPEN:
    case WM_QUERYNEWPALETTE:
    case WM_ENABLE:
    case WM_CAPTURECHANGED:
    case WM_WINDOWPOSCHANGED:
    case WM_WINDOWPOSCHANGING:
    case WM_ACTIVATE:
    case WM_MOUSEACTIVATE:
    case WM_CHILDACTIVATE:
    case WM_ACTIVATEAPP:
    case WM_CLOSE:
      return DefWindowProc(hwnd,iMsg,wParam,lParam);
    default:
      // XDND message types are not constants, so we can't put them in a switch() block
      if(iMsg==xdndEnter){
        return DefWindowProc(hwnd,iMsg,wParam,lParam);
        }
      else if(iMsg==xdndLeave){
        return DefWindowProc(hwnd,iMsg,wParam,lParam);
        }
      else if(iMsg==xdndPosition){
        return DefWindowProc(hwnd,iMsg,wParam,lParam);
        }
      else if(iMsg==xdndStatus){
        return DefWindowProc(hwnd,iMsg,wParam,lParam);
        }
      else if(iMsg==xdndDrop){
        return DefWindowProc(hwnd,iMsg,wParam,lParam);
        }
      else if(iMsg==xdndFinished){
        return DefWindowProc(hwnd,iMsg,wParam,lParam);
        }
      else{
        fxwarning("Caught unrecognized message %4x for %p\n",iMsg,window);
        }
      return DefWindowProc(hwnd,iMsg,wParam,lParam);
    }
  }

#endif


// Handle quit
long FXApp::onCmdQuit(FXObject*,FXSelector,void*){
  exit(0); 
  return 1;
  }


// Register DND type
FXDragType FXApp::registerDragType(const FXString& name) const {
#ifndef FX_NATIVE_WIN32
  if(display==NULL){ fxerror("%s::registerDragType: Should open display first.\n",getClassName()); }
  if(!name.text()){ fxerror("%s::registerDragType: NULL name argument.\n",getClassName()); }
  return (FXDragType)XInternAtom(display,name.text(),0);
#else
  return RegisterClipboardFormat(name.text());
#endif
  }


// Get name of registered drag type
FXString FXApp::getDragTypeName(FXDragType type) const {
#ifndef FX_NATIVE_WIN32
  if(display==NULL){ fxerror("%s::getDragTypeName: Should open display first\n",getClassName()); }
  FXchar *name=XGetAtomName(display,type);
  FXString dragtypename(name);
  XFree(name);
  return dragtypename;
#else
  char pBuffer[256];
  GetClipboardFormatName(type,pBuffer,sizeof(pBuffer));
  return FXString(pBuffer);
#endif
  }


/*******************************************************************************/


// Show a modal error message
FXuint FXApp::showModalErrorBox(FXuint opts,const char* title,const char* message,...){
  const unsigned char errorIcon[]={
    0x47,0x49,0x46,0x38,0x37,0x61,0x20,0x00,0x20,0x00,0xf1,0x00,0x00,0x80,0x80,0x80,
    0x80,0x00,0x00,0xff,0x00,0x00,0xff,0xff,0xff,0x2c,0x00,0x00,0x00,0x00,0x20,0x00,
    0x20,0x00,0x00,0x02,0x91,0x84,0x8f,0x10,0xcb,0x9b,0x0f,0x0f,0x13,0xb4,0x52,0x16,
    0x73,0xb0,0x9c,0x87,0x9c,0x2c,0xdd,0x58,0x39,0xe0,0x46,0xa6,0x97,0xa6,0xb6,0xc2,
    0x07,0xa1,0x2e,0x09,0x87,0xdd,0xa0,0xe2,0x9e,0xcd,0x0d,0x3e,0xe9,0xd3,0x59,0x6a,
    0x8a,0x5e,0x50,0x58,0x39,0x22,0x5f,0x08,0x59,0xf2,0x68,0x51,0x76,0x6a,0xce,0x28,
    0x54,0x20,0x1d,0xc1,0xaa,0x56,0x65,0x30,0xb5,0xcd,0x79,0x97,0xbb,0xa2,0xf8,0x0a,
    0x36,0x70,0x8d,0x68,0x9a,0xfa,0x9c,0x75,0x9b,0x47,0xe3,0xaf,0x7c,0x8d,0xbd,0xc6,
    0xcb,0xf3,0xa7,0x9d,0xb2,0xb7,0x22,0xc1,0xf6,0x43,0xb8,0x44,0xc4,0xf5,0xc7,0x36,
    0xc5,0x13,0x95,0xc3,0xf8,0x80,0x37,0xc3,0x14,0x33,0x29,0x57,0x69,0x39,0x04,0xa2,
    0x20,0xa9,0x45,0x74,0x32,0xf9,0xb9,0xc9,0x89,0x87,0x31,0x1a,0xd1,0xd0,0x70,0xba,
    0xca,0xda,0xea,0x6a,0x50,0x00,0x00,0x3b
    };
  FXGIFIcon icon(this,errorIcon);
  va_list arguments;
  FXchar buf[1000];
  va_start(arguments,message);
#ifdef _GNU_SOURCE
  vsnprintf(buf,sizeof(buf),message,arguments);
#else
  vsprintf(buf,message,arguments);
#endif
  va_end(arguments);
  FXMessageBox box(this,title,buf,&icon,opts|DECOR_TITLE|DECOR_BORDER|DIALOG_MODAL);
  return box.execute();
  }


// Show a modal warning message
FXuint FXApp::showModalWarningBox(FXuint opts,const char* title,const char* message,...){
  const unsigned char warningIcon[]={
    0x47,0x49,0x46,0x38,0x37,0x61,0x20,0x00,0x20,0x00,0xf2,0x00,0x00,0x80,0x80,0x80,
    0x80,0x80,0x00,0xff,0xff,0x00,0xc0,0xc0,0xc0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x20,0x00,0x20,0x00,0x00,0x03,
    0x8f,0x08,0xba,0xbc,0xf1,0x2d,0xca,0xa9,0x82,0x10,0x83,0xd0,0x2d,0xed,0xc5,0x1a,
    0x27,0x02,0xde,0x17,0x8e,0x54,0xf9,0x65,0x68,0xfa,0xbd,0x67,0xcb,0xa8,0x2f,0x2b,
    0xcf,0x6f,0x1e,0xcb,0x74,0x6e,0xdf,0xa5,0x0c,0x81,0x30,0x30,0xdd,0x48,0xb0,0xe1,
    0x70,0xb5,0x13,0xa9,0x94,0x4b,0x63,0x8b,0x06,0x25,0xd4,0x9a,0xae,0xa4,0x52,0x87,
    0xea,0x55,0x7d,0xd8,0x48,0x2f,0xa0,0xf4,0x3a,0x73,0x17,0x21,0x11,0xfd,0x9b,0xf4,
    0x2e,0x5b,0xb4,0x20,0x5c,0x91,0x5f,0xc8,0x6f,0x50,0x56,0x9e,0x29,0xda,0xc3,0x79,
    0x26,0x76,0x69,0x58,0x81,0x83,0x5c,0x0d,0x86,0x20,0x6b,0x76,0x6d,0x48,0x87,0x50,
    0x90,0x38,0x87,0x73,0x71,0x8d,0x27,0x8a,0x8b,0x7e,0x83,0x98,0x94,0x9e,0x30,0x0e,
    0x9f,0x9f,0x6d,0x99,0x9e,0x8e,0x24,0x55,0xa9,0xaa,0xa9,0x47,0xad,0xae,0x22,0x09,
    0x00,0x3b
    };
  FXGIFIcon icon(this,warningIcon);
  va_list arguments;
  FXchar buf[1000];
  va_start(arguments,message);
#ifdef _GNU_SOURCE
  vsnprintf(buf,sizeof(buf),message,arguments);
#else
  vsprintf(buf,message,arguments);
#endif
  va_end(arguments);
  FXMessageBox box(this,title,buf,&icon,opts|DECOR_TITLE|DECOR_BORDER|DIALOG_MODAL);
  return box.execute();
  }


// Show a modal question dialog
FXuint FXApp::showModalQuestionBox(FXuint opts,const char* title,const char* message,...){
  const unsigned char questionIcon[]={
    0x47,0x49,0x46,0x38,0x37,0x61,0x20,0x00,0x20,0x00,0xf2,0x00,0x00,0x80,0x80,0x80,
    0xc0,0xc0,0xc0,0xff,0xff,0xff,0x00,0x00,0xff,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x20,0x00,0x20,0x00,0x00,0x03,
    0x97,0x08,0xba,0xdc,0xfe,0xf0,0x05,0x41,0x6b,0x88,0x58,0xcd,0xca,0xf9,0xcd,0xcc,
    0xd6,0x8d,0x16,0x08,0x90,0x02,0x30,0x0c,0x80,0xc8,0x11,0x18,0x19,0xac,0xf4,0xe0,
    0x56,0xf0,0x23,0xdb,0x54,0x7d,0x53,0xb9,0x10,0x69,0xc5,0xa9,0xc9,0x82,0x0a,0x94,
    0x51,0xb0,0x1c,0x21,0x7f,0x3d,0x9a,0x60,0x26,0x3d,0x2e,0x50,0xd3,0xd5,0xa4,0x49,
    0x0a,0x62,0x3d,0x55,0xac,0xf7,0x1b,0x1d,0x90,0x05,0x63,0xb2,0xf6,0x9c,0xfe,0x12,
    0xd9,0x0b,0xe8,0xd9,0x7a,0xfd,0x52,0xe5,0xb8,0x06,0x9e,0x19,0xa6,0x33,0xdc,0x7d,
    0x4e,0x0f,0x04,0x58,0x77,0x62,0x11,0x84,0x73,0x82,0x18,0x04,0x7b,0x23,0x01,0x48,
    0x88,0x8d,0x67,0x90,0x91,0x19,0x04,0x98,0x3f,0x95,0x96,0x26,0x0a,0x98,0x93,0x40,
    0x9d,0x26,0x89,0x68,0xa2,0x20,0xa4,0x9c,0xa6,0x0d,0x89,0xa9,0xaa,0x0c,0x84,0xad,
    0xae,0x0b,0x98,0xb2,0xb5,0xb6,0xaa,0x09,0x00,0x3b
    };
  FXGIFIcon icon(this,questionIcon);
  va_list arguments;
  FXchar buf[1000];
  va_start(arguments,message);
#ifdef _GNU_SOURCE
  vsnprintf(buf,sizeof(buf),message,arguments);
#else
  vsprintf(buf,message,arguments);
#endif
  va_end(arguments);
  FXMessageBox box(this,title,buf,&icon,opts|DECOR_TITLE|DECOR_BORDER|DIALOG_MODAL);
  return box.execute();
  }


// Show a modal information dialog
FXuint FXApp::showModalInformationBox(FXuint opts,const char* title,const char* message,...){
  const unsigned char infoIcon[]={
    0x47,0x49,0x46,0x38,0x37,0x61,0x20,0x00,0x20,0x00,0xf2,0x00,0x00,0x80,0x80,0x80,
    0xc0,0xc0,0xc0,0xff,0xff,0xff,0x00,0x00,0xff,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x20,0x00,0x20,0x00,0x00,0x03,
    0x8c,0x08,0xba,0xdc,0xfe,0xf0,0x05,0x41,0x6b,0x88,0x58,0xcd,0xca,0xf9,0xcd,0xcc,
    0xd6,0x8d,0x16,0x08,0x90,0xc1,0xa0,0x0e,0x22,0x47,0x60,0xa4,0xb0,0xae,0xf1,0xfb,
    0xc4,0xf2,0x3c,0xe0,0x76,0x88,0xa7,0xab,0x56,0x27,0xd0,0x53,0xe0,0x8e,0xa4,0xa2,
    0x10,0x79,0x24,0x2e,0x8e,0x3a,0x15,0xb2,0xc7,0xd4,0x4d,0x9f,0x48,0xeb,0x91,0x9a,
    0x9d,0x5d,0x8d,0x5d,0xda,0x76,0xb1,0xec,0x68,0x63,0x4e,0x2c,0xee,0x9c,0x6c,0x94,
    0x2b,0xec,0x61,0x11,0x3c,0x8a,0x8a,0x3b,0x73,0x05,0x81,0x39,0x86,0xec,0xf9,0x23,
    0x79,0x0d,0x04,0x6f,0x68,0x82,0x0e,0x04,0x84,0x4c,0x44,0x87,0x7e,0x89,0x4b,0x8c,
    0x8d,0x20,0x89,0x8a,0x14,0x92,0x26,0x0c,0x7f,0x02,0x97,0x98,0x7a,0x15,0x9c,0x9d,
    0x00,0x7f,0xa0,0xa1,0x7b,0xa4,0xa5,0xa7,0xa1,0xaa,0xab,0x19,0x09,0x00,0x3b
    };
  FXGIFIcon icon(this,infoIcon);
  va_list arguments;
  FXchar buf[1000];
  va_start(arguments,message);
#ifdef _GNU_SOURCE
  vsnprintf(buf,sizeof(buf),message,arguments);
#else
  vsprintf(buf,message,arguments);
#endif
  va_end(arguments);
  FXMessageBox box(this,title,buf,&icon,opts|DECOR_TITLE|DECOR_BORDER|DIALOG_MODAL);
  return box.execute();
  }


// Beep
void FXApp::beep(){
#ifndef FX_NATIVE_WIN32
  XBell(display,0);
#else
  MessageBeep(0);
#endif
  }


// Check if we're saving as expected
void FXApp::save(FXStream& store) const {
  FXObject::save(store);
  store << clickSpeed;
  store << scrollSpeed;
  store << blinkSpeed;
  store << menuPause;
  store << tooltipPause;
  store << tooltipTime;
  store << scrollbarWidth;
  store << dragDelta;
  store << backColor;
  store << foreColor;
  store << hiliteColor;
  store << shadowColor;
  store << borderColor;
  store << selforeColor;
  store << selbackColor;
  }


// Check if we're loading as expected
void FXApp::load(FXStream& store){ 
  FXObject::load(store);
  store >> clickSpeed;
  store >> scrollSpeed;
  store >> blinkSpeed;
  store >> menuPause;
  store >> tooltipPause;
  store >> tooltipTime;
  store >> scrollbarWidth;
  store >> dragDelta;
  store >> backColor;
  store >> foreColor;
  store >> hiliteColor;
  store >> shadowColor;
  store >> borderColor;
  store >> selforeColor;
  store >> selbackColor;
  }


// Virtual destructor
FXApp::~FXApp(){
  app=NULL;
  }
