/********************************************************************************
*                                                                               *
*                         C o l o r W e l l   C l a s s                         *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXColorWell.cpp,v 1.46 1999/01/15 05:29:25 jeroen Exp $                  *
********************************************************************************/
#include "xincs.h"
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDC.h"
#include "FXFont.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXComposite.h"
#include "FXLabel.h"
#include "FXColorWell.h"

/*
  To do:
  - Check drop types.
  - Need to popup color dialog on double click.
  - Is there any reason why one wouldn't ^C ^V on the clipboard colors same as text?
*/

#define WELLSIZE    12              // Minimum well size
#define FOCUSBORDER 3               // Focus border


/*******************************************************************************/


// Map
FXDEFMAP(FXColorWell) FXColorWellMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXColorWell::onPaint),
  FXMAPFUNC(SEL_FOCUSIN,0,FXColorWell::onFocusIn),
  FXMAPFUNC(SEL_FOCUSOUT,0,FXColorWell::onFocusOut),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,0,FXColorWell::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,0,FXColorWell::onLeftBtnRelease),
  FXMAPFUNC(SEL_UNGRABBED,0,FXColorWell::onUngrabbed),
  FXMAPFUNC(SEL_MOTION,0,FXColorWell::onMotion),
  FXMAPFUNC(SEL_DND_ENTER,0,FXColorWell::onDNDEnter),
  FXMAPFUNC(SEL_DND_LEAVE,0,FXColorWell::onDNDLeave),
  FXMAPFUNC(SEL_DND_DROP,0,FXColorWell::onDNDDrop),
  FXMAPFUNC(SEL_DND_MOTION,0,FXColorWell::onDNDMotion),
  FXMAPFUNC(SEL_SELECTION_REQUEST,0,FXColorWell::onSelectionRequest),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_QUERY_TIP,FXColorWell::onQueryTip),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_QUERY_HELP,FXColorWell::onQueryHelp),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_SETVALUE,FXColorWell::onCmdSetValue),
  };


// Object implementation
FXIMPLEMENT(FXColorWell,FXCell,FXColorWellMap,ARRAYNUMBER(FXColorWellMap))


/*******************************************************************************/


// Init
FXColorWell::FXColorWell(){
  wellColor[0]=wellColor[1]=0;
  rgba=0;
  }

  
// Make a color well
FXColorWell::FXColorWell(FXComposite* p,FXColor clr,FXObject* tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h,FXint pl,FXint pr,FXint pt,FXint pb):
  FXCell(p,opts,x,y,w,h,pl,pr,pt,pb){
  flags|=FLAG_ENABLED;
  target=tgt;
  message=sel;
  rgba=clr;
  wellColor[0]=rgbaoverwhite(rgba);
  wellColor[1]=rgbaoverblack(rgba);
  dragCursor=getApp()->swatchCursor;
  }


// Create X window
void FXColorWell::create(){
  FXCell::create();
  if(!targetType){targetType=getApp()->registerDragType(targetTypeName);}
  if(!colorType){colorType=getApp()->registerDragType(colorTypeName);}
  FXTRACE((110,"targetType=%08x\n",targetType));
  FXTRACE((110,"colorType=%08x\n",colorType));
  dropEnable();
  show();
  }


// If window can have focus
FXbool FXColorWell::canFocus() const { return 1; }


// Compute color over black
FXColor FXColorWell::rgbaoverblack(FXColor clr){
  FXint r,g,b,mul=FXALPHAVAL(clr);
  r=(FXREDVAL(clr)*mul)/255;
  g=(FXGREENVAL(clr)*mul)/255;
  b=(FXBLUEVAL(clr)*mul)/255;
  return FXRGB(r,g,b);
  }


// Compute color over white
FXColor FXColorWell::rgbaoverwhite(FXColor clr){
  FXint r,g,b,mul=FXALPHAVAL(clr);
  r=((255-mul)*255+FXREDVAL(clr)*mul)/255;
  g=((255-mul)*255+FXGREENVAL(clr)*mul)/255;
  b=((255-mul)*255+FXBLUEVAL(clr)*mul)/255;
  return FXRGB(r,g,b);
  }


// Get default size
FXint FXColorWell::getDefaultWidth(){
  return WELLSIZE+FOCUSBORDER+padleft+padright+4;
  }


FXint FXColorWell::getDefaultHeight(){
  return WELLSIZE+FOCUSBORDER+padtop+padbottom+4;
  }


// Handle repaint 
long FXColorWell::onPaint(FXObject*,FXSelector,void* ptr){
  FXEvent *ev=(FXEvent*)ptr;
  FXPoint points[3];
  FXDC *dc=DC();
  dc->begin(this);
  dc->setClipRectangle(ev->rect.x,ev->rect.y,ev->rect.w,ev->rect.h);
  dc->setForeground(backColor);
  dc->fillRectangle(0,0,width,padtop+FOCUSBORDER);
  dc->fillRectangle(0,padtop+FOCUSBORDER,padleft+FOCUSBORDER,height-padtop-padbottom-(FOCUSBORDER<<1));
  dc->fillRectangle(width-padright-FOCUSBORDER,padtop+FOCUSBORDER,padright+FOCUSBORDER,height-padtop-padbottom-(FOCUSBORDER<<1));
  dc->fillRectangle(0,height-padbottom-FOCUSBORDER,width,padbottom+FOCUSBORDER);
  dc->setForeground(wellColor[0]);
  points[0].x=points[1].x=padleft+FOCUSBORDER+2; 
  points[2].x=width-padright-FOCUSBORDER-2;
  points[0].y=points[2].y=padtop+FOCUSBORDER+2;
  points[1].y=height-padbottom-FOCUSBORDER-2;
  dc->fillPolygon(points,3);
  dc->setForeground(wellColor[1]);
  points[0].x=padleft+FOCUSBORDER+2; 
  points[1].x=points[2].x=width-padright-FOCUSBORDER-2;
  points[0].y=points[1].y=height-padbottom-FOCUSBORDER-2;
  points[2].y=padtop+FOCUSBORDER+2;
  dc->fillPolygon(points,3);
  drawDoubleSunkenRectangle(padleft+FOCUSBORDER,padtop+FOCUSBORDER,width-padright-padleft-(FOCUSBORDER<<1),height-padbottom-padtop-(FOCUSBORDER<<1));
  if(hasFocus()){
    drawFocusRectangle(padleft,padtop,width-padright-padleft,height-padbottom-padtop);
    }
  dc->clearClipRectangle();
  dc->end();
  return 1;
  }


// Gained focus
long FXColorWell::onFocusIn(FXObject* sender,FXSelector sel,void* ptr){
  FXCell::onFocusIn(sender,sel,ptr);
  update(0,0,width,height);
  return 1;
  }

  
// Lost focus
long FXColorWell::onFocusOut(FXObject* sender,FXSelector sel,void* ptr){
  FXCell::onFocusOut(sender,sel,ptr);
  update(0,0,width,height);
  return 1;
  }


// Handle drag-and-drop enter
long FXColorWell::onDNDEnter(FXObject* sender,FXSelector sel,void* ptr){
  FXCell::onDNDEnter(sender,sel,ptr);
//   FXuchar *data;
//   FXuint   len;
//   if(getDNDData(FROM_DRAGNDROP,FXColorWell::targetType,data,len)){
// for(int i=0; i<len; i+=4){ FXTRACE((10,"type=%08x\n",*((FXDragType*)&data[i]))); }
//     if(*((FXDragType*)data) == FXColorWell::colorType){
//       acceptDrop(DRAG_COPY);
//       return 1;
//       }
//     }
  return 1;
  }

// Handle drag-and-drop leave
long FXColorWell::onDNDLeave(FXObject* sender,FXSelector sel,void* ptr){
  FXCell::onDNDLeave(sender,sel,ptr);
  return 1;
  }


// Handle drag-and-drop motion
long FXColorWell::onDNDMotion(FXObject* sender,FXSelector sel,void* ptr){
  
  // Handle base class first
  if(FXCell::onDNDMotion(sender,sel,ptr)) return 1;
  
  // No more messages while inside
  setDragRectangle(0,0,width,height,FALSE);
  
  // Is it a color being dropped?
  if(offeredDNDType(FXColorWell::colorType)){
    acceptDrop(DRAG_COPY);
    return 1;
    }
  return 0;
  }


// Handle drag-and-drop drop
long FXColorWell::onDNDDrop(FXObject* sender,FXSelector sel,void* ptr){
  FXuchar *data; FXuint len,r,g,b,a;
  
  // Try handling it in base class first
  if(FXCell::onDNDDrop(sender,sel,ptr)) return 1;
  
  // Try handle here
  if(getDNDData(FROM_DRAGNDROP,colorType,data,len)){
    sscanf((char*)data,"#%02x%02x%02x%02x",&r,&g,&b,&a);
    setRGBA(FXRGBA(r,g,b,a));
    FXFREE(&data);
    if(target) target->handle(this,MKUINT(message,SEL_COMMAND),(void*)rgba); 
    return 1;
    }
  return 0;
  }


// Somebody wants our selection
long FXColorWell::onSelectionRequest(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  FXDragType *type;
  FXchar *value;
  
  // Try handling it in base class first
  if(FXCell::onSelectionRequest(sender,sel,ptr)) return 1;
  
  // When asked, supply the list of types
  if(event->target==targetType){
    FXTRACE((100,"%s: Get types list\n",getClassName()));
    FXMALLOC(&type,FXDragType,1);
    type[0]=colorType;
    setDNDData(event->origin,targetType,(FXuchar*)type,sizeof(FXDragType));
    return 1;
    }
    
  // Try handle here
  if(event->target==colorType){
    FXTRACE((100,"%s: Get color\n",getClassName()));
    FXMALLOC(&value,FXchar,50);
    sprintf(value,"#%02x%02x%02x%02x",FXREDVAL(rgba),FXGREENVAL(rgba),FXBLUEVAL(rgba),FXALPHAVAL(rgba));
    setDNDData(event->origin,colorType,(FXuchar*)value,50);
    return 1;
    }
  return 0;
  }


// Drag start
long FXColorWell::onLeftBtnPress(FXObject*,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  flags&=~FLAG_TIP;
  if(isEnabled()){
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONPRESS),ptr)) return 1;
    handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
    if(event->click_count==1){
      beginDrag(&FXColorWell::colorType,1);
      flags&=~FLAG_UPDATE;
      flags|=FLAG_PRESSED;
      }
    }
  return 1;
  }


// Drop 
long FXColorWell::onLeftBtnRelease(FXObject*,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  if(isEnabled()){
    ungrab();
    flags&=~FLAG_PRESSED;
    flags|=FLAG_UPDATE;
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr)) return 1;
    if(event->click_count==1){
      endDrag(didAccept()==DRAG_COPY);
      }
    else{
      }
// {
//   FXEvent *ev=(FXEvent*)ptr;
//   XEvent se;
//   FXint tox,toy;
//   Window tmp;
//   Window www=getWindowAt(ev->root_x,ev->root_y);
//   
//   XTranslateCoordinates(getDisplay(),XDefaultRootWindow(getDisplay()),www,ev->root_x,ev->root_y,&tox,&toy,&tmp);
// 
//   se.xbutton.type=ButtonPress;
//   se.xbutton.serial=0;
//   se.xbutton.send_event=1;
//   se.xbutton.display=getDisplay();
//   se.xbutton.window=www;
//   se.xbutton.root=XDefaultRootWindow(getDisplay());
//   se.xbutton.subwindow=None;
//   se.xbutton.time=ev->time;
//   se.xbutton.x=tox;
//   se.xbutton.y=toy;
//   se.xbutton.x_root=ev->root_x;
//   se.xbutton.y_root=ev->root_y;
//   se.xbutton.state=0;
//   se.xbutton.button=2;
//   se.xbutton.same_screen=TRUE;
// fprintf(stderr,"SendEvent to window %d\n",se.xbutton.window);
//   XSendEvent(getDisplay(),se.xbutton.window,True,NoEventMask,&se);
//   
//   se.xbutton.type=ButtonRelease;
//   se.xbutton.serial=0;
//   se.xbutton.send_event=1;
//   se.xbutton.display=getDisplay();
//   se.xbutton.window=www;
//   se.xbutton.root=XDefaultRootWindow(getDisplay());
//   se.xbutton.subwindow=None;
//   se.xbutton.time=ev->time+1;
//   se.xbutton.x=tox;
//   se.xbutton.y=toy;
//   se.xbutton.x_root=ev->root_x;
//   se.xbutton.y_root=ev->root_y;
//   se.xbutton.state=Button2Mask;
//   se.xbutton.button=2;
//   se.xbutton.same_screen=TRUE;
// fprintf(stderr,"SendEvent to window %d\n",se.xbutton.window);
//   XSendEvent(getDisplay(),se.xbutton.window,True,NoEventMask,&se);
// }
    }
  return 1;
  }


// The widget lost the grab for some reason
long FXColorWell::onUngrabbed(FXObject* sender,FXSelector sel,void* ptr){
  FXCell::onUngrabbed(sender,sel,ptr);
  flags&=~FLAG_PRESSED;
  flags|=FLAG_UPDATE;
  endDrag(FALSE);
  return 1;
  }


// Moving 
long FXColorWell::onMotion(FXObject*,FXSelector,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  if(isEnabled()){
    if(isDragging()){
      handleDrag(event->root_x,event->root_y,DRAG_COPY);
      if(didAccept()==DRAG_COPY){
        setDragCursor(getApp()->swatchCursor);
        }
      else{
        setDragCursor(getApp()->dontdropCursor);
        }
      }
    }
  return 1;
  }


// We were asked about status text
long FXColorWell::onQueryHelp(FXObject* sender,FXSelector,void*){
  if(help.text() && (flags&FLAG_HELP)){
    sender->handle(this,MKUINT(ID_SETSTRINGVALUE,SEL_COMMAND),(void*)&help);
    return 1;
    }
  return 0;
  }


// We were asked about tip text
long FXColorWell::onQueryTip(FXObject* sender,FXSelector,void*){
  if(tip.text() && (flags&FLAG_TIP)){
    sender->handle(this,MKUINT(ID_SETSTRINGVALUE,SEL_COMMAND),(void*)&tip);
    return 1;
    }
  return 0;
  }


// Change help text
void FXColorWell::setHelpText(const FXString& text){
  help=text;
  }


// Change tip text
void FXColorWell::setTipText(const FXString& text){
  tip=text;
  }


// Change RGBA color 
void FXColorWell::setRGBA(FXColor clr){
  if(clr!=rgba){
    rgba=clr;
    wellColor[0]=rgbaoverwhite(rgba);
    wellColor[1]=rgbaoverblack(rgba);
    update(0,0,width,height);
    }
  }


// Set color
long FXColorWell::onCmdSetValue(FXObject*,FXSelector,void* ptr){
  setRGBA((FXColor)ptr);
  return 1;
  }


// Get RGBA color
FXColor FXColorWell::getRGBA() const { 
  return rgba; 
  }


// Save data
void FXColorWell::save(FXStream& store) const {
  FXCell::save(store);
  store << wellColor[0] << wellColor[1];
  store << rgba;
  store << tip;
  store << help;
  }


// Load data
void FXColorWell::load(FXStream& store){ 
  FXCell::load(store);
  store >> wellColor[0] >> wellColor[1];
  store >> rgba;
  store >> tip;
  store >> help;
  }
