/********************************************************************************
*                                                                               *
*                       C o m b o   B o x   O b j e c t                         *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXComboBox.cpp,v 1.9 1999/01/15 04:54:12 jeroen Exp $                    *
********************************************************************************/
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXObjectList.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDrawable.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXLabel.h"
#include "FXTextField.h"
#include "FXButton.h"
#include "FXArrowButton.h"
#include "FXComposite.h"
#include "FXPacker.h"
#include "FXShell.h"
#include "FXPopup.h"
#include "FXScrollbar.h"
#include "FXScrollWindow.h"
#include "FXList.h"
#include "FXComboBox.h"

// Private header files
#include "FXListItem.h"


/*
  Notes:
  - Keyboard-driven operation.
  - Pass through styles for left/right align of list, text editable, etc.
  - Should the arrow button really be a menubutton?
*/


/*******************************************************************************/

// Map
FXDEFMAP(FXComboBox) FXComboBoxMap[]={
  FXMAPFUNC(SEL_MOTION,0,FXComboBox::onMotion),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,0,FXComboBox::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,0,FXComboBox::onLeftBtnRelease),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,FXComboBox::ID_LIST,FXComboBox::onListLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,FXComboBox::ID_LIST,FXComboBox::onListLeftBtnRelease),
  FXMAPFUNC(SEL_CLICKED,FXComboBox::ID_LIST,FXComboBox::onListClicked),
  FXMAPFUNC(SEL_COMMAND,FXComboBox::ID_LIST,FXComboBox::onListCommand),
  FXMAPFUNC(SEL_SELECTED,FXComboBox::ID_LIST,FXComboBox::onListSelected),
  FXMAPFUNC(SEL_DESELECTED,FXComboBox::ID_LIST,FXComboBox::onListDeselected),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,FXComboBox::ID_BUTTON,FXComboBox::onButtonDown),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,FXComboBox::ID_BUTTON,FXComboBox::onButtonUp),
  FXMAPFUNC(SEL_COMMAND,FXComboBox::ID_TEXT,FXComboBox::onCmdText),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_POST,FXComboBox::onCmdPost),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_UNPOST,FXComboBox::onCmdUnpost),
  };


// Object implementation
FXIMPLEMENT(FXComboBox,FXPacker,FXComboBoxMap,ARRAYNUMBER(FXComboBoxMap))


// List box
FXComboBox::FXComboBox(FXComposite *p,FXint cols,FXObject* tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h,FXint pl,FXint pr,FXint pt,FXint pb):
  FXPacker(p, opts, x,y,w,h, 0,0,0,0, 0,0){
  flags|=FLAG_ENABLED;
  target=tgt;
  message=sel;
  text=new FXTextField(this,cols,this,FXComboBox::ID_TEXT,0, 0,0,0,0, pl,pr,pt,pb);
  button=new FXArrowButton(this,this,FXComboBox::ID_BUTTON,FRAME_RAISED|FRAME_THICK|ARROW_DOWN, 0,0,0,0, 0,0,0,0);
  pane=new FXPopup(getApp(),this,FRAME_LINE);
  list=new FXList(pane,this,FXComboBox::ID_LIST,LIST_BROWSESELECT|LIST_AUTOSELECT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT|SCROLLERS_TRACK,0,0,10,10);
  dragCursor=getApp()->rarrowCursor;
  }


// Create X window
void FXComboBox::create(){
  FXPacker::create();
  pane->create();
  }


// Destroy X window
void FXComboBox::destroy(){
  pane->destroy();
  FXPacker::destroy();
  }


// Get default width
FXint FXComboBox::getDefaultWidth(){
  FXint ww,pw;
  ww=text->getDefaultWidth()+button->getDefaultWidth();
  pw=pane->getDefaultWidth();
  return FXMAX(ww,pw)+(border<<1);
  }


// Get default height
FXint FXComboBox::getDefaultHeight(){
  return text->getDefaultHeight()+(border<<1);
  }


// Recalculate layout
void FXComboBox::layout(){
  FXint buttonWidth,textWidth,itemHeight;
  itemHeight=height-(border<<1);
  buttonWidth=button->getDefaultWidth();
  textWidth=width-buttonWidth-(border<<1);
  text->position(border,border,textWidth,itemHeight);
  button->position(border+textWidth,border,buttonWidth,itemHeight);
  flags&=~FLAG_DIRTY;
  }


// Post the menu
long FXComboBox::onCmdPost(FXObject*,FXSelector,void*){
  if(pane && !pane->shown()){
    FXint x,y,hh;
    translateCoordinatesTo(x,y,getRoot(),0,0);
    y=y+height;
    hh=list->getContentHeight();
    if(hh>120) hh=120;
    list->setWidth(width);
    list->setHeight(hh);
    pane->popup(this,x,y,pane->getDefaultWidth(),pane->getDefaultHeight());
    if(!grabbed()) grab();
    flags&=~FLAG_UPDATE;
    }
  return 1;
  }


// Unpost the menu
long FXComboBox::onCmdUnpost(FXObject*,FXSelector,void*){
  if(pane && pane->shown()){
    pane->popdown();
    if(grabbed()) ungrab();
    flags|=FLAG_UPDATE;
    }
  return 1;
  }


// Pressed the button
long FXComboBox::onButtonDown(FXObject* sender,FXSelector,void* ptr){
  if(!pane->shown()){
    handle(this,MKUINT(ID_POST,SEL_COMMAND),ptr);
    }
  else{
    handle(this,MKUINT(ID_UNPOST,SEL_COMMAND),ptr);
    }
  return 1;
  }


// Pressed the button
long FXComboBox::onButtonUp(FXObject* sender,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  if(event->moved){ handle(this,MKUINT(ID_UNPOST,SEL_COMMAND),ptr); }
  return 1;
  }


// Pressed left button
long FXComboBox::onLeftBtnPress(FXObject*,FXSelector,void* ptr){
  handle(this,MKUINT(ID_UNPOST,SEL_COMMAND),ptr);
  return 1;
  }


// Released left button
long FXComboBox::onLeftBtnRelease(FXObject*,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  if(event->moved){ handle(this,MKUINT(ID_UNPOST,SEL_COMMAND),ptr); }
  return 1;
  }


// Mousing around
long FXComboBox::onMotion(FXObject*,FXSelector,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  if(pane->shown()){
    if(pane){
      if(pane->contains(ev->root_x,ev->root_y)){
        if(grabbed()) ungrab(); 
        }
      else{
        if(!grabbed()) grab();
        }
      return 1;
      }
    }
  return 0;
  }


// Logically inside pane
FXbool FXComboBox::contains(FXint parentx,FXint parenty) const {
  if(pane->shown() && pane->contains(parentx,parenty)) return 1;
  return 0;
  }


// Forward left button press from list to target
long FXComboBox::onListLeftBtnPress(FXObject* sender,FXSelector,void* ptr){
  FXTRACE((200,"onListLeftBtnPress\n"));
  return target && target->handle(sender,MKUINT(message,SEL_LEFTBUTTONPRESS),ptr);
  }


// Forward left button release from list to target
long FXComboBox::onListLeftBtnRelease(FXObject* sender,FXSelector,void* ptr){
  FXTRACE((200,"onListLeftBtnRelease\n"));
  handle(this,MKUINT(ID_UNPOST,SEL_COMMAND),ptr);
  return target && target->handle(sender,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr);
  }


// Forward clicked message from list to target
long FXComboBox::onListClicked(FXObject* sender,FXSelector,void* ptr){
  FXTRACE((200,"onListClicked\n"));
  return target && target->handle(sender,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr);
  }


// Forward command message from list to target
long FXComboBox::onListCommand(FXObject* sender,FXSelector,void* ptr){
  FXTRACE((200,"onListCommand\n"));
  text->setText(list->getItemText((FXListItem*)ptr));
  return target && target->handle(sender,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr);
  }


// Forward selected message from list to target
long FXComboBox::onListSelected(FXObject* sender,FXSelector,void* ptr){
  FXTRACE((200,"onListSelected\n"));
  return target && target->handle(sender,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr);
  }


// Forward deselected message from list to target
long FXComboBox::onListDeselected(FXObject* sender,FXSelector,void* ptr){
  FXTRACE((200,"onListSelected\n"));
  return target && target->handle(sender,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr);
  }



// Text has changed
long FXComboBox::onCmdText(FXObject* sender,FXSelector,void* ptr){
  return 1;
  }


// Set text cursor position
void FXComboBox::setCursorPos(FXint pos){
  text->setCursorPos(pos);
  }


// Get text cursor position
FXint FXComboBox::getCursorPos() const { 
  return text->getCursorPos(); 
  }


// Set anchor pos
void FXComboBox::setAnchorPos(FXint pos){
  text->setAnchorPos(pos);
  }


// Get anchor pos
FXint FXComboBox::getAnchorPos() const { 
  return text->getAnchorPos(); 
  }


// Set text
void FXComboBox::setText(const FXString& txt){
  text->setText(txt);
  }


// Obtain text
FXString FXComboBox::getText() const {
  return text->getText(); 
  }


// Set number of text columns
void FXComboBox::setNumColumns(FXuint cols){
  text->setNumColumns(cols);
  }


// Get number of text columns
FXuint FXComboBox::getNumColumns() const {
  return text->getNumColumns(); 
  }


// Kill selection
void FXComboBox::killSelection(){
  text->killSelection();
  }


// Get number of items
FXint FXComboBox::getNumItems() const {
  return list->getNumItems(); 
  }


// Get number of selected items
FXint FXComboBox::getNumSelectedItems() const { 
  return list->getNumSelectedItems(); 
  }


// Get NULL terminated list of selected items
FXListItem** FXComboBox::getSelectedItems() const { 
  return list->getSelectedItems(); 
  }


// Get first item
FXListItem* FXComboBox::getFirstItem() const { 
  return list->getFirstItem(); 
  }


// Get last item
FXListItem* FXComboBox::getLastItem() const { 
  return list->getLastItem(); 
  }


// Get next item
FXListItem* FXComboBox::getNextItem(const FXListItem* item) const { 
  if(!item){ fxerror("%s::getNextItem: item is NULL.\n",getClassName()); } 
  return item->next; 
  }


// Get previous item
FXListItem* FXComboBox::getPrevItem(const FXListItem* item) const { 
  if(!item){ fxerror("%s::getPrevItem: item is NULL.\n",getClassName()); } 
  return item->prev; 
  }


// Add item as first
FXListItem* FXComboBox::addItemFirst(const FXString& txt,void* ptr){ 
  return list->addItemFirst(txt,ptr); 
  }


// Add item as last
FXListItem* FXComboBox::addItemLast(const FXString& txt,void* ptr){ 
  return list->addItemLast(txt,ptr); 
  }


// Add item after another
FXListItem* FXComboBox::addItemAfter(FXListItem* other,const FXString& txt,void* ptr){ 
  return list->addItemAfter(other,txt,ptr); 
  }


// Add item before another
FXListItem* FXComboBox::addItemBefore(FXListItem* other,const FXString& txt,void* ptr){ 
  return list->addItemBefore(other,txt,ptr); 
  }


// Remove given item
void FXComboBox::removeItem(FXListItem* item){ 
  list->removeItem(item); 
  }


// Remove sequence of items
void FXComboBox::removeItems(FXListItem* fm,FXListItem* to){
  list->removeItems(fm,to);
  }


// Remove all items
void FXComboBox::removeAllItems(){
  list->removeAllItems();
  }


// Get item at x, y  
FXListItem* FXComboBox::getItemAt(FXint x,FXint y) const {
  return list->getItemAt(x,y);
  }


// Get item x position
FXint FXComboBox::getItemX(const FXListItem* item) const {
  if(!item){ fxerror("%s::getItemX: item is NULL.\n",getClassName()); } 
  return 0; 
  }


// Get item y position
FXint FXComboBox::getItemY(const FXListItem* item) const {
  if(!item){ fxerror("%s::getItemY: item is NULL.\n",getClassName()); } 
  return item->y; 
  }


// Get item width
FXint FXComboBox::getItemWidth(const FXListItem* item) const {
  if(!item){ fxerror("%s::getItemWidth: item is NULL.\n",getClassName()); }
  return list->getItemWidth(item);
  }


// Get item height
FXint FXComboBox::getItemHeight(const FXListItem* item) const {
  if(!item){ fxerror("%s::getItemHeight: item is NULL.\n",getClassName()); } 
  return list->getItemHeight(item);
  }


// Update item
void FXComboBox::updateItem(FXListItem* item){
  list->updateItem(item);
  }



// Make item visible
void FXComboBox::makeItemVisible(FXListItem* item){
  list->makeItemVisible(item);
  }


// Is item selected
FXbool FXComboBox::isItemSelected(const FXListItem* item) const {
  if(!item){ fxerror("%s::isItemSelected: item is NULL.\n",getClassName()); } 
  return (item->state&LISTITEM_SELECTED)!=0; 
  }


// Is item current
FXbool FXComboBox::isItemCurrent(const FXListItem* item) const {
  if(!item){ fxerror("%s::isItemCurrent: item is NULL.\n",getClassName()); } 
  return (item->state&LISTITEM_CURRENT)!=0; 
  }


// Is item visible
FXbool FXComboBox::isItemVisible(const FXListItem* item) const {
  if(!item){ fxerror("%s::isItemVisible: item is NULL.\n",getClassName()); } 
  return list->isItemVisible(item);
  }


// Mark items
void FXComboBox::markItems(){
  list->markItems();
  }


// Select range of items
FXbool FXComboBox::selectItemRange(FXListItem* beg,FXListItem* end){
  return list->selectItemRange(beg,end);
  }


// Deselect range of items
FXbool FXComboBox::deselectItemRange(FXListItem* beg,FXListItem* end){
  return list->deselectItemRange(beg,end);
  }


// Toggle range of items
FXbool FXComboBox::toggleItemRange(FXListItem* beg,FXListItem* end){
  return list->toggleItemRange(beg,end);
  }


// Restore range of items
FXbool FXComboBox::restoreItemRange(FXListItem* beg,FXListItem* end){
  return list->restoreItemRange(beg,end);
  }


// Select item
FXbool FXComboBox::selectItem(FXListItem* item){
  return list->selectItem(item);
  }


// Deselect item
FXbool FXComboBox::deselectItem(FXListItem* item){
  return list->deselectItem(item);
  }


// Toggle item
FXbool FXComboBox::toggleItem(FXListItem* item){
  return list->toggleItem(item);
  }


// Restore item
FXbool FXComboBox::restoreItem(FXListItem* item){
  return list->restoreItem(item);
  }


// Sort item list
void FXComboBox::sortItems(){
  list->sortItems();
  }


// Change current item
void FXComboBox::setCurrentItem(FXListItem* item){
  /////
  list->setCurrentItem(item);
  }


// Get current item
FXListItem* FXComboBox::getCurrentItem() const {
  return list->getCurrentItem();
  }


// Set anchor item
void FXComboBox::setAnchorItem(FXListItem* item){
  list->setAnchorItem(item);
  }


// Get anchor item
FXListItem* FXComboBox::getAnchorItem() const {
  return list->getAnchorItem();
  }


// Extend selection
FXbool FXComboBox::extendSelection(FXListItem* item){
  return list->extendSelection(item);
  }


// Set item text;
void FXComboBox::setItemText(FXListItem* item,const FXString& text){
  if(item==NULL){ fxerror("%s::setItemText: item is NULL\n",getClassName()); }
  item->label=text;
  recalc(); // That's right, the ComboBox may get larger if the list does!!
  }


// Get item text
FXString FXComboBox::getItemText(const FXListItem* item) const {
  if(item==NULL){ fxerror("%s::getItemText: item is NULL\n",getClassName()); }
  return item->label;
  }


// Set item data
void FXComboBox::setItemData(FXListItem* item,void* ptr) const {
  if(item==NULL){ fxerror("%s::setItemData: item is NULL\n",getClassName()); }
  item->data=ptr;
  }


// Get item data 
void* FXComboBox::getItemData(const FXListItem* item) const {
  if(item==NULL){ fxerror("%s::getItemData: item is NULL\n",getClassName()); }
  return item->data;
  }


// Get sort function
FXItemSortFunc FXComboBox::getSortFunc() const {
  return list->getSortFunc();
  }


// Change sort function
void FXComboBox::setSortFunc(FXItemSortFunc func){
  list->setSortFunc(func);
  }  


// Delete it
FXComboBox::~FXComboBox(){
  delete pane;
  pane=(FXPopup*)-1;
  text=(FXTextField*)-1;
  button=(FXArrowButton*)-1;
  list=(FXList*)-1;
  }

