/********************************************************************************
*                                                                               *
*                     D i r e c t o r y   L i s t   O b j e c t                 *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXDirList.cpp,v 1.64 1999/01/15 05:29:25 jeroen Exp $                    *
********************************************************************************/
#include "xincs.h"
#include "fxdefs.h"
#include "fxkeys.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXFont.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXGIFIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXLabel.h"
#include "FXButton.h"
#include "FXComposite.h"
#include "FXCanvas.h"
#include "FXShell.h"
#include "FXScrollbar.h"
#include "FXScrollWindow.h"
#include "FXTreeList.h"
#include "FXDirList.h"
#include "FXFileDict.h"


// Private headers
#include "FXTreeItem.h"
#include "FXDirItem.h"


/*
  Notes:
  - One can never create items in constructor:- createItem() may be overloaded!
  - Instead of FXTreeItems, callbacks should pass pointer to directory.
  - Clipboard of a filenames.
  - Should do drag and drop and such.
  - Should be able to set initial directory/filename path, i.e. open all from
    root through last pathname segment which really exists.
  - Clipboard, DND, etc. support.
*/


#define REFRESHINTERVAL     1000          // Interval between refreshes
#define REFRESHFREQUENCY    32            // File systems not supporting mod-time, refresh every nth time

/*******************************************************************************/

/* Generated by reswrap from file minifolderopen.gif */
const unsigned char minifolderopen[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf2,0x00,0x00,0xb2,0xc0,0xdc,
  0x00,0x00,0x00,0x7f,0x7f,0x7f,0xff,0xff,0xff,0xd9,0xd9,0xd9,0xff,0xff,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x03,
  0x42,0x08,0xba,0xdc,0x2c,0x10,0xba,0x37,0x6a,0x15,0x13,0x88,0x41,0x4a,0x27,0x43,
  0x14,0x29,0x9b,0x67,0x82,0x56,0x18,0x68,0xdc,0xe9,0x12,0x42,0x20,0xce,0x62,0x11,
  0x6f,0x69,0x1e,0xc3,0x72,0xfb,0xb9,0xb2,0x18,0xeb,0x47,0xbc,0xad,0x4a,0xc4,0x93,
  0x6c,0xc5,0x7a,0x99,0x62,0x4c,0x1a,0x2d,0xc0,0x04,0x50,0xaf,0x58,0x6c,0x66,0xcb,
  0x6d,0x24,0x00,0x00,0x3b
  };

/* Generated by reswrap from file minifolderclosed.gif */
const unsigned char minifolderclosed[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf2,0x00,0x00,0xb2,0xc0,0xdc,
  0x80,0x80,0x80,0xc0,0xc0,0xc0,0xff,0xff,0x00,0x00,0x00,0x00,0xff,0xff,0xff,0x00,
  0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x03,
  0x3b,0x08,0xba,0xdc,0x1b,0x10,0x3a,0x16,0xc4,0xb0,0x22,0x4c,0x50,0xaf,0xcf,0x91,
  0xc4,0x15,0x64,0x69,0x92,0x01,0x31,0x7e,0xac,0x95,0x8e,0x58,0x7b,0xbd,0x41,0x21,
  0xc7,0x74,0x11,0xef,0xb3,0x5a,0xdf,0x9e,0x1c,0x6f,0x97,0x03,0xba,0x7c,0xa1,0x64,
  0x48,0x05,0x20,0x38,0x9f,0x50,0xe8,0x66,0x4a,0x75,0x24,0x00,0x00,0x3b
  };

/* Generated by reswrap from file minidoc.gif */
const unsigned char minidoc[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf2,0x00,0x00,0xbf,0xbf,0xbf,
  0x80,0x80,0x80,0xff,0xff,0xff,0xc0,0xc0,0xc0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x03,
  0x36,0x08,0x10,0xdc,0xae,0x70,0x89,0x49,0xe7,0x08,0x51,0x56,0x3a,0x04,0x86,0xc1,
  0x46,0x11,0x24,0x01,0x8a,0xd5,0x60,0x2a,0x21,0x6a,0xad,0x9a,0xab,0x9e,0xae,0x30,
  0xb3,0xb5,0x0d,0xb7,0xf2,0x9e,0xdf,0x31,0x14,0x90,0x27,0xf4,0xd5,0x86,0x83,0xa4,
  0x72,0x09,0x2c,0x39,0x9f,0xa6,0x04,0x00,0x3b
  };

/* Generated by reswrap from file miniapp.gif */
const unsigned char miniapp[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf2,0x00,0x00,0xb2,0xc0,0xdc,
  0x80,0x80,0x80,0xc0,0xc0,0xc0,0x00,0x00,0x00,0xff,0xff,0xff,0x00,0x00,0x80,0x00,
  0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x03,
  0x38,0x08,0xba,0xdc,0x10,0x30,0xca,0x09,0x85,0xbd,0xf8,0x86,0x11,0x44,0xf9,0x60,
  0xf8,0x6d,0x9d,0x58,0x10,0x03,0x8a,0x92,0x02,0xe5,0x72,0x02,0x21,0xcf,0xb4,0xcc,
  0xd6,0x38,0x71,0xe7,0xf4,0xce,0xdb,0xb0,0xdf,0xcc,0xf7,0x23,0xf2,0x48,0xae,0xd7,
  0x60,0xc9,0x6c,0x3a,0x07,0x8e,0xe8,0x22,0x01,0x00,0x3b
  };
  
/*******************************************************************************/

// Map
FXDEFMAP(FXDirList) FXDirListMap[]={
  FXMAPFUNC(SEL_TIMEOUT,FXDirList::ID_REFRESH,FXDirList::onRefresh),
  FXMAPFUNC(SEL_EXPANDED,0,FXDirList::onItemExpanded),
  FXMAPFUNC(SEL_COLLAPSED,0,FXDirList::onItemCollapsed),
  FXMAPFUNC(SEL_SELECTION_REQUEST,0,FXDirList::onSelectionRequest),
  FXMAPFUNC(SEL_UPDATE,FXDirList::ID_SHOW_HIDDEN,FXDirList::onUpdShowHidden),
  FXMAPFUNC(SEL_UPDATE,FXDirList::ID_HIDE_HIDDEN,FXDirList::onUpdHideHidden),
  FXMAPFUNC(SEL_UPDATE,FXDirList::ID_TOGGLE_HIDDEN,FXDirList::onUpdToggleHidden),
  FXMAPFUNC(SEL_UPDATE,FXDirList::ID_SHOW_FILES,FXDirList::onUpdShowFiles),
  FXMAPFUNC(SEL_UPDATE,FXDirList::ID_HIDE_FILES,FXDirList::onUpdHideFiles),
  FXMAPFUNC(SEL_UPDATE,FXDirList::ID_TOGGLE_FILES,FXDirList::onUpdToggleFiles),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_SETVALUE,FXDirList::onCmdSetValue),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_SETSTRINGVALUE,FXDirList::onCmdSetStringValue),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_GETSTRINGVALUE,FXDirList::onCmdGetStringValue),
  FXMAPFUNC(SEL_COMMAND,FXDirList::ID_SHOW_HIDDEN,FXDirList::onCmdShowHidden),
  FXMAPFUNC(SEL_COMMAND,FXDirList::ID_HIDE_HIDDEN,FXDirList::onCmdHideHidden),
  FXMAPFUNC(SEL_COMMAND,FXDirList::ID_TOGGLE_HIDDEN,FXDirList::onCmdToggleHidden),
  FXMAPFUNC(SEL_COMMAND,FXDirList::ID_SHOW_FILES,FXDirList::onCmdShowFiles),
  FXMAPFUNC(SEL_COMMAND,FXDirList::ID_HIDE_FILES,FXDirList::onCmdHideFiles),
  FXMAPFUNC(SEL_COMMAND,FXDirList::ID_TOGGLE_FILES,FXDirList::onCmdToggleFiles),
  };


// Object implementation
FXIMPLEMENT(FXDirList,FXTreeList,FXDirListMap,ARRAYNUMBER(FXDirListMap))

  
// For serialization
FXDirList::FXDirList(){
  refresh=NULL;
  counter=0;
  }


// Directory List Widget
FXDirList::FXDirList(FXComposite *p,FXObject* tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h):
  FXTreeList(p,tgt,sel,opts,x,y,w,h){
  open_folder=new FXGIFIcon(getApp(),minifolderopen);
  closed_folder=new FXGIFIcon(getApp(),minifolderclosed);
  mini_doc=new FXGIFIcon(getApp(),minidoc);
  mini_app=new FXGIFIcon(getApp(),miniapp);
  refresh=NULL;
  sortfunc=cmpName;
  associations=NULL;
  counter=0;
associations=new FXFileDict(getApp());///// This should be NULL by default
  }


// Create X window
void FXDirList::create(){
  FXTreeList::create();
  refresh=getApp()->addTimeout(REFRESHINTERVAL,this,ID_REFRESH);
  open_folder->create();
  closed_folder->create();
  mini_doc->create();
  mini_app->create();
  scanRootDir(FALSE);
  }


// Destroy zaps the icons
void FXDirList::destroy(){
  open_folder->destroy();
  closed_folder->destroy();
  mini_doc->destroy();
  mini_app->destroy();
  FXTreeList::destroy();
  }


// Create item
FXTreeItem* FXDirList::createItem(){ 
  FXTRACE((150,"%s::createItem\n",getClassName()));
  return (FXTreeItem*) new FXDirItem; 
  }


// Destroy item
void FXDirList::deleteItem(FXTreeItem* item){ 
  FXTRACE((150,"%s::deleteItem\n",getClassName()));
  delete ((FXDirItem*)item); 
  }


// Compare file names
FXbool FXDirList::cmpName(const FXItem* a,const FXItem* b){
  return strcmp(((FXDirItem*)a)->label.text(),((FXDirItem*)b)->label.text())>=0;
  }


// Helper function 
FXchar *FXDirList::getpath(const FXTreeItem* item,FXchar* pathname) const {
  register const FXchar *ptr;
  if(item){
    if(!item->parent){
      *pathname++=PATHSEP;
      }
    else{
      pathname=getpath(item->parent,pathname);
      if(*(pathname-1)!=PATHSEP){*pathname++=PATHSEP;}
      ptr=item->label.text();
      if(ptr){ while(*ptr){*pathname++=*ptr++;} }
      }
    }
  *pathname='\0';
  return pathname;
  }


// Determine path of item
FXchar*  FXDirList::itempath(FXDirItem* item,FXchar* pathname){
  getpath(item,pathname);
  return pathname;
  }


// Somebody wants our selection
long FXDirList::onSelectionRequest(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  FXuchar *data; 
  FXDragType *type;
  
  // Try handle through base class first
  if(FXTreeList::onSelectionRequest(sender,sel,ptr)) return 1;
  
  // When asked, supply the list of types
  if(event->target==targetType){
    FXTRACE((100,"%s: Get types list\n",getClassName()));
    FXMALLOC(&type,FXDragType,2);
    type[0]=stringType;
    type[1]=textType;
    setDNDData(event->origin,targetType,(FXuchar*)type,2*sizeof(FXDragType));
    return 1;
    }
  
  // Otherwise (most likely) its the string from this text we return
  // Strictly speaking, we should return ALL selected items, not just the current one
  if(currentitem && isItemSelected(currentitem)){
    if(event->target==stringType){
      FXCALLOC(&data,FXuchar,MAXPATHLEN+1);
      getpath(currentitem,(FXchar*)data);
      setDNDData(event->origin,stringType,data,strlen((FXchar*)data));
      }
    }
  return 0;
  }


// Open up the path down to the given string
long FXDirList::onCmdSetValue(FXObject*,FXSelector,void* ptr){
  //////////////////
  return 1;
  }


// Open up the path down to the given string
long FXDirList::onCmdSetStringValue(FXObject*,FXSelector,void* ptr){
  if(ptr==NULL){ fxerror("%s::onCmdSetStringValue: NULL pointer.\n",getClassName()); }
/////////////////////////
  return 1;
  }


// Obtain value of the current item
long FXDirList::onCmdGetStringValue(FXObject*,FXSelector,void* ptr){
  if(ptr==NULL){ fxerror("%s::onCmdGetStringValue: NULL pointer.\n",getClassName()); }
  if(currentitem){ 
    *((FXString*)ptr)=getItemPathname(currentitem);
    }
  else{
    *((FXString*)ptr)=NULL;
    }
  return 1;
  }


// Toggle hidden files
long FXDirList::onCmdToggleHidden(FXObject*,FXSelector,void*){
  showHiddenFiles(!showHiddenFiles());
  return 1;
  }


// Update toggle hidden files widget
long FXDirList::onUpdToggleHidden(FXObject* sender,FXSelector,void*){
  if(showHiddenFiles())
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Show hidden files
long FXDirList::onCmdShowHidden(FXObject*,FXSelector,void*){
  showHiddenFiles(TRUE);
  return 1;
  }


// Update show hidden files widget
long FXDirList::onUpdShowHidden(FXObject* sender,FXSelector,void*){
  if(showHiddenFiles())
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Hide hidden files
long FXDirList::onCmdHideHidden(FXObject*,FXSelector,void*){
  showHiddenFiles(FALSE);
  return 1;
  }


// Update hide hidden files widget
long FXDirList::onUpdHideHidden(FXObject* sender,FXSelector,void*){
  if(!showHiddenFiles())
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Toggle files display
long FXDirList::onCmdToggleFiles(FXObject*,FXSelector,void*){
  showFiles(!showFiles());
  return 1;
  }


// Update toggle files widget
long FXDirList::onUpdToggleFiles(FXObject* sender,FXSelector,void*){
  if(showFiles())
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Show files
long FXDirList::onCmdShowFiles(FXObject*,FXSelector,void*){
  showFiles(TRUE);
  return 1;
  }


// Update show files widget
long FXDirList::onUpdShowFiles(FXObject* sender,FXSelector,void*){
  if(showFiles())
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Hide files
long FXDirList::onCmdHideFiles(FXObject*,FXSelector,void*){
  showFiles(FALSE);
  return 1;
  }


// Update hide files widget
long FXDirList::onUpdHideFiles(FXObject* sender,FXSelector,void*){
  if(!showFiles())
    sender->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Item opened
long FXDirList::onItemExpanded(FXObject* sender,FXSelector sel,void* ptr){
  FXDirItem *item=(FXDirItem*)ptr;
  FXchar path[MAXPATHLEN];
  FXbool changed;
  if(!(item->state&DIRITEM_DIRECTORY)) return 0;

  FXTRACE((100,"%s::onItemExpanded: %s\n",getClassName(),item->label.text()));
  
  // Expand tree item
  FXTreeList::onItemExpanded(sender,sel,ptr);
  
  // Get path to item
  getpath(item,path);
  
  // List stuff in item directory
  changed=listSubDir(item,path);
  
  // Now we know for sure whether we really have subitems or not
  if(!item->first) item->state&=~DIRITEM_HASITEMS; else item->state|=DIRITEM_HASITEMS;
  
  // Sort items
  if(changed){
    sortChildItems(item);
    recalc();
    }
  return 1;
  }


// Item closed
long FXDirList::onItemCollapsed(FXObject* sender,FXSelector sel,void* ptr){
  FXDirItem *item=(FXDirItem*)ptr;
  if(!(item->state&DIRITEM_DIRECTORY)) return 0;

  FXTRACE((100,"%s::onItemCollapsed: %s\n",getClassName(),item->label.text()));
  
  // Collapse tree item
  FXTreeList::onItemCollapsed(sender,sel,ptr);
  
  // As a memory saving feature, all knowledge below this item
  // is deleted; we'll just recreate it when its reexpanded!
  //removeItems(item->first,item->last);
  //recalc();
  return 1;
  }


// Refresh
long FXDirList::onRefresh(FXObject* sender,FXSelector sel,void*){
  FXbool changed;
  FXTRACE((400,"%s::onRefresh\n",getClassName()));
  changed=scanRootDir(FALSE);
  if(changed) recalc();
  refresh=getApp()->addTimeout(REFRESHINTERVAL,this,ID_REFRESH);
  counter=(counter+1)%REFRESHFREQUENCY;
  return 1;
  }




// Scan root directory for changes
FXbool FXDirList::scanRootDir(FXbool relist){
  FXbool changed=FALSE;
  FXchar pathname[MAXPATHLEN];
  FXDirItem *item;
#ifndef WIN32
  struct stat info;
#else
  HANDLE hFile;
  FILETIME ftLastWriteTime;
#endif
  
  // Root directory name
  pathname[0]=PATHSEP;
  pathname[1]='\0';
  
  // Create root item if we don't have one yet
  if(!firstitem){
    item=(FXDirItem*)createItem();
    item->parent=NULL;
    item->next=NULL;
    item->prev=NULL;
    item->inext=NULL;
    item->iprev=NULL;
    item->list=NULL;
#ifndef WIN32
    item->date=0;
#else
    item->date.dwLowDateTime=0;
    item->date.dwHighDateTime=0;
#endif
    item->state=DIRITEM_DIRECTORY|DIRITEM_HASITEMS;
    item->label=pathname;
    item->first=NULL;
    item->last=NULL;
    item->openIcon=open_folder;
    item->closedIcon=closed_folder;
    firstitem=lastitem=item;
    changed=TRUE;
    }
  
  item=(FXDirItem*)firstitem;

  // Item is directory; check regardless of children
  if((item->state&DIRITEM_DIRECTORY) && (item->state&DIRITEM_EXPANDED)){
    
#ifndef WIN32
    
    // Refresh subitems of root if necessary
    if(stat(pathname,&info)==0){
      if(relist || (item->date<info.st_mtime) || (item->date<info.st_ctime) || ((info.st_mtime==0) && (counter==0))){
        if(listSubDir(item,pathname)){
          sortChildItems(item);
          changed=TRUE;
          }
        }
      item->date=FXMAX(info.st_mtime,info.st_ctime);
      }

#else
    
    // Refresh subitems of root if necessary
    hFile=CreateFile(pathname,GENERIC_READ,0,NULL,OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS,NULL);
    if(hFile!=INVALID_HANDLE_VALUE){
      if(GetFileTime(hFile,NULL,NULL,&ftLastWriteTime)){
	if(relist || (CompareFileTime(&item->date,&ftLastWriteTime)!=0)){
	  if(listSubDir(item,pathname)){
	    sortChildItems(item);
	    changed=TRUE;
	    }
	  }
	item->date=ftLastWriteTime;
	}
      CloseHandle(hFile);
      }

#endif
    
    // Then check subdirectories
    changed|=scanSubDir(item,pathname,relist);
    }
  return changed;
  }


// Scan sub directories
FXbool FXDirList::scanSubDir(FXDirItem *par,FXchar *pathname,FXbool relist){
  FXbool changed=FALSE;
  FXchar *pathtail,*pathend;
  FXDirItem *item;
#ifndef WIN32
  struct stat info;
#else
  HANDLE hFile;
  FILETIME ftLastWriteTime;
#endif

  if(par->first){
    
    // Build path prefix
    pathend=pathtail=pathname+strlen(pathname);
    if(*(pathtail-1)!=PATHSEP) *pathtail++=PATHSEP;
    
    // Loop over contents
    for(item=(FXDirItem*)par->first; item; item=(FXDirItem*)item->next){
      
      // Item is directory; check regardless of children
      if((item->state&DIRITEM_DIRECTORY) && (item->state&DIRITEM_EXPANDED)){
        
        // Sub item path
        strcpy(pathtail,item->label.text());
        
#ifndef WIN32
        
        // Refresh subitems of item if necessary
        if(stat(pathname,&info)==0){
          if(relist || (item->date<info.st_mtime) || (item->date<info.st_ctime) || ((info.st_mtime==0) && (counter==0))){
            if(listSubDir(item,pathname)){
              sortChildItems(item);
              changed=TRUE;
              }
            }
          item->date=FXMAX(info.st_mtime,info.st_ctime);
          }
        
#else
        
        // Refresh subitems of item if necessary
	hFile=CreateFile(pathname,GENERIC_READ,0,NULL,OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS,NULL);
	if(hFile!=INVALID_HANDLE_VALUE){
	  if(GetFileTime(hFile,NULL,NULL,&ftLastWriteTime)){
	    if(relist || (CompareFileTime(&item->date,&ftLastWriteTime)!=0)){
	      if(listSubDir(item,pathname)){
		sortChildItems(item);
		changed=TRUE;
		}
	      }
	    item->date=ftLastWriteTime;
	    }
	  CloseHandle(hFile);
	  }
        
#endif
        
        // Then check subdirectories
        changed|=scanSubDir(item,pathname,relist);
        }
      }
    *pathend='\0';
    }
  return changed;
  }


// List subdirectories
FXbool FXDirList::listSubDir(FXDirItem *par,FXchar *pathname){
  FXDirItem *after,*before,*newlist,*item,*it;
  FXIcon *openicon,*closedicon;
  FXchar *pathtail,*pathend,*name,*ext;
  FXbool changed=FALSE;
  FXFileAssoc *fileassoc;
#ifndef WIN32
  struct dirent *dp;
  struct stat info;
  DIR *dirp;
  int islink;
#else
  WIN32_FIND_DATA ffData;
  HANDLE hFindFile;
  DWORD dwBinaryType;
#endif
  
  // Find end of original pathname
  pathend=pathtail=pathname+strlen(pathname);
  
#ifndef WIN32

  // Get directory stream pointer
  dirp=opendir(pathname);
  if(!dirp) return FALSE;

  // Insert a '/' if needed
  if(*(pathtail-1)!=PATHSEP) *pathtail++=PATHSEP;
  
#else

  // Insert a '\' if needed
  if(*(pathtail-1)!=PATHSEP) *pathtail++=PATHSEP;
  
  // Get file find handle and first file's info
  strcpy(pathtail,"*");
  hFindFile=FindFirstFile(pathname,&ffData);
  if(hFindFile==INVALID_HANDLE_VALUE){ *pathend='\0'; return FALSE; }

#endif
  
  // Build new insert-order list
  after=NULL;
  before=par->list;
  newlist=NULL;
  if(!par->first) before=NULL;

#ifndef WIN32

  // Process directory entries
  while((dp=readdir(dirp))!=NULL){
    name=dp->d_name;

    // A dot special file?
    if(name[0]=='.' && (name[1]==0 || (name[1]=='.' && name[2]==0))) continue;

    // Hidden file or directory normally not shown
    if(name[0]=='.' && !(options&DIRLIST_SHOWHIDDEN)) continue;
    
    // Build full pathname
    strcpy(pathtail,name);

    // Get File info
    if(lstat(pathname,&info)!=0) continue;

    // If its a link, get the file info
    islink=S_ISLNK(info.st_mode);
    if(islink && stat(pathname,&info)!=0) continue;

    // If not showing files and it is not a directory, skip it
    if(!(options&DIRLIST_SHOWFILES) && !S_ISDIR(info.st_mode)) continue;

#else

  // Loop over directory entries
  do{
    name=ffData.cFileName;

    // A dot special file?
    if(name[0]=='.' && (name[1]==0 || (name[1]=='.' && name[2]==0))) continue;

    // Hidden file or directory normally not shown
    if((ffData.dwFileAttributes&FILE_ATTRIBUTE_HIDDEN || name[0]=='.') && !(options&DIRLIST_SHOWHIDDEN)) continue;

    // If not showing files and it is not a directory, skip it
    if(!(options&DIRLIST_SHOWFILES) && !(ffData.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY)) continue;
    
    // Build full pathname
    strcpy(pathtail,name);

#endif
    
    // Find it
    for(item=before; item; item=item->inext){
      if(strcmp(item->label.text(),name)==0){
        while(before!=item){
          it=before;
          before=before->inext;
          if(it->prev) it->prev->next=it->next; else it->parent->first=it->next;
          if(it->next) it->next->prev=it->prev; else it->parent->last=it->prev;
          if(currentitem==it) currentitem=NULL;
          if(anchoritem==it) anchoritem=NULL;
          removeItems(it->first,it->last);
          changed=TRUE;
          deleteItem(it);
          }
        before=item->inext;
        if(before) before->iprev=NULL;
        item->inext=NULL;
        item->iprev=NULL;
        goto fnd;
        }
      }

    // Not found; prepend before list
    item=(FXDirItem*)createItem();
    item->prev=par->last;
    item->next=NULL;
    item->parent=par;
    item->first=NULL;
    item->last=NULL;
    item->label=name;
    item->iprev=NULL;
    item->inext=NULL;
    item->list=NULL;
    item->state=DIRITEM_HASITEMS;
#ifndef WIN32
    item->date=0;
#else
    item->date.dwLowDateTime=0;
    item->date.dwHighDateTime=0;
#endif
    if(item->prev) item->prev->next=item; else par->first=item;
    par->last=item;
    changed=TRUE;

    // Next gets hung after this one
fnd:item->iprev=after;
    if(after) after->inext=item; else newlist=item;

    // Fill item with updated information, if data has changed
    
#ifndef WIN32
    
    // Anything about the file changed?
    if((item->date!=info.st_mtime) || (info.st_mtime==0)){
      
      // Item flags
      if(info.st_mode&(S_IXUSR|S_IXGRP|S_IXOTH)) item->state|=DIRITEM_EXECUTABLE; else item->state&=~DIRITEM_EXECUTABLE;
      if(S_ISDIR(info.st_mode)) item->state|=DIRITEM_DIRECTORY; else item->state&=~(DIRITEM_DIRECTORY|DIRITEM_HASITEMS);
      if(S_ISLNK(info.st_mode)) item->state|=DIRITEM_SYMLINK; else item->state&=~DIRITEM_SYMLINK;
      if(S_ISCHR(info.st_mode)) item->state|=DIRITEM_CHARDEV; else item->state&=~DIRITEM_CHARDEV;
      if(S_ISBLK(info.st_mode)) item->state|=DIRITEM_BLOCKDEV; else item->state&=~DIRITEM_BLOCKDEV;
      if(S_ISFIFO(info.st_mode)) item->state|=DIRITEM_FIFO; else item->state&=~DIRITEM_FIFO;
      if(S_ISSOCK(info.st_mode)) item->state|=DIRITEM_SOCK; else item->state&=~DIRITEM_SOCK;
    
      // Is it a directory?
      if(item->state&DIRITEM_DIRECTORY){
        openicon=open_folder;
        closedicon=closed_folder;
        }
      
      // Document
      else{
        
        // Is it an executable?
        if(item->state&DIRITEM_EXECUTABLE){
          openicon=mini_app;
          closedicon=mini_app;
          }
        else{
          openicon=mini_doc;
          closedicon=mini_doc;
          }
        
        // Try use association table
        if(associations){

          fileassoc=NULL;

          // Have file extension
          ext=strchr(name,'.');
          if(ext){
            fileassoc=associations->insert(ext+1);
            if(!fileassoc){
              ext=strchr(ext+1,'.');
              if(ext){
                fileassoc=associations->insert(ext+1);
                if(!fileassoc){
                  ext=strchr(ext+1,'.');
                  if(ext){
                    fileassoc=associations->insert(ext+1);
                    }
                  }
                }
              }
            }
          
          // Use whole filename; this is damn convenient!!
          if(!fileassoc){
            fileassoc=associations->insert(name);
            }
          
          // If we have an association, grab the icon
          if(fileassoc){
            if(fileassoc->miniicon){openicon=closedicon=fileassoc->miniicon;openicon->create();}
            }
          }
        }
      
      // Set the icons
      item->openIcon=openicon;
      item->closedIcon=closedicon;

      // Mod-time
      item->date=info.st_mtime;
      
      // Info has changed
      changed=TRUE;
      }

#else

    // Anything about the file changed?
    if(CompareFileTime(&item->date,&ffData.ftLastWriteTime)!=0){

      // Item flags
      if(ffData.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY) item->state|=DIRITEM_DIRECTORY; else item->state&=~(DIRITEM_DIRECTORY|DIRITEM_HASITEMS);
      if(!GetBinaryType(pathname,&dwBinaryType)) item->state&=~DIRITEM_EXECUTABLE; else item->state|=DIRITEM_EXECUTABLE;
      
      // Is it a directory?
      if(item->state&DIRITEM_DIRECTORY){
        openicon=open_folder;
        closedicon=closed_folder;
        }
      
      // Document
      else{
        
        // Is it an executable?
        if(item->state&DIRITEM_EXECUTABLE){
          openicon=mini_app;
          closedicon=mini_app;
          }
        else{
          openicon=mini_doc;
          closedicon=mini_doc;
          }
        
        // Try use association table
        if(associations){

          fileassoc=NULL;

          // Have file extension; NOTE: we don't expect multiple extensions under MS-Windows
          ext=strchr(name,'.');
          if(ext){
            fileassoc=associations->insert(ext+1);
            }
          
          // Use whole filename; this is damn convenient!!
          if(!fileassoc){
            fileassoc=associations->insert(name);
            }
          
          // If we have an association, grab the icon
          if(fileassoc){
            if(fileassoc->miniicon){openicon=closedicon=fileassoc->miniicon;openicon->create();}
            }
          }
        }
      
      // Set the icons
      item->openIcon=openicon;
      item->closedIcon=closedicon;

      // Mod-time
      item->date=ffData.ftLastWriteTime;
  
      // Info has changed
      changed=TRUE;
      }

#endif

    // Next one goes after item
    after=item;
    }
#ifdef WIN32
  while(FindNextFile(hFindFile,&ffData));
#endif

  // Wipe items remaining in list:- they have disappeared!!
  while(before){
    it=before;
    before=before->inext;
    if(it->prev) it->prev->next=it->next; else it->parent->first=it->next;
    if(it->next) it->next->prev=it->prev; else it->parent->last=it->prev;
    if(currentitem==it) currentitem=NULL;
    if(anchoritem==it) anchoritem=NULL;
    removeItems(it->first,it->last);
    changed=TRUE;
    deleteItem(it);
    }

  // Remember new list
  par->list=newlist;

  // Close it
#ifndef WIN32
  closedir(dirp);
#else
  FindClose(hFindFile);
#endif
  
  // Restore original path
  *pathend='\0';
  return changed;
  }


// Is directory
FXbool FXDirList::isItemDirectory(const FXTreeItem* item) const {
  if(item==NULL){ fxerror("%s::isItemDirectory: item is NULL.\n",getClassName()); }
  return (item->state&DIRITEM_DIRECTORY)!=0;
  }


// Is file
FXbool FXDirList::isItemFile(const FXTreeItem* item) const {
  if(item==NULL){ fxerror("%s::isItemFile: item is NULL.\n",getClassName()); }
  return (item->state&(DIRITEM_DIRECTORY|DIRITEM_CHARDEV|DIRITEM_BLOCKDEV|DIRITEM_FIFO|DIRITEM_SOCK))==0;
  }


// Is executable
FXbool FXDirList::isItemExecutable(const FXTreeItem* item) const {
  if(item==NULL){ fxerror("%s::isItemExecutable: item is NULL.\n",getClassName()); }
  return (item->state&DIRITEM_EXECUTABLE)!=0;
  }


// Obtain full pathname down from root
FXString FXDirList::getItemPathname(const FXTreeItem* item) const {
  FXchar path[MAXPATHLEN];
  if(item==NULL){ fxerror("%s::getItemPathname: item is NULL.\n",getClassName()); }
  getpath(item,path);
  return FXString(path);
  }


// Helper 
FXTreeItem *FXDirList::getNamedChild(FXTreeItem* par,const FXchar* name,FXuint len) const {
  FXTreeItem *item;
  if(par){
    item=par->first;
    while(item){
      if(strncmp(item->label.text(),name,len)==0) return item;
      item=item->next;
      }
    }
  return NULL;
  }


// Obtain item from full pathname
FXTreeItem* FXDirList::getPathnameItem(const FXString& file) const {
  if(file.text()==NULL){ fxerror("%s::getPathnameItem: empty file name specified.\n",getClassName()); }
  
  
////////////////////////
  return NULL;
  }


// Get list style
FXbool FXDirList::showFiles() const { 
  return (options&DIRLIST_SHOWFILES)!=0; 
  }


// Change list style
void FXDirList::showFiles(FXbool shown){
  FXuint opts=options;
  if(shown) opts|=DIRLIST_SHOWFILES; else opts&=~DIRLIST_SHOWFILES;
  if(options!=opts){
    options=opts;
    scanRootDir(TRUE);
    }
  }


// Return TRUE if showing hidden files
FXbool FXDirList::showHiddenFiles() const {
  return (options&DIRLIST_SHOWHIDDEN)!=0;
  }


// Change show hidden files mode
void FXDirList::showHiddenFiles(FXbool shown){
  FXuint opts=options;
  if(shown) opts|=DIRLIST_SHOWHIDDEN; else opts&=~DIRLIST_SHOWHIDDEN;
  if(opts!=options){
    options=opts;
    scanRootDir(TRUE);
    }
  }


// Save data
void FXDirList::save(FXStream& store) const {
  FXTreeList::save(store);
  store << associations;
  store << closed_folder;
  store << open_folder;
  store << mini_doc;
  store << mini_app;
  }


// Load data
void FXDirList::load(FXStream& store){ 
  FXTreeList::load(store);
  store >> associations; 
  store >> closed_folder;
  store >> open_folder;  
  store >> mini_doc;     
  store >> mini_app;     
  }


// Cleanup
FXDirList::~FXDirList(){
  removeAllItems();
  if(refresh) getApp()->removeTimeout(refresh);
delete associations;//////////////////////////// Really should not delete this...
  delete closed_folder;
  delete open_folder;
  delete mini_doc;
  delete mini_app;
  associations=(FXFileDict*)-1;
  closed_folder=(FXGIFIcon*)-1;
  open_folder=(FXGIFIcon*)-1;
  mini_doc=(FXGIFIcon*)-1;
  mini_app=(FXGIFIcon*)-1;
  refresh=(FXTimer*)-1;
  }


