/********************************************************************************
*                                                                               *
*                               F o n t   O b j e c t                           *
*                                                                               *
*********************************************************************************
* Copyright (C) 1997 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXFont.cpp,v 1.16 1999/01/12 04:17:06 jeroen Exp $                       *
********************************************************************************/
#include "xincs.h"
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXFont.h"


/*
  To do:
  - Instead of a font name only, should be taking different params.
  - We have hidden TEXTMETRICS or XFontStruct knowlegde from the public interface
*/


/*******************************************************************************/

// Object implementation
FXIMPLEMENT(FXFont,FXId,NULL,0)


// Deserialization
FXFont::FXFont(){
  font=(void*)-1;
  }


// Initialize
FXFont::FXFont(FXApp* a,const FXString& nm):FXId(a),name(nm){
  FXTRACE((100,"FXFont::FXFont %08x\n",this));
  font=NULL;
  }


// Create font
void FXFont::create(){
  if(!xid){
    FXTRACE((100,"%s::create %08x\n",getClassName(),this));
#ifndef FX_NATIVE_WIN32

    // Gotta have display open!
    if(!getDisplay()){ fxerror("%s::create: trying to create font before opening display.\n",getClassName()); }
  
    // Should have non-NULL font name
    if(!name.text()){ fxerror("%s::create: font name should not be NULL.\n",getClassName()); }

    // Try load the font
    font=XLoadQueryFont(getDisplay(),name.text());

    // Test if it worked    
    if(font){
      xid=((XFontStruct*)font)->fid;
      }
#else
    HDC hdc=GetDC(GetDesktopWindow());
    SaveDC(hdc);

    /*
    SetGraphicsMode(hdc,GM_ADVANCED);
    ModifyWorldTransform(hdc,NULL,MWT_IDENTITY);
    SetViewportOrgEx(hdc,0,0,NULL);
    SetWindowOrgEx(hdc,0,0,NULL);

    FLOAT cxDpi=(FLOAT)(25.4*GetDeviceCaps(hdc,HORZRES)/GetDeviceCaps(hdc,HORZSIZE));
    FLOAT cyDpi=(FLOAT)(25.4*GetDeviceCaps(hdc,VERTRES)/GetDeviceCaps(hdc,VERTSIZE));
    POINT pt;
    pt.x=0;
    pt.y=(int)(80*cyDpi/72);
    DPtoLP(hdc,&pt,1);

    LOGFONT lf;
    lf.lfHeight=-(int)(fabs(pt.y)/10.0+0.5);
    lf.lfWidth=0;
    lf.lfEscapement=0;
    lf.lfOrientation=0;
    lf.lfWeight=700;
    lf.lfItalic=0;
    lf.lfUnderline=0;
    lf.lfStrikeOut=0;
    lf.lfCharSet=0;
    lf.lfOutPrecision=0;
    lf.lfClipPrecision=0;
    lf.lfQuality=0;
    lf.lfPitchAndFamily=0;
    strcpy(lf.lfFaceName,"MS Sans Serif");
    xid=CreateFontIndirect(&lf);
    */
    xid=GetStockObject(DEFAULT_GUI_FONT);
    if(xid){
      FXCALLOC(&font,TEXTMETRIC,1);
      SelectObject(hdc,xid);
      GetTextMetrics(hdc,(TEXTMETRIC*)font);
      }
    RestoreDC(hdc,-1);
    ReleaseDC(GetDesktopWindow(),hdc);
#endif
    
    // Uh-oh, we failed
    if(!xid){ fxerror("%s::create: unable to create font.\n",getClassName()); }
    }
  }


// Destroy font
void FXFont::destroy(){
  if(xid){
    FXTRACE((100,"%s::destroy %08x\n",getClassName(),this));
    FXASSERT(font);
#ifndef FX_NATIVE_WIN32
    XFreeFont(getDisplay(),(XFontStruct*)font);
#else
    DeleteObject((HFONT)xid);
    FXFREE(&font);
#endif
    font=NULL;
    xid=0;
    }
  }


// Is it a mono space font
FXbool FXFont::isFontMono() const {
  if(!font){ fxerror("%s::isFontMono: font has not been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  return ((XFontStruct*)font)->per_char==NULL;
#else
  return !(((TEXTMETRIC*)font)->tmPitchAndFamily&TMPF_FIXED_PITCH);
#endif
  }


// Get font width
FXint FXFont::getFontWidth() const {
  if(!font){ fxerror("%s::getFontWidth: font has not been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  return ((XFontStruct*)font)->max_bounds.width;
#else
  return ((TEXTMETRIC*)font)->tmMaxCharWidth;
#endif
  }


// Get font height
FXint FXFont::getFontHeight() const {
  if(!font){ fxerror("%s::getFontHeight: font has not been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  return ((XFontStruct*)font)->max_bounds.ascent+((XFontStruct*)font)->max_bounds.descent;
#else
  return ((TEXTMETRIC*)font)->tmHeight;
#endif
  }


// Get font ascent
FXint FXFont::getFontAscent() const {
  if(!font){ fxerror("%s::getFontAscent: font has not been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  return ((XFontStruct*)font)->max_bounds.ascent;
#else
  return ((TEXTMETRIC*)font)->tmAscent;
#endif
  }


// Get font descent
FXint FXFont::getFontDescent() const {
  if(!font){ fxerror("%s::getFontDescent: font has not been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  return ((XFontStruct*)font)->max_bounds.descent;
#else
  return ((TEXTMETRIC*)font)->tmDescent;
#endif
  }


// Text width
FXint FXFont::getTextWidth(const FXchar *text,FXuint n) const {
  if(!text && n){ fxerror("%s::getTextWidth: NULL string argument\n",getClassName()); }
  if(!font){ fxerror("%s::getTextWidth: font has not been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  return XTextWidth((XFontStruct*)font,text,n);
#else
  HDC hdc=GetDC(GetDesktopWindow());
  SaveDC(hdc);
  SelectObject(hdc,xid);
  SIZE size;
  GetTextExtentPoint32(hdc,text,n,&size);
  RestoreDC(hdc,-1);
  ReleaseDC(GetDesktopWindow(),hdc);
  return size.cx;
#endif
  }


// Text height
FXint FXFont::getTextHeight(const FXchar *text,FXuint n) const {
  if(!font){ fxerror("%s::getTextHeight: font has not been created.\n",getClassName()); }
  if(!text && n){ fxerror("%s::getTextHeight: NULL string argument\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  XCharStruct chst; int dir,asc,desc;
  XTextExtents((XFontStruct*)font,text,n,&dir,&asc,&desc,&chst);
  return asc+desc;
#else
  HDC hdc=GetDC(GetDesktopWindow());
  SaveDC(hdc);
  SelectObject(hdc,xid);
  SIZE size;
  GetTextExtentPoint32(hdc,text,n,&size);
  RestoreDC(hdc,-1);
  ReleaseDC(GetDesktopWindow(),hdc);
  return size.cy;
#endif
  }


// Save font to stream
void FXFont::save(FXStream& store) const {
  FXId::save(store);
  store << name;
  }


// Load font from stream; create() should be called later
void FXFont::load(FXStream& store){
  FXId::load(store);
  store >> name;
#ifndef FX_NATIVE_WIN32
  if(font){XFreeFont(getDisplay(),(XFontStruct*)font);}
#else
  if(font){DeleteObject((HFONT)xid);FXFREE(&font);}
#endif
  font=NULL;
  xid=0;
  }  


// Clean up
FXFont::~FXFont(){
  FXTRACE((100,"FXFont::~FXFont %08x\n",this));
#ifndef FX_NATIVE_WIN32
  if(font){XFreeFont(getApp()->display,(XFontStruct*)font);}
#else
  if(font){FXFREE(&font);DeleteObject((HFONT)xid);}
#endif
  font=(void*)-1;
  }
  
