/********************************************************************************
*                                                                               *
*                        F r a m e   W i n d o w   O b j e c t                  *
*                                                                               *
*********************************************************************************
* Copyright (C) 1997 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXFrame.cpp,v 1.25 1999/01/07 03:38:42 jeroen Exp $                      *
********************************************************************************/
#include "xincs.h"
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXObjectList.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXCursor.h"
#include "FXDrawable.h"
#include "FXDC.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXComposite.h"
#include "FXShell.h"

/*
  Notes:
  - Update to use new FXWDC stuff.
*/

/*******************************************************************************/

// Map
FXDEFMAP(FXFrame) FXFrameMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXFrame::onPaint),
  };


// Object implementation
FXIMPLEMENT(FXFrame,FXWindow,FXFrameMap,ARRAYNUMBER(FXFrameMap))


// Deserialization
FXFrame::FXFrame(){
  baseColor=0;
  hiliteColor=0;
  shadowColor=0;
  borderColor=0;
  border=0;
  }
  

// Create child frame window
FXFrame::FXFrame(FXComposite* p,FXuint opts,FXint x,FXint y,FXint w,FXint h):
  FXWindow(p,opts,x,y,w,h){
  baseColor=getApp()->backColor;
  hiliteColor=getApp()->hiliteColor;
  shadowColor=getApp()->shadowColor;
  borderColor=getApp()->borderColor;
  border=(options&FRAME_THICK)?2:(options&(FRAME_SUNKEN|FRAME_RAISED))?1:0;
  }


// Save data
void FXFrame::save(FXStream& store) const {
  FXWindow::save(store);
  store << baseColor;
  store << hiliteColor;
  store << shadowColor;
  store << borderColor;
  store << border;
  }


// Load data
void FXFrame::load(FXStream& store){ 
  FXWindow::load(store);
  store >> baseColor;
  store >> hiliteColor;
  store >> shadowColor;
  store >> borderColor;
  store >> border;
  }


// Create window
void FXFrame::create(){
  FXWindow::create();
  show();
  }


void FXFrame::drawBorderRectangle(FXint x,FXint y,FXint w,FXint h){
  FXDC *dc=DC();
  dc->setForeground(borderColor);
  dc->drawRectangle(x,y,w,h);
  }


void FXFrame::drawRaisedRectangle(FXint x,FXint y,FXint w,FXint h){
  FXDC *dc=DC();
  dc->setForeground(shadowColor);
  dc->drawLine(x,y+h-1,x+w-1,y+h-1);
  dc->drawLine(x+w-1,y,x+w-1,y+h-1);
  dc->setForeground(hiliteColor);
  dc->drawLine(x,y,x+w-1,y);
  dc->drawLine(x,y,x,y+h-1);
  }


void FXFrame::drawSunkenRectangle(FXint x,FXint y,FXint w,FXint h){
  FXDC *dc=DC();
  dc->setForeground(shadowColor);
  dc->drawLine(x,y,x+w-2,y);
  dc->drawLine(x,y,x,y+h-2);
  dc->setForeground(hiliteColor);
  dc->drawLine(x,y+h-1,x+w-1,y+h-1);
  dc->drawLine(x+w-1,y,x+w-1,y+h-1);
  }


void FXFrame::drawRidgeRectangle(FXint x,FXint y,FXint w,FXint h){
  FXDC *dc=DC();
  dc->setForeground(hiliteColor);
  dc->drawLine(x,y,x+w-1,y);
  dc->drawLine(x,y,x,y+h-1);
  dc->drawLine(x+1,y+h-2,x+w-2,y+h-2);
  dc->drawLine(x+w-2,y+1,x+w-2,y+h-2);
  dc->setForeground(shadowColor);
  dc->drawLine(x+1,y+1,x+w-3,y+1);
  dc->drawLine(x+1,y+1,x+1,y+h-3);
  dc->drawLine(x,y+h-1,x+w-1,y+h-1);
  dc->drawLine(x+w-1,y,x+w-1,y+h-1);
  }


void FXFrame::drawGrooveRectangle(FXint x,FXint y,FXint w,FXint h){
  FXDC *dc=DC();
  dc->setForeground(shadowColor);
  dc->drawLine(x,y,x+w-1,y);
  dc->drawLine(x,y,x,y+h-1);
  dc->drawLine(x+1,y+h-2,x+w-2,y+h-2);
  dc->drawLine(x+w-2,y+1,x+w-2,y+h-2);
  dc->setForeground(hiliteColor);
  dc->drawLine(x+1,y+1,x+w-2,y+1);
  dc->drawLine(x+1,y+1,x+1,y+h-2);
  dc->drawLine(x+1,y+h-1,x+w-1,y+h-1);
  dc->drawLine(x+w-1,y+1,x+w-1,y+h-1);
  }


void FXFrame::drawDoubleRaisedRectangle(FXint x,FXint y,FXint w,FXint h){
  FXDC *dc=DC();
  dc->setForeground(hiliteColor);
  dc->drawLine(x,y,x+w-2,y);
  dc->drawLine(x,y,x,y+h-2);
  dc->setForeground(baseColor);
  dc->drawLine(x+1,y+1,x+w-3,y+1);
  dc->drawLine(x+1,y+1,x+1,y+h-3);
  dc->setForeground(shadowColor);
  dc->drawLine(x+1,y+h-2,x+w-2,y+h-2);
  dc->drawLine(x+w-2,y+h-2,x+w-2,y+1);
  dc->setForeground(borderColor);
  dc->drawLine(x,y+h-1,x+w-1,y+h-1);
  dc->drawLine(x+w-1,y,x+w-1,y+h-1);
  }

void FXFrame::drawDoubleSunkenRectangle(FXint x,FXint y,FXint w,FXint h){
  FXDC *dc=DC();
  dc->setForeground(shadowColor);
  dc->drawLine(x,y,x+w-1,y);
  dc->drawLine(x,y,x,y+h-1);
  dc->setForeground(borderColor);
  dc->drawLine(x+1,y+1,x+w-2,y+1);
  dc->drawLine(x+1,y+1,x+1,y+h-2);
  dc->setForeground(hiliteColor);
  dc->drawLine(x+1,y+h-1,x+w-1,y+h-1);
  dc->drawLine(x+w-1,y+h-1,x+w-1,y+1);
  dc->setForeground(baseColor);
  dc->drawLine(x+2,y+h-2,x+w-2,y+h-2);
  dc->drawLine(x+w-2,y+2,x+w-2,y+h-2);
  }


// Draw dashed focus rectangle
void FXFrame::drawFocusRectangle(FXint x,FXint y,FXint w,FXint h){
#ifndef FX_NATIVE_WIN32
  static const char onoff[]={1,2};
  FXDC *dc=DC();
  dc->setForeground(borderColor);
  dc->setDashes(0,onoff,2);
  dc->setLineStyle(LINE_ONOFF_DASH);
  dc->drawLine(x,y,x+w-1,y);
  dc->drawLine(x,y,x,y+h-1);
  dc->drawLine(x,y+h-1,x+w-1,y+h-1);
  dc->drawLine(x+w-1,y,x+w-1,y+h-1);
  dc->setLineStyle(LINE_SOLID);
#else
  
  FXDC *dc=DC();
  dc->setForeground(borderColor);
  HDC hdc=GetDC((HWND)xid);
  SaveDC(hdc);
  WORD b[8]={0xAAAA,0x5555,0xAAAA,0x5555,0xAAAA,0x5555,0xAAAA,0x5555};
  BITMAP bm;
  bm.bmType=0;
  bm.bmWidth=16;
  bm.bmHeight=8;
  bm.bmWidthBytes=2;
  bm.bmPlanes=1;
  bm.bmBitsPixel=1;
  bm.bmBits=b;
  HBITMAP hbm=CreateBitmapIndirect(&bm);
  HBRUSH hPatternBrush=CreatePatternBrush(hbm);
  DeleteObject(hbm);
  SelectObject(hdc,hPatternBrush);
  PatBlt(hdc,x,y,w,1,PATCOPY);
  PatBlt(hdc,x,y,1,h,PATCOPY);
  PatBlt(hdc,x,y+h-1,w,1,PATCOPY);
  PatBlt(hdc,x+w-1,y,1,h,PATCOPY);
  RestoreDC(hdc,-1);
  DeleteObject(hPatternBrush);

#endif
  }
  


// Draw border
void FXFrame::drawFrame(FXint x,FXint y,FXint w,FXint h){
  switch(options&FRAME_MASK) {
    case FRAME_LINE: drawBorderRectangle(x,y,w,h); break;
    case FRAME_SUNKEN: drawSunkenRectangle(x,y,w,h); break;
    case FRAME_RAISED: drawRaisedRectangle(x,y,w,h); break;
    case FRAME_GROOVE: drawGrooveRectangle(x,y,w,h); break;
    case FRAME_RIDGE: drawRidgeRectangle(x,y,w,h); break;
    case FRAME_SUNKEN|FRAME_THICK: drawDoubleSunkenRectangle(x,y,w,h); break;
    case FRAME_RAISED|FRAME_THICK: drawDoubleRaisedRectangle(x,y,w,h); break;
    }
  }


// Handle repaint 
long FXFrame::onPaint(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *ev=(FXEvent*)ptr;
  FXDC *dc=DC();
  dc->begin(this);
  dc->setForeground(backColor);
  dc->fillRectangle(ev->rect.x,ev->rect.y,ev->rect.w,ev->rect.h);
  drawFrame(0,0,width,height);
  dc->end();
  return 1;
  }


// Change frame border style
void FXFrame::setFrameStyle(FXuint style){
  FXuint opts=(options&~FRAME_MASK) | (style&FRAME_MASK);
  if(options!=opts){
    FXint b=(opts&FRAME_THICK) ? 2 : (opts&(FRAME_SUNKEN|FRAME_RAISED)) ? 1 : 0;
    options=opts;
    if(border!=b){
      border=b;
      recalc();
      }
    update(0,0,width,height);
    }
  }


// Get frame style
FXuint FXFrame::getFrameStyle() const { 
  return (options&FRAME_MASK); 
  }


// Set base color
void FXFrame::setBaseColor(FXColor clr){
  baseColor=clr;
  update(0,0,width,height);
  }


// Set highlight color
void FXFrame::setHiliteColor(FXColor clr){
  hiliteColor=clr;
  update(0,0,width,height);
  }


// Set shadow color
void FXFrame::setShadowColor(FXColor clr){
  shadowColor=clr;
  update(0,0,width,height);
  }


// Set border color
void FXFrame::setBorderColor(FXColor clr){
  borderColor=clr;
  update(0,0,width,height);
  }


// Get default width
FXint FXFrame::getDefaultWidth(){ 
  return (border<<1); 
  }


// Get default height
FXint FXFrame::getDefaultHeight(){ 
  return (border<<1); 
  }
