/********************************************************************************
*                                                                               *
*                    O p e n G L   C a n v a s   O b j e c t                    *
*                                                                               *
*********************************************************************************
* Copyright (C) 1997 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXGLCanvas.cpp,v 1.23 1999/01/14 02:33:26 jeroen Exp $                   *
********************************************************************************/
#include "xincs.h"
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDrawable.h"
#include "FXWindow.h"
#include "FXCursor.h"
#include "FXGLCanvas.h"

/*
  To do:
  - Better solution for onConfigure?
*/


#define ATTRIBLIST_SIZE 100

/*******************************************************************************/

#ifdef FX_NATIVE_WIN32
static FXuchar defSysClr[20][3] = {                         // System colors to match against
    { 0,   0,   0 },
    { 0x80,0,   0 },
    { 0,   0x80,0 },
    { 0x80,0x80,0 },
    { 0,   0,   0x80 },

    { 0x80,0,   0x80 },
    { 0,   0x80,0x80 },
    { 0xC0,0xC0,0xC0 },

    { 192, 220, 192 },
    { 166, 202, 240 },
    { 255, 251, 240 },
    { 160, 160, 164 },

    { 0x80,0x80,0x80 },
    { 0xFF,0,   0 },
    { 0,   0xFF,0 },
    { 0xFF,0xFF,0 },
    { 0,   0,   0xFF },
    { 0xFF,0,   0xFF },
    { 0,   0xFF,0xFF },
    { 0xFF,0xFF,0xFF }
    };
    
// Color mapping
static const FXuchar one_to_8[]={0,255};
static const FXuchar two_to_8[]={0,85,170,255};
static const FXuchar three_to_8[]={0,36,73,109,146,182,219,255};

#endif
    
/*******************************************************************************/

// Map
FXDEFMAP(FXGLCanvas) FXGLCanvasMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXGLCanvas::onPaint),
  FXMAPFUNC(SEL_CONFIGURE,0,FXGLCanvas::onConfigure),
  FXMAPFUNC(SEL_MOTION,0,FXGLCanvas::onMotion),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,0,FXGLCanvas::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,0,FXGLCanvas::onLeftBtnRelease),
  FXMAPFUNC(SEL_MIDDLEBUTTONPRESS,0,FXGLCanvas::onMiddleBtnPress),
  FXMAPFUNC(SEL_MIDDLEBUTTONRELEASE,0,FXGLCanvas::onMiddleBtnRelease),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS,0,FXGLCanvas::onRightBtnPress),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,0,FXGLCanvas::onRightBtnRelease),
  FXMAPFUNC(SEL_KEYPRESS,0,FXGLCanvas::onKeyPress),
  FXMAPFUNC(SEL_KEYRELEASE,0,FXGLCanvas::onKeyRelease),
  };


// Object implementation
FXIMPLEMENT(FXGLCanvas,FXWindow,FXGLCanvasMap,ARRAYNUMBER(FXGLCanvasMap))



// Make a canvas
FXGLCanvas::FXGLCanvas(FXComposite* p,FXObject* tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h):
  FXWindow(p,opts,x,y,w,h){
  flags|=FLAG_ENABLED;
  target=tgt;
  message=sel;
  redSize=1;
  greenSize=1;
  blueSize=1;
  alphaSize=1;
  depthSize=12;
  stencilSize=0;
  accumRedSize=0;
  accumGreenSize=0;
  accumBlueSize=0;
  accumAlphaSize=0;
#ifndef FX_NATIVE_WIN32
  visualInfo=NULL;
  colormap=0;
#endif
  ctx=0;
  }


// Canvas is an object drawn by another
long FXGLCanvas::onPaint(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_PAINT),ptr);
  }


// Handle configure notify
long FXGLCanvas::onConfigure(FXObject* sender,FXSelector sel,void* ptr){
  FXWindow::onConfigure(sender,sel,ptr);
  return target && target->handle(this,MKUINT(message,SEL_CONFIGURE),ptr);
  }



// Handle buttons if not handled in derived class
long FXGLCanvas::onLeftBtnPress(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONPRESS),ptr);
  }

long FXGLCanvas::onLeftBtnRelease(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr);
  }

long FXGLCanvas::onMiddleBtnPress(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_MIDDLEBUTTONPRESS),ptr);
  }

long FXGLCanvas::onMiddleBtnRelease(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_MIDDLEBUTTONRELEASE),ptr);
  }

long FXGLCanvas::onRightBtnPress(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_RIGHTBUTTONPRESS),ptr);
  }

long FXGLCanvas::onRightBtnRelease(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_RIGHTBUTTONRELEASE),ptr);
  }


// Mouse moved
long FXGLCanvas::onMotion(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_MOTION),ptr);
  }


// Handle keyboard press/release 
long FXGLCanvas::onKeyPress(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_KEYPRESS),ptr);
  }

long FXGLCanvas::onKeyRelease(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_KEYRELEASE),ptr);
  }


// Check for OpenGL extension
FXbool FXGLCanvas::checkOpenGL(){
#ifdef HAVE_OPENGL
#ifndef FX_NATIVE_WIN32
  return glXQueryExtension(getApp()->display,NULL,NULL);
#else
  return TRUE;
#endif
#else
  return FALSE;
#endif
  }


#ifndef FX_NATIVE_WIN32


// Dump visual class
static char* visualClass(int cls){
  if(cls==TrueColor) return "TrueColor";
  if(cls==DirectColor) return "DirectColor";
  if(cls==PseudoColor) return "PseudoColor";
  if(cls==StaticColor) return "StaticColor";
  if(cls==GrayScale) return "GrayScale";
  if(cls==StaticGray) return "StaticGray";
  return "Unknown";
  }

// Obtain visual 
Visual* FXGLCanvas::getDefaultVisual(){
  if(!visualInfo){
#ifdef HAVE_OPENGL
    if(checkOpenGL()){
      int attribList[ATTRIBLIST_SIZE];
      int *ptr=attribList;
      *ptr++=GLX_RGBA;
      if(options&GL_DOUBLE_BUFFER) *ptr++=GLX_DOUBLEBUFFER;
      if(options&GL_STEREO_BUFFER) *ptr++=GLX_STEREO;
      *ptr++=GLX_RED_SIZE;
      *ptr++=redSize;
      *ptr++=GLX_GREEN_SIZE;
      *ptr++=greenSize;
      *ptr++=GLX_BLUE_SIZE;
      *ptr++=blueSize;
      *ptr++=GLX_DEPTH_SIZE;
      *ptr++=depthSize;
//    *ptr++=GLX_BUFFER_SIZE;
//    *ptr++=bufferSize;
//    *ptr++=GLX_AUX_BUFFERS;
//    *ptr++=auxBuffers;
//    *ptr++=GLX_ALPHA_SIZE;
//    *ptr++=alphaSize;
//    *ptr++=GLX_STENCIL_SIZE;
//    *ptr++=stencilSize;
//    *ptr++=GLX_ACCUM_RED_SIZE;
//    *ptr++=accumRedSize;
//    *ptr++=GLX_ACCUM_GREEN_SIZE;
//    *ptr++=accumGreenSize;
//    *ptr++=GLX_ACCUM_BLUE_SIZE;
//    *ptr++=accumBlueSize;
//    *ptr++=GLX_ACCUM_ALPHA_SIZE;
//    *ptr++=accumAlphaSize;
      *ptr++=None;
      visualInfo=glXChooseVisual(getApp()->display,DefaultScreen(getApp()->display),attribList);
      if(visualInfo){
        FXTRACE((100,"visualID=0x%lx\n",visualInfo->visualid));
        FXTRACE((100,"  Screen: %d\n",visualInfo->screen));
        FXTRACE((100,"  Depth: %d\n",visualInfo->depth));
        FXTRACE((100,"  Bits: %d\n",visualInfo->bits_per_rgb));
        FXTRACE((100,"  Class: %s\n",visualClass(visualInfo->c_class)));
        return visualInfo->visual;
        }
      fxerror("%s::getDefaultVisual: requested OpenGL visual unavailable.\n",getClassName());
      }
#endif
    fxerror("%s::getDefaultVisual: display does not support OpenGL.\n",getClassName());
    }
  return visualInfo->visual;
  }

#endif


#if 0
#ifdef FX_NATIVE_WIN32

// Compute color component from map index
static FXuchar compX(int i,FXuchar nbits,FXuchar shift){
  unsigned char val = i>>shift;
  switch(nbits){
    case 1: return one_to_8[val&1];
    case 2: return two_to_8[val&3];
    case 3: return three_to_8[val&7];
    }
  return 0;
  }


// Make a logical palette
static LOGPALETTE *makeLogPalette(PIXELFORMATDESCRIPTOR& pfd){
  unsigned char index[256],closest;
  int mindif,dif,dr,dg,db,n,i,j;
  LOGPALETTE *pal;

  n = 1 << pfd.cColorBits;
  FXMALLOC(&pal,char,sizeof(LOGPALETTE)+sizeof(PALETTEENTRY)*n);
  pal->palVersion = 0x300;
  pal->palNumEntries = n;

  for(i=0; i<n; i++) {
    pal->palPalEntry[i].peRed = compX(i,pfd.cRedBits,pfd.cRedShift);
    pal->palPalEntry[i].peGreen = compX(i,pfd.cGreenBits,pfd.cGreenShift);
    pal->palPalEntry[i].peBlue = compX(i,pfd.cBlueBits,pfd.cBlueShift);
    pal->palPalEntry[i].peFlags = PC_NOCOLLAPSE;
    }

  // Find best mapping of default colors to new map
  if(n==256){
    for(i=0; i<256; i++) index[i]=255;
    for(j=19; j>=0; j--){
      mindif = 3*256*256;
      for(i=0; i<256; i++){
        if(index[i]==255){
          dr = (int)pal->palPalEntry[i].peRed - defSysClr[j][0];
          dg = (int)pal->palPalEntry[i].peGreen - defSysClr[j][1];
          db = (int)pal->palPalEntry[i].peBlue - defSysClr[j][2];
          dif = dr*dr+dg*dg+db*db;
          if(dif<mindif){
            closest = i;
            mindif = dif;
            }
          }
        }
//      TRACE("system color %d mapped to %d: dif = %f\n",j,closest,sqrt((double)mindif));
      ASSERT(index[closest]==255);
      index[closest] = j;
      }

    // Change proposed map to include already existing colors
    for(i=0; i<256; i++){
      j = index[i];
      if(j<20){
        pal->palPalEntry[i].peRed = defSysClr[j][0];
        pal->palPalEntry[i].peGreen = defSysClr[j][1];
        pal->palPalEntry[i].peBlue = defSysClr[j][2];
        pal->palPalEntry[i].peFlags = 0;
        }
      }
    }
  return pal;
  }  



/**************************  Choose Pixel Format  ****************************/


// Choose best pixel format matching given flags
static int choosePFD(HDC hdc,PIXELFORMATDESCRIPTOR& pxfmt){
  PIXELFORMATDESCRIPTOR pfd;
  int i,pfdnum,pfdno,bpp,hw,db,wantdb,depth,hardw,dblbf;

  // Get number of bits/pixel 
  bpp = GetDeviceCaps(hdc,BITSPIXEL);


  // Get number of supported pixel formats
  pfd.nSize=sizeof(PIXELFORMATDESCRIPTOR);   
  pfd.nVersion=1;
  pfdnum=DescribePixelFormat(hdc,1,sizeof(PIXELFORMATDESCRIPTOR),&pfd);
  if(pfdnum==0) return 0;

  // Find pixel format for RGBA mode
  pfdno = 0;
  depth = 0;
  hardw = 0;
  dblbf = 0;
  wantdb = (pxfmt.dwFlags&PFD_DOUBLEBUFFER) != 0;
  for(i=1; i<=pfdnum; i++) {
    DescribePixelFormat(hdc,i,sizeof(PIXELFORMATDESCRIPTOR),&pfd);

    // Without OpenGL, RGBA mode, and Z-buffer, it's pointless!
    if((pfd.dwFlags&PFD_SUPPORT_OPENGL) && (pfd.iPixelType==PFD_TYPE_RGBA) && (pfd.cDepthBits>=16)){

      // Same bits/pixel as device?
      if(pfd.cColorBits==bpp){

        // Can draw on desired device surface
        if((pxfmt.dwFlags&pfd.dwFlags&PFD_DRAW_TO_WINDOW) ||
           (pxfmt.dwFlags&pfd.dwFlags&PFD_DRAW_TO_BITMAP)){

          hw = (pfd.dwFlags&PFD_GENERIC_FORMAT) == 0;
          db = (pfd.dwFlags&PFD_DOUBLEBUFFER) != 0;

          // We prefer hardware and double buffer over depth
          if(hw>hardw || (hw==hardw && (wantdb==db && (dblbf!=wantdb || pfd.cDepthBits>depth)))){
            pfdno = i;
            depth = pfd.cDepthBits;
            hardw = hw;
            dblbf = db;
            }
          }
        }
      }
    }

  // Check if successful
  if(pfdno==0) return 0;

  // Fetch the pixel format which was chosen
  pxfmt.nSize=sizeof(PIXELFORMATDESCRIPTOR);   
  pxfmt.nVersion=1;
  DescribePixelFormat(hdc,pfdno,sizeof(PIXELFORMATDESCRIPTOR),&pxfmt);

  return pfdno;
  }

#endif
#endif


// Get depth from visual
FXint FXGLCanvas::getDefaultDepth(){
#ifndef FX_NATIVE_WIN32
  return visualInfo->depth;
#else
  HDC hdc=GetDC((HWND)xid);
  FXint nDepth=GetDeviceCaps(hdc,BITSPIXEL);
  ReleaseDC((HWND)xid,hdc);
  return nDepth;
#endif
  }

#ifndef FX_NATIVE_WIN32

// Obtain colormap
FXID FXGLCanvas::getDefaultColormap(){
  static struct CmapCache { Visual *visual; FXID cmap; } *cmapCache=NULL;
  static int cacheMalloced=0;
  static int cacheEntries=0;
  Visual* visual;
  
  if(!colormap){
    
    // Get visual for which to allocate colormap
    visual=getDefaultVisual();
    
    // Only make a new colormap if the visual is different from the default visual
    if((options&GL_PRIVATE_MAP) || (visual!=DefaultVisual(getDisplay(),DefaultScreen(getDisplay())))){
      
      // See if we can find it in the cache
      for(int i=0; i<cacheEntries; i++){
        if(cmapCache[i].visual==visual){
          colormap=cmapCache[i].cmap;
          return colormap;
          }
        }

      // not in the cache, create a new entry
      if(cacheEntries>=cacheMalloced){
        if(cacheMalloced==0){
          cacheMalloced=1;
          cmapCache=(CmapCache*)malloc(sizeof(CmapCache));
          }
        else{
          cacheMalloced<<=1;
          cmapCache=(CmapCache*)realloc(cmapCache,sizeof(CmapCache)*cacheMalloced);
          }
        }
 
      FXASSERT(cmapCache);
 
      // Add new entry
      colormap=XCreateColormap(getApp()->display,RootWindow(getApp()->display,0),visual,AllocNone);
      cmapCache[cacheEntries].cmap=colormap;
      cmapCache[cacheEntries].visual=visual;
      cacheEntries++;
      }
    
    // Get colormap for default visual
    else{
      FXASSERT(visual == DefaultVisual(getDisplay(),DefaultScreen(getDisplay())));
      colormap=DefaultColormap(getDisplay(),DefaultScreen(getDisplay())); 
      }
    }
  return colormap;
  }

#endif


// Create X window (GL CANVAS)
void FXGLCanvas::create(){
  FXWindow::create();
  if(!ctx){
#ifdef HAVE_OPENGL
#ifndef FX_NATIVE_WIN32
    if(options&GL_INSTALL_COLORMAP) addColormapWindows();
    ctx=glXCreateContext(getDisplay(),visualInfo,NULL,TRUE);
    if(!ctx){ fxerror("%s::create(): unable to create graphics rendering context.\n",getClassName()); }
    glXMakeCurrent(getDisplay(),xid,(GLXContext)ctx);
#else
    DWORD dwFlags=PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL;
    if(options&GL_DOUBLE_BUFFER) dwFlags|=PFD_DOUBLEBUFFER;
    if(options&GL_STEREO_BUFFER) dwFlags|=PFD_STEREO;

    PIXELFORMATDESCRIPTOR pfd={
      sizeof(PIXELFORMATDESCRIPTOR), 1,
      dwFlags,
      PFD_TYPE_RGBA,            // RGBA type 
      getDefaultDepth(),
      redSize, 0, greenSize, 0, blueSize, 0,
      0,
      0,
      0,
      0, 0, 0, 0,
      depthSize,
      0,
      0,
      PFD_MAIN_PLANE,
      0,
      0, 0, 0
      };

    // Get the best available match of pixel format for the device context
    HDC hdc=GetDC((HWND)xid);
    int iPixelFormat=ChoosePixelFormat(hdc,&pfd);

    // Make that the pixel format of the device context 
    SetPixelFormat(hdc,iPixelFormat,&pfd);

    // Now try to get a GL context
    ctx=wglCreateContext(hdc);
    if(!ctx){ fxerror("%s::create(): unable to create graphics rendering context.\n",getClassName()); }
    wglMakeCurrent(hdc,(HGLRC)ctx);
#endif
#endif
    }
  show();
  }


// Destroy the GL Canvas
void FXGLCanvas::destroy(){
#ifdef HAVE_OPENGL
#ifndef FX_NATIVE_WIN32
  if(ctx) glXDestroyContext(getDisplay(),(GLXContext)ctx);
  if(options&GL_INSTALL_COLORMAP) removeColormapWindows();
#else
  wglDeleteContext((HGLRC)ctx);
#endif
  ctx=0;
#endif
  FXWindow::destroy();
  }


//  Make the rendering context of GL Canvas current
FXbool FXGLCanvas::makeCurrent(){
#ifdef HAVE_OPENGL
#ifndef FX_NATIVE_WIN32
  glXMakeCurrent(getDisplay(),xid,(GLXContext)ctx);
#else
  return wglMakeCurrent(GetDC((HWND)xid),(HGLRC)ctx);
#endif
#endif
  return TRUE;
  }
  
  

// Used by GL to swap the buffers in double buffer mode, or flush a single buffer
void FXGLCanvas::swapBuffers(){
#ifdef HAVE_OPENGL
#ifndef FX_NATIVE_WIN32
  glXSwapBuffers(getApp()->display,xid);
#else
  SwapBuffers(GetDC((HWND)xid));
#endif
#endif
  }


// This function only available on Mesa
void FXGLCanvas::swapSubBuffers(FXint x,FXint y,FXint w,FXint h){
#ifdef HAVE_OPENGL
#ifdef MESA
#ifdef GLX_MESA_copy_sub_buffer
  glXCopySubBufferMESA(getApp()->display,xid,x,height-y-h-1,w,h);
#else
  glXSwapBuffers(getApp()->display,xid);
#endif
#endif
#endif
  }


// Close and release any resources
FXGLCanvas::~FXGLCanvas(){
#ifdef HAVE_OPENGL
#ifndef FX_NATIVE_WIN32
  glXDestroyContext(getDisplay(),(GLXContext)ctx);
  if(options&GL_INSTALL_COLORMAP) removeColormapWindows();
  if(visualInfo) XFree(visualInfo);
#else
  wglDeleteContext((HGLRC)ctx);
#endif
#endif
  }
