/********************************************************************************
*                                                                               *
*          M u l t i p l e  D o c u m e n t   C l i e n t   W i n d o w         *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXMDIClient.cpp,v 1.58 1999/01/07 03:38:43 jeroen Exp $                  *
********************************************************************************/
#include "xincs.h"
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXLabel.h"
#include "FXButton.h"
#include "FXMenuButton.h"
#include "FXComposite.h"
#include "FXShell.h"
#include "FXPopup.h"
#include "FXMenu.h"
#include "FXMDIButton.h"
#include "FXMenuBar.h"
#include "FXMDIChild.h"
#include "FXScrollbar.h"
#include "FXScrollWindow.h"
#include "FXMDIClient.h"


/*
  Notes:
  - This brings up the question of the cascade design. Do you want the windows 
    to be cascaded/tiled in the order that they were created, or do you want them 
    to be cascaded in some sort of focus order, with the one that had the focus 
    to be on top, keeping its focus.
    In the test app, if you click within any of the child windows, it pops to 
    the top with the focus, with the exception of "TEST3", which has a button. 
    If this button is clicked, that window dos not pop to the top when it 
    gets the focus. Seems like it should...
  - Minor problems with mdi.  When a window is maximized and then
    deleted, the next window that is created is created in "normal" mode.
    When any adjustment is made to the size of the new window, it pops to
    "almost" maximized mode.  (The child frame is visible, but as large as
    possible).  Seems like the child windows should be created maximized if
    the mdi is in maximized mode.  Also, if there are two windows and one
    gets maximized and then deleted, the second window pops to "almost"
    maximixed mode.
  - We make MDIClient get a first crack at the messages, so that the MDIChild
    can not shadow any messages really directed at the MDIClient.
  - Need ``arrange icons'' feature.
*/

#define CASCADE_XOFF  24
#define CASCADE_YOFF  24


/*******************************************************************************/


FXDEFMAP(FXMDIClient) FXMDIClientMap[]={
  FXMAPFUNC(SEL_CHANGED,0,FXMDIClient::onChanged),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_MDI_NEXT,FXMDIClient::onUpdActivateNext),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_MDI_PREV,FXMDIClient::onUpdActivatePrev),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_MDI_TILEHORIZONTAL,FXMDIClient::onUpdTileHorizontal),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_MDI_TILEVERTICAL,FXMDIClient::onUpdTileVertical),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_MDI_CASCADE,FXMDIClient::onUpdCascade),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_MDI_DELETE,FXMDIClient::onUpdDelete),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_MDI_MENUDELETE,FXMDIClient::onUpdMenuDelete),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_MDI_MINIMIZE,FXMDIClient::onUpdMinimize),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_MDI_MENUMINIMIZE,FXMDIClient::onUpdMenuMinimize),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_MDI_RESTORE,FXMDIClient::onUpdRestore),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_MDI_MENURESTORE,FXMDIClient::onUpdMenuRestore),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_MDI_MAXIMIZE,FXMDIClient::onUpdMaximize),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_MDI_MENUWINDOW,FXMDIClient::onUpdMenuWindow),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_MDI_NEXT,FXMDIClient::onCmdActivateNext),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_MDI_PREV,FXMDIClient::onCmdActivatePrev),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_MDI_TILEHORIZONTAL,FXMDIClient::onCmdTileHorizontal),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_MDI_TILEVERTICAL,FXMDIClient::onCmdTileVertical),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_MDI_CASCADE,FXMDIClient::onCmdCascade),
  };


// Object implementation
FXIMPLEMENT(FXMDIClient,FXScrollArea,FXMDIClientMap,ARRAYNUMBER(FXMDIClientMap))


// Construct and init
FXMDIClient::FXMDIClient(){
  mdifirst=NULL;
  mdilast=NULL;
  active=NULL;
  cascadex=CASCADE_XOFF;
  cascadey=CASCADE_YOFF;
  xmin=0;
  xmax=0;
  ymin=0;
  ymax=0;
  }


// Construct and init
FXMDIClient::FXMDIClient(FXComposite* p,FXuint opts,FXint x,FXint y,FXint w,FXint h):
  FXScrollArea(p,opts,x,y,w,h){
  backColor=getApp()->shadowColor;
  mdifirst=NULL;
  mdilast=NULL;
  active=NULL;
  cascadex=CASCADE_XOFF;
  cascadey=CASCADE_YOFF;
  xmin=0;
  xmax=0;
  ymin=0;
  ymax=0;
  }


// Set other background color
void FXMDIClient::create(){
  FXScrollArea::create();
  }


// Propagate size change
void FXMDIClient::recalc(){ 
  FXScrollArea::recalc();
  flags|=FLAG_RECALC;
  }


// Update restore; gray if no active
long FXMDIClient::onUpdRestore(FXObject* sender,FXSelector,void*){
  if(!active){ return sender->handle(this,MKUINT(ID_DISABLE,SEL_COMMAND),NULL); }
  return 0;
  }


// Update maximized; gray if no active
long FXMDIClient::onUpdMaximize(FXObject* sender,FXSelector,void* ptr){
  if(!active){ return sender->handle(this,MKUINT(ID_DISABLE,SEL_COMMAND),NULL); }
  return 0;
  }


// Update minimized
long FXMDIClient::onUpdMinimize(FXObject* sender,FXSelector,void* ptr){
  if(!active){ return sender->handle(this,MKUINT(ID_DISABLE,SEL_COMMAND),NULL); }
  return 0;
  }


// Update delete active child
long FXMDIClient::onUpdDelete(FXObject* sender,FXSelector,void* ptr){
  if(!active){ return sender->handle(this,MKUINT(ID_DISABLE,SEL_COMMAND),NULL); }
  return 0;
  }


// Update menu's restore button
long FXMDIClient::onUpdMenuWindow(FXObject* sender,FXSelector,void* ptr){
  if(!active){ return sender->handle(this,MKUINT(ID_HIDE,SEL_COMMAND),NULL); }
  return 0;
  }


// Update menu's restore button
long FXMDIClient::onUpdMenuRestore(FXObject* sender,FXSelector,void* ptr){
  if(!active){ return sender->handle(this,MKUINT(ID_HIDE,SEL_COMMAND),NULL); }
  return 0;
  }


// Update menu's minimized button
long FXMDIClient::onUpdMenuMinimize(FXObject* sender,FXSelector,void* ptr){
  if(!active){ return sender->handle(this,MKUINT(ID_HIDE,SEL_COMMAND),NULL); }
  return 0;
  }


// Update menu's delete button
long FXMDIClient::onUpdMenuDelete(FXObject* sender,FXSelector,void* ptr){
  if(!active){ return sender->handle(this,MKUINT(ID_HIDE,SEL_COMMAND),NULL); }
  return 0;
  }


// Set the active child
FXbool FXMDIClient::setActiveChild(FXMDIChild* child){
  FXbool wasmax=FALSE;
  if(active!=child){
    if(active){
      wasmax=active->isMaximized();
      active->handle(this,MKUINT(0,SEL_DEACTIVATE),NULL);
      if(wasmax) active->handle(this,MKUINT(0,SEL_RESTORE),NULL);
      active=NULL;
      }
    if(child){
      active=child;
      active->handle(this,MKUINT(0,SEL_ACTIVATE),NULL);
      if(wasmax) active->handle(this,MKUINT(0,SEL_MAXIMIZE),NULL);
      }
    recalc();
    return TRUE;
    }
  return FALSE;
  }


// Tile horizontally (actually, prefer wider windows)
long FXMDIClient::onCmdTileHorizontal(FXObject*,FXSelector,void* ptr){
  register FXMDIChild* child;
  register FXint n,nr,nc,hroom,vroom,r,c;
  for(n=0,child=mdifirst; child; child=child->getMDINext()){
    if(child->shown() && !child->isMinimized()) n++;
    }
  nr=n;
  nc=1;
  if(n>3){
    nc=(int)sqrt((double)n);
    nr=(n+nc-1)/nc;
    }
  hroom=0;
  vroom=0;
  if(nc>0) hroom=width/nc;
  if(nr>0) vroom=height/nr;
  for(child=mdifirst,n=0; child; child=child->getMDINext()){
    if(child->shown() && !child->isMinimized()){
      r=n/nc;
      c=n%nc;
      child->handle(this,MKUINT(0,SEL_RESTORE),NULL);
      child->position(c*hroom,r*vroom,hroom,vroom);
      n++;
      }
    }
  if(active && active->shown()) active->raise();
  recalc();
  return 1;
  }


// Update tile horizontally
long FXMDIClient::onUpdTileHorizontal(FXObject* sender,FXSelector,void* ptr){
  FXuint msg=mdifirst ? ID_ENABLE : ID_DISABLE;
  sender->handle(this,MKUINT(msg,SEL_COMMAND),ptr);
  return 1;
  }


// Tile vertically (actually, prefer taller windows)
long FXMDIClient::onCmdTileVertical(FXObject*,FXSelector,void* ptr){
  register FXMDIChild* child;
  register FXint n,nr,nc,hroom,vroom,r,c;
  for(n=0,child=mdifirst; child; child=child->getMDINext()){
    if(child->shown() && !child->isMinimized()) n++;
    }
  nc=n;
  nr=1;
  if(n>3){
    nr=(int)sqrt((double)n);
    nc=(n+nr-1)/nr;
    }
  hroom=0;
  vroom=0;
  if(nc>0) hroom=width/nc;
  if(nr>0) vroom=height/nr;
  for(child=mdifirst,n=0; child; child=child->getMDINext()){
    if(child->shown() && !child->isMinimized()){
      r=n/nc;
      c=n%nc;
      child->handle(this,MKUINT(0,SEL_RESTORE),NULL);
      child->position(c*hroom,r*vroom,hroom,vroom);
      n++;
      }
    }
  if(active && active->shown()) active->raise();
  recalc();
  return 1;
  }


// Update tile vertically
long FXMDIClient::onUpdTileVertical(FXObject* sender,FXSelector,void* ptr){
  FXuint msg=mdifirst ? ID_ENABLE : ID_DISABLE;
  sender->handle(this,MKUINT(msg,SEL_COMMAND),ptr);
  return 1;
  }


// Pass message to all MDI Child windows
long FXMDIClient::forallWindows(FXObject* sender,FXSelector sel,void* ptr){
  register FXbool handled=FALSE;
  register FXMDIChild* child;
  for(child=mdifirst; child; child=child->getMDINext()){
    handled|=child->handle(sender,sel,ptr);
    }
  return handled;
  }


// Pass message to all MDI Child windows whose target is document
long FXMDIClient::forallDocWindows(FXObject* document,FXObject* sender,FXSelector sel,void* ptr){
  register FXbool handled=FALSE;
  register FXMDIChild* child;
  if(document){
    for(child=mdifirst; child; child=child->getMDINext()){
      if(document==child->getTarget()){ handled|=child->handle(sender,sel,ptr); }
      }
    }
  return handled;
  }


// Cascade windows
long FXMDIClient::onCmdCascade(FXObject*,FXSelector,void* ptr){
  register FXMDIChild* child;
  FXint childx,childy,childw,childh;
  childx=5;
  childy=5;
  childw=(2*width)/3;
  childh=(2*height)/3;
  for(child=mdifirst; child; child=child->getMDINext()){
    if(child==active) continue;
    if(child->shown() && !child->isMinimized()){
      child->handle(this,MKUINT(0,SEL_RESTORE),NULL);
      child->position(childx,childy,childw,childh);
      child->raise();
      childx+=cascadex;
      childy+=cascadey;
      if(childx+child->getWidth()>width){ childx=5; childy=5; }
      if(childy+child->getHeight()>height){ childy=5; }
      }
    }
  if(active && active->shown() && !active->isMinimized()){
    active->restore();
    active->position(childx,childy,childw,childh);
    active->raise();
    }
  recalc();
  return 1;
  }


// Update cascade
long FXMDIClient::onUpdCascade(FXObject* sender,FXSelector,void* ptr){
  FXuint msg=mdifirst ? ID_ENABLE : ID_DISABLE;
  sender->handle(this,MKUINT(msg,SEL_COMMAND),ptr);
  return 1;
  }


// Activate next child
long FXMDIClient::onCmdActivateNext(FXObject*,FXSelector,void*){
  if(active && active->getMDINext()){
    handle(this,MKUINT(0,SEL_CHANGED),active->getMDINext());
    }
  return 1;
  }


// Activate next child
long FXMDIClient::onUpdActivateNext(FXObject* sender,FXSelector,void*){
  if(active  && active->getMDINext())
    sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_DISABLE,SEL_COMMAND),NULL);
  return 1;
  }


// Activate previous child
long FXMDIClient::onCmdActivatePrev(FXObject*,FXSelector,void*){
  if(active && active->getMDIPrev()){
    handle(this,MKUINT(0,SEL_CHANGED),active->getMDIPrev());
    }
  return 1;
  }


// Activate previous child
long FXMDIClient::onUpdActivatePrev(FXObject* sender,FXSelector,void* ptr){
  if(active  && active->getMDIPrev())
    sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(ID_DISABLE,SEL_COMMAND),NULL);
  return 1;
  }


// Current child changed
long FXMDIClient::onChanged(FXObject*,FXSelector,void* ptr){
  setActiveChild((FXMDIChild*)ptr);
  if(target) target->handle(this,MKUINT(message,SEL_CHANGED),ptr);
  return 1;
  }


// Recompute interior 
void FXMDIClient::recompute(){
  register FXMDIChild* child;
  register FXint xl,yl,xh,yh;
  xmin= 1000000;
  xmax=-1000000;
  ymin= 1000000;
  ymax=-1000000;
  for(child=mdifirst; child; child=child->getMDINext()){
    if(child->shown()){
      if(child->isMaximized()){
        xl=0;
        yl=0;
        xh=width;
        yh=height;
        }
      else if(child->isMinimized()){
        xl=child->iconPosX;
        yl=child->iconPosY;
        xh=xl+child->getDefaultWidth();
        yh=yl+child->getDefaultHeight();
        }
      else{
        xl=child->normalPosX;
        yl=child->normalPosY;
        xh=xl+child->normalWidth;
        yh=yl+child->normalHeight;
        }
      if(xl<xmin) xmin=xl;
      if(xh>xmax) xmax=xh;
      if(yl<ymin) ymin=yl;
      if(yh>ymax) ymax=yh;
      }
    }
  flags&=~FLAG_RECALC;
  }



// Determine content width of scroll area
FXint FXMDIClient::getContentWidth(){

  // If active child is maximized, client is non-scrollable
  if(active && active->isMaximized()){
    xmin=0;
    ymin=0;
    xmax=width;
    ymax=height;
    flags&=~FLAG_RECALC;
    return width;
    }
  
  // Recalculate contents
  if(flags&FLAG_RECALC) recompute();
  
  // Return the right size
  if(xmin>xmax) return 1;
  if(xmin<0 && xmax<=width) return width-xmin;
  if(0<=xmin && width<xmax) return xmax;
  return xmax-xmin;
  }


// Determine content height of scroll area
FXint FXMDIClient::getContentHeight(){ 
  
  // If active child is maximized, client is non-scrollable
  if(active && active->isMaximized()){
    xmin=0;
    ymin=0;
    xmax=width;
    ymax=height;
    flags&=~FLAG_RECALC;
    return height;
    }
  
  // Recalculate contents
  if(flags&FLAG_RECALC) recompute();
  
  // Return the right size
  if(ymin>ymax) return 1;
  if(ymin<0 && ymax<=height) return height-ymin;
  if(0<=ymin && height<ymax) return ymax;
  return ymax-ymin;
  }


// Move contents; moves child window
void FXMDIClient::moveContents(FXint x,FXint y){
  FXMDIChild *child;
  for(child=mdifirst; child; child=child->getMDINext()){
    child->move(child->getX()+x-pos_x,child->getY()+y-pos_y);
    }
  pos_x=x;
  pos_y=y;
  }


// Recalculate layout
void FXMDIClient::layout(){
  register FXMDIChild* child;
  register FXint xx,yy,ww,hh;
  
  // Place children
  for(child=mdifirst; child; child=child->getMDINext()){
    if(child->shown()){
      if(child->isMaximized()){
        xx=0;
        yy=0;
        ww=width;
        hh=height;
        }
      else if(child->isMinimized()){
        xx=child->iconPosX;
        yy=child->iconPosY;
        ww=child->getDefaultWidth();
        hh=child->getDefaultHeight();
        }
      else{
        xx=child->normalPosX;
        yy=child->normalPosY;
        ww=child->normalWidth;
        hh=child->normalHeight;
        }
      child->position(xx,yy,ww,hh);
      }
    }
    
  // Raise active child
  if(active && active->shown()) active->raise();
  
  // Recompute scroll bars, and raise them on top of everything
  FXScrollArea::layout();
  
  // With new range, adjust scroll bar position X
  if(xmin<0){
    horizontal->setPosition(-xmin);
    pos_x=-horizontal->getPosition();
    }
  
  // With new range, adjust scroll bar position Y
  if(ymin<0){
    vertical->setPosition(-ymin);
    pos_y=-vertical->getPosition();
    }
  
  // No more dirty
  flags&=~FLAG_DIRTY;
  }


// Handle message
long FXMDIClient::handle(FXObject* sender,FXSelector key,void* data){
  
  // Filter out messages for MDIClient; the active MDIChild
  // should not be able to catch any of MDIClient's messages.
  if(FXScrollArea::handle(sender,key,data)) return 1;
  
  // Unknown messages get forwarded to active MDIChild
  return active && active->handle(sender,key,data);
  }


// Save object to stream
void FXMDIClient::save(FXStream& store) const {
  FXScrollArea::save(store);
  store << mdifirst;
  store << mdilast;
  store << active;
  store << cascadex;
  store << cascadey;
  store << xmin;
  store << xmax;
  store << ymin;
  store << ymax;
  }


// Load object from stream
void FXMDIClient::load(FXStream& store){
  FXScrollArea::load(store);
  store >> mdifirst;
  store >> mdilast;
  store >> active;
  store >> cascadex;
  store >> cascadey;
  store >> xmin;
  store >> xmax;
  store >> ymin;
  store >> ymax;
  }  


// Destruct thrashes object
FXMDIClient::~FXMDIClient(){
  mdifirst=(FXMDIChild*)-1;
  mdilast=(FXMDIChild*)-1;
  active=(FXMDIChild*)-1;
  }
