/********************************************************************************
*                                                                               *
*                           T e x t F i e l d   O b j e c t                     *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXTextField.cpp,v 1.88 1999/01/15 04:54:13 jeroen Exp $                  *
********************************************************************************/
#include "xincs.h"
#include "fxdefs.h"
#include "fxkeys.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDC.h"
#include "FXFont.h"
#include "FXCursor.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXLabel.h"
#include "FXTextField.h"


/*
  To do:
  - Add numeric (float/integer) input modes.
  - shift arrow/shift end should highlight between anchor and cursor.
  - Focus into TextField should select all.
  - Need to generalize GUI Update scheme.
  - Double click should select word, not whole text.
  - TextField passes string as the ptr argument in the callback.
  - Should add some messages that allow it to update.
  - GUI Updatable as long as its in the focus chain
  - Double-click should select word, triple click all of text field.
  - TextField should return 0 for all unhandled keys!
  - Emulate FXText more closely, esp as far as key messages
  - Pressing mouse button will set the focus w/o claiming selection!
  - Clipboard types handling.
*/

/*******************************************************************************/

// Map
FXDEFMAP(FXTextField) FXTextFieldMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXTextField::onPaint),
  FXMAPFUNC(SEL_MOTION,0,FXTextField::onMotion),
  FXMAPFUNC(SEL_TIMEOUT,FXWindow::ID_CARETBLINK,FXTextField::onBlink),
  FXMAPFUNC(SEL_TIMEOUT,FXWindow::ID_AUTOSCROLL,FXTextField::onAutoScroll),
  FXMAPFUNC(SEL_ENTER,0,FXTextField::onEnter),
  FXMAPFUNC(SEL_LEAVE,0,FXTextField::onLeave),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,0,FXTextField::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,0,FXTextField::onLeftBtnRelease),
  FXMAPFUNC(SEL_MIDDLEBUTTONPRESS,0,FXTextField::onMiddleBtnPress),
  FXMAPFUNC(SEL_MIDDLEBUTTONRELEASE,0,FXTextField::onMiddleBtnRelease),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS,0,FXTextField::onRightBtnPress),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,0,FXTextField::onRightBtnRelease),
  FXMAPFUNC(SEL_KEYPRESS,0,FXTextField::onKeyPress),
  FXMAPFUNC(SEL_KEYRELEASE,0,FXTextField::onKeyRelease),
  FXMAPFUNC(SEL_CHANGED,0,FXTextField::onChanged),
  FXMAPFUNC(SEL_SELECTION_LOST,0,FXTextField::onSelectionLost),
  FXMAPFUNC(SEL_SELECTION_GAINED,0,FXTextField::onSelectionGained),
  FXMAPFUNC(SEL_SELECTION_REQUEST,0,FXTextField::onSelectionRequest),
  FXMAPFUNC(SEL_FOCUSIN,0,FXTextField::onFocusIn),
  FXMAPFUNC(SEL_FOCUSOUT,0,FXTextField::onFocusOut),
  FXMAPFUNC(SEL_FOCUS_SELF,0,FXTextField::onFocusSelf),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_QUERY_TIP,FXTextField::onQueryTip),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_QUERY_HELP,FXTextField::onQueryHelp),
  FXMAPFUNC(SEL_UPDATE,FXTextField::ID_TEXT_COLOR,FXTextField::onUpdTextColor),
  FXMAPFUNC(SEL_UPDATE,FXTextField::ID_SELTEXT_COLOR,FXTextField::onUpdSelTextColor),
  FXMAPFUNC(SEL_UPDATE,FXTextField::ID_SELBACK_COLOR,FXTextField::onUpdSelBackColor),
  FXMAPFUNC(SEL_UPDATE,FXTextField::ID_TOGGLE_EDITABLE,FXTextField::onUpdToggleEditable),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_SETVALUE,FXTextField::onCmdSetValue),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_SETINTVALUE,FXTextField::onCmdSetIntValue),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_SETREALVALUE,FXTextField::onCmdSetRealValue),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_SETSTRINGVALUE,FXTextField::onCmdSetStringValue),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_GETINTVALUE,FXTextField::onCmdGetIntValue),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_GETREALVALUE,FXTextField::onCmdGetRealValue),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_GETSTRINGVALUE,FXTextField::onCmdGetStringValue),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_CURSOR_HOME,FXTextField::onCmdCursorHome),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_CURSOR_END,FXTextField::onCmdCursorEnd),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_CURSOR_RIGHT,FXTextField::onCmdCursorRight),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_CURSOR_LEFT,FXTextField::onCmdCursorLeft),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_MARK,FXTextField::onCmdMark),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_EXTEND,FXTextField::onCmdExtend),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_SELECT_ALL,FXTextField::onCmdSelectAll),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_DESELECT_ALL,FXTextField::onCmdDeselectAll),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_CUT_SEL,FXTextField::onCmdCutSel),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_COPY_SEL,FXTextField::onCmdCopySel),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_PASTE_SEL,FXTextField::onCmdPasteSel),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_DELETE_SEL,FXTextField::onCmdDeleteSel),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_INSERT_CHAR,FXTextField::onCmdInsertChar),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_INSERT_STRING,FXTextField::onCmdInsertString),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_BACKSPACE,FXTextField::onCmdBackspace),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_DELETE,FXTextField::onCmdDelete),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_TEXT_COLOR,FXTextField::onCmdTextColor),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_SELTEXT_COLOR,FXTextField::onCmdSelTextColor),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_SELBACK_COLOR,FXTextField::onCmdSelBackColor),
  FXMAPFUNC(SEL_COMMAND,FXTextField::ID_TOGGLE_EDITABLE,FXTextField::onCmdToggleEditable),
  };


// Object implementation
FXIMPLEMENT(FXTextField,FXCell,FXTextFieldMap,ARRAYNUMBER(FXTextFieldMap))

  

/*******************************************************************************/


// For serialization
FXTextField::FXTextField(){
  flags|=FLAG_ENABLED;
  font=(FXFont*)-1;
  textColor=0;
  selbackColor=0;
  seltextColor=0;
  cursor=0;
  anchor=0;
  blinker=NULL;
  columns=0;
  scroll=0;
  }


// Construct and init
FXTextField::FXTextField(FXComposite* p,FXuint cols,FXObject* tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h,FXint pl,FXint pr,FXint pt,FXint pb):
  FXCell(p,opts,x,y,w,h,pl,pr,pt,pb){
  contents.fill('\0',cols+1);
  flags|=FLAG_ENABLED;
  target=tgt;
  message=sel;
  defaultCursor=getApp()->textCursor;
  dragCursor=getApp()->textCursor;
  font=getApp()->normalFont;
  backColor=FXRGB(255,255,255);
  textColor=getApp()->foreColor;
  selbackColor=getApp()->selbackColor;
  seltextColor=getApp()->selforeColor;
  cursor=0;
  anchor=0;
  blinker=NULL;
  columns=cols;
  scroll=0;
  }


// Create X window
void FXTextField::create(){
  FXCell::create();
  font->create();
  if(!targetType){targetType=getApp()->registerDragType(targetTypeName);}
  if(!deleteType){deleteType=getApp()->registerDragType(deleteTypeName);}
  if(!textType){textType=getApp()->registerDragType(textTypeName);}
  }


// Change the font
void FXTextField::setFont(FXFont* fnt){
  if(!fnt){ fxerror("%s::setFont: NULL font specified.\n",getClassName()); }
  font=fnt;
  recalc();
  update(0,0,width,height);
  }


// Enable the window
void FXTextField::enable(){
  if(!(flags&FLAG_ENABLED)){
    FXCell::enable();
    update(0,0,width,height);
    }
  }


// Disable the window
void FXTextField::disable(){
  if(flags&FLAG_ENABLED){
    FXCell::disable();
    update(0,0,width,height);
    }
  }


// Get default size
FXint FXTextField::getDefaultWidth(){
  return padleft+padright+(border<<1)+columns*font->getTextWidth("M",1);
  }


FXint FXTextField::getDefaultHeight(){
  return padtop+padbottom+(border<<1)+font->getFontHeight();
  }


// Enter textfield
long FXTextField::onEnter(FXObject* sender,FXSelector sel,void* ptr){
  FXCell::onEnter(sender,sel,ptr);
  return 1;
  }


// Leave textfield
long FXTextField::onLeave(FXObject* sender,FXSelector sel,void* ptr){
  FXCell::onLeave(sender,sel,ptr);
  return 1; 
  }


// We now really do have the selection; repaint the text field
long FXTextField::onSelectionGained(FXObject* sender,FXSelector sel,void* ptr){
  FXCell::onSelectionGained(sender,sel,ptr);
  update(0,0,width,height);
  return 1;
  }


// We lost the selection somehow; repaint the text field
long FXTextField::onSelectionLost(FXObject* sender,FXSelector sel,void* ptr){
  FXCell::onSelectionLost(sender,sel,ptr);
  update(0,0,width,height);
  return 1;
  }



// Somebody wants our selection; the text field will furnish it if the target doesn't
long FXTextField::onSelectionRequest(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  FXDragType *type;
  FXuchar *data; 
  FXuint l,st;
  
  // First, perhaps the target wants to supply its own data for the request
  if(FXCell::onSelectionRequest(sender,sel,ptr)) return 1;
  
  // When asked, supply the list of types
  if(event->target==targetType){
    FXTRACE((100,"%s: Get types list\n",getClassName()));
    FXMALLOC(&type,FXDragType,2);
    type[0]=stringType;
    type[1]=textType;
    setDNDData(event->origin,targetType,(FXuchar*)type,2*sizeof(FXDragType));
    return 1;
    }
    
#ifndef FX_NATIVE_WIN32
  
  // Paste of the selection
  if(event->origin==FROM_SELECTION){
    
    // Otherwise (most likely) its the string from this text we return
    if(event->target==stringType || event->target==textType){
      FXTRACE((100,"%s: Get selection string\n",getClassName()));
      FXuint st;
      if(anchor<cursor){st=anchor;l=cursor-anchor;}else{st=cursor;l=anchor-cursor;}
      FXCALLOC(&data,FXuchar,l+1);
      if(options&TEXTFIELD_PASSWD){
        memset((FXchar*)data,'*',l);      // We shall not reveal the password!
        }
      else{
        memcpy(data,&contents[st],l);
        }
      setDNDData(FROM_SELECTION,stringType,data,l);
      return 1;
      }
    
    // Text delete
    if(event->target==deleteType){
      if(isEditable()){
        FXTRACE((100,"%s: Delete selection\n",getClassName()));
        handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
        }
      return 1;
      }
    return 0;
    }

#endif
  
  // Request for the clipped text
  if(event->origin==FROM_CLIPBOARD){
    
    // Asked for clipped text
    if(event->target==stringType || event->target==textType){
      FXTRACE((100,"%s: Get clipboard string\n",getClassName()));
      l=clipped.length();
      FXCALLOC(&data,FXuchar,l+1);
      if(options&TEXTFIELD_PASSWD){
        memset((FXchar*)data,'*',l);      // We shall not reveal the password!
        }
      else{
        memcpy(data,clipped.text(),l);
        }
      setDNDData(FROM_CLIPBOARD,stringType,data,l);
      return 1;
      }
    
    // Delete clipped text, lost the clipboard
    if(event->target==deleteType){
      FXTRACE((100,"%s: Delete clipboard\n",getClassName()));
      clipped.clear();
      return 1;
      }
    return 0;
    }

  return 0;
  }


// Gained focus
long FXTextField::onFocusIn(FXObject* sender,FXSelector sel,void* ptr){
  FXCell::onFocusIn(sender,sel,ptr);
  if(!blinker) blinker=getApp()->addTimeout(getApp()->blinkSpeed,this,FXWindow::ID_CARETBLINK);
  drawCursor(FLAG_CARET);
  if(hasSelection()){
    update(border,border,width-(border<<1),height-(border<<1));
    }
  return 1;
  }

  
// Lost focus
long FXTextField::onFocusOut(FXObject* sender,FXSelector sel,void* ptr){
  FXCell::onFocusOut(sender,sel,ptr);
  if(blinker){getApp()->removeTimeout(blinker);blinker=NULL;}
  drawCursor(0);
  if(flags&FLAG_CHANGED){
    if(target) target->handle(this,MKUINT(message,SEL_COMMAND),(void*)contents.text());
    flags&=~FLAG_CHANGED;
    }
  if(hasSelection()){
    update(border,border,width-(border<<1),height-(border<<1));
    }
  flags|=FLAG_UPDATE;
  return 1;
  }


// If window can have focus
FXbool FXTextField::canFocus() const { return 1; }


// Focus on widget itself
long FXTextField::onFocusSelf(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  FXTRACE((200,"%s::onFocusSelf %08x\n",getClassName(),this));
  FXCell::onFocusSelf(sender,sel,ptr);
  if(event->type==SEL_KEYPRESS || event->type==SEL_KEYRELEASE){
    handle(this,MKUINT(ID_SELECT_ALL,SEL_COMMAND),NULL);
    }
  return 1;
  }


// We were asked about status text
long FXTextField::onQueryHelp(FXObject* sender,FXSelector,void*){
  if(help.text() && (flags&FLAG_HELP)){ 
    sender->handle(this,MKUINT(ID_SETSTRINGVALUE,SEL_COMMAND),&help);
    return 1;
    }
  return 0;
  }


// We were asked about tip text
long FXTextField::onQueryTip(FXObject* sender,FXSelector,void*){
  if(tip.text() && (flags&FLAG_TIP)){
    sender->handle(this,MKUINT(ID_SETSTRINGVALUE,SEL_COMMAND),&tip);
    return 1;
    }
  return 0;
  }


// Update value from a message
long FXTextField::onCmdSetValue(FXObject*,FXSelector,void* ptr){
  if(ptr){ setText((const FXchar*)ptr); }
  return 1;
  }


// Update value from a message
long FXTextField::onCmdSetIntValue(FXObject*,FXSelector,void* ptr){
  FXchar buf[16];
  if(ptr==NULL){ fxerror("%s::onCmdSetIntValue: NULL pointer.\n",getClassName()); }
  sprintf(buf,"%d",*((FXint*)ptr));
  setText(buf);
  return 1;
  }


// Update value from a message
long FXTextField::onCmdSetRealValue(FXObject*,FXSelector,void* ptr){
  FXchar buf[128];
  if(ptr==NULL){ fxerror("%s::onCmdSetRealValue: NULL pointer.\n",getClassName()); }
  sprintf(buf,"%.10g",*((FXdouble*)ptr));
  setText(buf);
  return 1;
  }


// Update value from a message
long FXTextField::onCmdSetStringValue(FXObject*,FXSelector,void* ptr){
  if(ptr==NULL){ fxerror("%s::onCmdSetStringValue: NULL pointer.\n",getClassName()); }
  setText(*((FXString*)ptr));
  return 1;
  }


// Obtain value from text field
long FXTextField::onCmdGetIntValue(FXObject*,FXSelector,void* ptr){
  if(ptr==NULL){ fxerror("%s::onCmdGetIntValue: NULL pointer.\n",getClassName()); }
  sscanf(contents.text(),"%d",(FXint*)ptr);
  return 1;
  }


// Obtain value from text field
long FXTextField::onCmdGetRealValue(FXObject*,FXSelector,void* ptr){
  if(ptr==NULL){ fxerror("%s::onCmdGetRealValue: NULL pointer.\n",getClassName()); }
  sscanf(contents.text(),"%lf",(FXdouble*)ptr);
  return 1;
  }


// Obtain value from text field
long FXTextField::onCmdGetStringValue(FXObject*,FXSelector,void* ptr){
  if(ptr==NULL){ fxerror("%s::onCmdGetStringValue: NULL pointer.\n",getClassName()); }
  *((FXString*)ptr) = contents;
  return 1;
  }


// Pressed left button
long FXTextField::onLeftBtnPress(FXObject*,FXSelector,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXint pos;
  flags&=~FLAG_TIP;
  if(isEnabled()){
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONPRESS),ptr)) return 1;
    handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
    if(ev->click_count==1){
      pos=index(ev->win_x);
      setCursorPos(pos);
      if(ev->state&SHIFTMASK){
        extendSelection(pos);
        }
      else{
        killSelection();
        setAnchorPos(pos);
        }
      makePositionVisible(pos);
      flags|=FLAG_PRESSED;
      flags&=~FLAG_UPDATE;
      }
    else{
      setAnchorPos(0);
      setCursorPos(contents.length());
      extendSelection(contents.length());
      makePositionVisible(cursor);
      }
    return 1;
    }
  return 0;
  }


// Released left button
long FXTextField::onLeftBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    ungrab();
    flags&=~FLAG_PRESSED;
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr)) return 1;
    return 1;
    }
  return 0;
  }


// Pressed middle button to paste
long FXTextField::onMiddleBtnPress(FXObject*,FXSelector,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXint pos;
  flags&=~FLAG_TIP;
  if(isEnabled()){
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_MIDDLEBUTTONPRESS),ptr)) return 1;
    handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
    pos=index(ev->win_x);
    setCursorPos(pos);
    setAnchorPos(pos);
    makePositionVisible(pos);
    update(border,border,width-(border<<1),height-(border<<1));
    flags&=~FLAG_UPDATE;
    return 1;
    }
  return 0;
  }


// Released middle button causes paste of selection
long FXTextField::onMiddleBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    ungrab();
    if(target && target->handle(this,MKUINT(message,SEL_MIDDLEBUTTONRELEASE),ptr)) return 1;
  
    // Paste text from selection (X-Windows only)
#ifndef FX_NATIVE_WIN32
    if(isEditable()){
      FXuchar *data;
      FXuint   size;
      if(getDNDData(FROM_SELECTION,stringType,data,size)){
        handle(this,MKUINT(ID_INSERT_STRING,SEL_COMMAND),(void*)data);
        FXFREE(&data);
        return 1;
        }
      }
    else{
      getApp()->beep();
      }
#endif
    }
  return 0;
  }


// Pressed right button
long FXTextField::onRightBtnPress(FXObject*,FXSelector,void* ptr){
  flags&=~FLAG_TIP;
  if(isEnabled()){
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_RIGHTBUTTONRELEASE),ptr)) return 1;
    handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
    flags&=~FLAG_UPDATE;
    return 1;
    }
  return 0;
  }


// Released right button
long FXTextField::onRightBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    ungrab();
    if(target && target->handle(this,MKUINT(message,SEL_RIGHTBUTTONRELEASE),ptr)) return 1;
    return 1;
    }
  return 0;
  }


// Moved
long FXTextField::onMotion(FXObject*,FXSelector,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXint t;
  if(flags&FLAG_PRESSED){
    if(ev->win_x<(border+padleft) || (width-padright-border)<ev->win_x){
      getApp()->addTimeout(getApp()->scrollSpeed,this,FXWindow::ID_AUTOSCROLL);
      }
    else{
      t=index(ev->win_x);
      if(t!=cursor){
        drawCursor(0);
        cursor=t;
        extendSelection(cursor);
        }
      }
    return 1;
    }
  return 0;
  }


// Content changed
long FXTextField::onChanged(FXObject*,FXSelector,void* ptr){
  if(target) target->handle(this,MKUINT(message,SEL_CHANGED),ptr);
  return 1;
  }


// Automatic scroll
long FXTextField::onAutoScroll(FXObject*,FXSelector,void*){
  FXuint buttons; FXint x,y,t,ll,rr,ww,tw,lim;
  if(flags&FLAG_PRESSED){
    t=cursor;
    getCursorPosition(x,y,buttons);
    ll=border+padleft;
    rr=width-padright-border;
    ww=rr-ll;
    
    if(options&TEXTFIELD_PASSWD)
      tw=font->getTextWidth("*",1)*contents.length();
    else
      tw=font->getTextWidth(contents.text(),contents.length());
    
    if(options&JUSTIFY_RIGHT){
      lim=tw-ww;
      
      // Scroll left
      if(x<ll){
        scroll+=ll-x;
        if(scroll>=lim) 
          scroll=lim;
        else
          getApp()->addTimeout(getApp()->scrollSpeed,this,FXWindow::ID_AUTOSCROLL);
        t=index(ll);
        }
      
      // Scroll right
      if(rr<x){
        scroll+=rr-x;
        if(scroll<=0) 
          scroll=0;
        else
          getApp()->addTimeout(getApp()->scrollSpeed,this,FXWindow::ID_AUTOSCROLL);
        t=index(rr);
        }
      }
    else{
      lim=ww-tw;
      
      // Scroll left
      if(x<ll){
        scroll+=ll-x;
        if(scroll>=0) 
          scroll=0;
        else
          getApp()->addTimeout(getApp()->scrollSpeed,this,FXWindow::ID_AUTOSCROLL);
        t=index(ll);
        }
  
      // Scroll right
      if(rr<x){
        scroll+=rr-x;
        if(scroll<=lim) 
          scroll=lim;
        else
          getApp()->addTimeout(getApp()->scrollSpeed,this,FXWindow::ID_AUTOSCROLL);
        t=index(rr);
        }
      }
    
    // Extend the selection
    if(t!=cursor){
      drawCursor(0);
      cursor=t;
      extendSelection(cursor);
      }
    }
  return 1;
  }


// Blink the cursor
long FXTextField::onBlink(FXObject*,FXSelector,void*){
  drawCursor(flags^FLAG_CARET);
  blinker=getApp()->addTimeout(getApp()->blinkSpeed,this,FXWindow::ID_CARETBLINK);
  return 0;
  }


// Editable toggle
long FXTextField::onCmdToggleEditable(FXObject*,FXSelector,void*){
  options^=TEXTFIELD_READONLY;
  return 1;
  }


// Update editable toggle
long FXTextField::onUpdToggleEditable(FXObject* sender,FXSelector,void*){
  FXuint msg=(options&TEXTFIELD_READONLY) ? ID_UNCHECK : ID_CHECK;
  sender->handle(this,MKUINT(msg,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(ID_SHOW,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  return 1;
  }


// Move the cursor
void FXTextField::setCursorPos(FXint pos){
  FXint len=contents.length();
  if(pos>len) pos=len; else if(pos<0) pos=0;
  if(cursor!=pos){
    drawCursor(0);
    cursor=pos;
    if(hasFocus()) drawCursor(FLAG_CARET);
    }
  }


// Set anchor position
void FXTextField::setAnchorPos(FXint pos){
  if(pos>(FXint)contents.length()) pos=contents.length(); else if(pos<0) pos=0;
  anchor=pos;
  }


// Draw the cursor
void FXTextField::drawCursor(FXuint state){
  FXint cl,ch,xx,len;
  if(!xid) return;
  if((state^flags)&FLAG_CARET){
    len=contents.length();
    FXASSERT(0<=cursor);
    FXASSERT(cursor<=len);
    xx=coord(cursor)-1;
    FXDC *dc=DC();
    dc->begin(this);
    dc->setClipRectangle(border,border,width-(border<<1),height-(border<<1));
    if(flags&FLAG_CARET){
      dc->setForeground(backColor);
      dc->drawLine(xx,padtop+border,xx,height-border-padbottom-1);
      dc->drawLine(xx-2,padtop+border,xx+2,padtop+border);
      dc->drawLine(xx-2,height-border-padbottom-1,xx+2,height-border-padbottom-1);
      cl=cursor-1;
      ch=cursor+1;
      drawTextRange(FXMAX(cl,0),FXMIN(ch,len));     // Gotta redraw these letters...
      flags&=~FLAG_CARET;
      }
    else{
      dc->setForeground(textColor);
      dc->drawLine(xx,padtop+border,xx,height-border-padbottom-1);
      dc->drawLine(xx-2,padtop+border,xx+2,padtop+border);
      dc->drawLine(xx-2,height-border-padbottom-1,xx+2,height-border-padbottom-1);
      flags|=FLAG_CARET;
      }
    dc->clearClipRectangle();
    dc->end();
    }
  }


// Fix scroll amount after text changes or widget resize
void FXTextField::layout(){
  register FXint ll,rr,ww,tw,oldscroll,len;
  if(!xid) return;
  oldscroll=scroll;
  len=contents.length();
  ll=border+padleft;
  rr=width-padright-border;
  ww=rr-ll;
  if(ww<=0){
    scroll=0;
    }
  else{
    if(options&TEXTFIELD_PASSWD)
      tw=font->getTextWidth("*",1)*len;
    else
      tw=font->getTextWidth(contents.text(),len);
    if(options&JUSTIFY_RIGHT){
      if(scroll<0) scroll=0;
      if(ww>=tw) scroll=0; 
      else if(scroll+ww>tw) scroll=tw-ww;
      FXASSERT(scroll>=0);
      }
    else{
      if(scroll>0) scroll=0;
      if(ww>=tw) scroll=0; 
      else if(scroll+tw<ww) scroll=ww-tw;
      }
    }
  if(scroll!=oldscroll){
    update(0,0,width,height);
    }
  flags&=~FLAG_DIRTY;
  }


// Force position to become fully visible; we assume layout is correct
void FXTextField::makePositionVisible(FXuint pos){
  FXint cw,ll,rr,ww,oldscroll;
  FXuint len;
  if(!xid) return;
  oldscroll=scroll;
  len=contents.length();
  ll=border+padleft;
  rr=width-padright-border;
  ww=rr-ll;
  if(pos>len) pos=len; else if(pos<0) pos=0;
  if(options&JUSTIFY_RIGHT){
    FXASSERT(scroll>=0);
    if(options&TEXTFIELD_PASSWD)
      cw=font->getTextWidth("*",1)*(len-pos);
    else
      cw=font->getTextWidth(&contents[pos],len-pos);
    if(scroll-cw>0) scroll=cw;
    else if(scroll-cw<-ww) scroll=cw-ww;
    FXASSERT(scroll>=0);
    }
  else{
    FXASSERT(scroll<=0);
    if(options&TEXTFIELD_PASSWD)
      cw=font->getTextWidth("*",1)*pos;
    else
      cw=font->getTextWidth(contents.text(),pos);
    if(scroll+cw<0) scroll=-cw;
    else if(scroll+cw>=ww) scroll=ww-cw;
    FXASSERT(scroll<=0);
    }
  if(scroll!=oldscroll){
    update(0,0,width,height);
    }
  }


// Find index from coord
FXint FXTextField::index(FXint x){
  FXint ci,len;
  FXint cx,cw; 
  len=contents.length();
  if(options&JUSTIFY_RIGHT){
    x=x-(scroll+width-padright-border);
    if(x>0) return len;
    if(options&TEXTFIELD_PASSWD){
      cw=font->getTextWidth("*",1);
      ci=len+(x-(cw>>1))/cw;
      if(ci<0) ci=0;
      FXASSERT(0<=ci && ci<=len);
      }
    else{
      cx=0; 
      ci=len;
      while(0<ci){
        FXASSERT(0<ci);
        cw=font->getTextWidth(&contents[ci-1],1);
        if(x>(cx-(cw>>1))) break;
        cx-=cw;
        ci-=1;
        }
      FXASSERT(0<=ci && ci<=len);
      }
    }
  else{
    x=x-scroll-padleft-border;
    if(x<0) return 0;
    if(options&TEXTFIELD_PASSWD){
      cw=font->getTextWidth("*",1);
      ci=(x+(cw>>1))/cw;
      if(ci>len) ci=len;
      FXASSERT(0<=ci && ci<=len);
      }
    else{
      cx=0;
      ci=0;
      while(ci<len){
        FXASSERT(ci<=len);
        cw=font->getTextWidth(&contents[ci],1);
        if(x<(cx+(cw>>1))) break;
        cx+=cw;
        ci+=1;
        }
      FXASSERT(0<=ci && ci<=len);
      }
    }
  return ci;
  }


// Find coordinate from index
FXint FXTextField::coord(FXuint i){
  FXint tx;
  FXASSERT(0<=i);
  FXASSERT(i<=contents.length());
  if(options&JUSTIFY_RIGHT){
    if(options&TEXTFIELD_PASSWD){
      tx=scroll+width-padright-border-font->getTextWidth("*",1)*(contents.length()-i);
      }
    else{
      tx=scroll+width-padright-border-font->getTextWidth(&contents[i],contents.length()-i);
      }
    }
  else{
    if(options&TEXTFIELD_PASSWD){
      tx=scroll+padleft+border+font->getTextWidth("*",1)*i;
      }
    else{
      tx=scroll+padleft+border+font->getTextWidth(contents.text(),i);
      }
    }
  return tx;
  }


// Draw text fragment
void FXTextField::drawTextFragment(FXint x,FXint y,FXuint fm,FXuint to){
  x+=font->getTextWidth(contents.text(),fm);
  y+=font->getFontAscent();
  DC()->drawText(x,y,&contents[fm],to-fm);
  }


// Draw text fragment in password mode
void FXTextField::drawPWDTextFragment(FXint x,FXint y,FXuint fm,FXuint to){
  register FXint cw=font->getTextWidth("*",1);
  register FXuint i;
  y+=font->getFontAscent();
  for(i=fm; i<to; i++){DC()->drawText(x+cw*i,y,"*",1);}
  }


// Draw range of text
void FXTextField::drawTextRange(FXuint fm,FXuint to){
  FXint sx,ex,xx,yy,cw,hh,ww; FXuint si,ei;
  FXDC *dc=DC();
  if(to<=fm) return;
  dc->setTextFont(font);
  dc->setForeground(textColor);

  // Height
  hh=font->getFontHeight();

  // Text centered in y
  if((options&JUSTIFY_TOP) && (options&JUSTIFY_BOTTOM)){
    yy=border+padtop+(height-padbottom-padtop-(border<<1)-hh)/2;
    }
  
  // Text sticks to top of field
  else if(options&JUSTIFY_TOP){
    yy=padtop+border;
    }
  
  // Text sticks to bottom of field
  else if(options&JUSTIFY_BOTTOM){
    yy=height-padbottom-border-hh;
    }
  
  // Text centered in y
  else{
    yy=border+padtop+(height-padbottom-padtop-(border<<1)-hh)/2;
    }

  xx=scroll+padleft+border;

  if(anchor<cursor){si=anchor;ei=cursor;}else{si=cursor;ei=anchor;}

  // Password mode
  if(options&TEXTFIELD_PASSWD){
    cw=font->getTextWidth("*",1);
    ww=cw*contents.length();
    
    // Text sticks to right of field
    if(options&JUSTIFY_RIGHT){
      xx=scroll+width-padright-border-ww;
      }
 
    // Text on left is the default
    else{
      xx=scroll+padleft+border;
      }
    
    // Nothing selected
    if(!hasSelection() || to<=si || ei<=fm){
      drawPWDTextFragment(xx,yy,fm,to);
      }
    
    // Stuff selected
    else{
      if(fm<si){
        drawPWDTextFragment(xx,yy,fm,si);
        }
      else{
        si=fm;
        }
      if(ei<to){
        drawPWDTextFragment(xx,yy,ei,to);
        }
      else{
        ei=to;
        }
      if(si<ei){
        sx=xx+cw*si;
        ex=xx+cw*ei;
        if(hasFocus()){
          dc->setForeground(selbackColor);
          dc->fillRectangle(sx,padtop+border,ex-sx,height-padtop-padbottom-(border<<1));
          dc->setForeground(seltextColor);
          drawPWDTextFragment(xx,yy,si,ei);
          }
        else{
          dc->setForeground(baseColor);
          dc->fillRectangle(sx,padtop+border,ex-sx,height-padtop-padbottom-(border<<1));
          dc->setForeground(textColor);
          drawPWDTextFragment(xx,yy,si,ei);
          }
        }
      }
    }
  
  // Normal mode
  else{
    ww=font->getTextWidth(contents.text(),contents.length());
  
    // Text sticks to right of field
    if(options&JUSTIFY_RIGHT){
      xx=scroll+width-padright-border-ww;
      }
 
    // Text on left is the default
    else{
      xx=scroll+padleft+border;
      }

    // Nothing selected
    if(!hasSelection() || to<=si || ei<=fm){
      drawTextFragment(xx,yy,fm,to);
      }
    
    // Stuff selected
    else{
      if(fm<si){
        drawTextFragment(xx,yy,fm,si);
        }
      else{
        si=fm;
        }
      if(ei<to){
        drawTextFragment(xx,yy,ei,to);
        }
      else{
        ei=to;
        }
      if(si<ei){
        sx=xx+font->getTextWidth(contents.text(),si);
        ex=xx+font->getTextWidth(contents.text(),ei);
        if(hasFocus()){
          dc->setForeground(selbackColor);
          dc->fillRectangle(sx,padtop+border,ex-sx,height-padtop-padbottom-(border<<1));
          dc->setForeground(seltextColor);
          drawTextFragment(xx,yy,si,ei);
          }
        else{
          dc->setForeground(baseColor);
          dc->fillRectangle(sx,padtop+border,ex-sx,height-padtop-padbottom-(border<<1));
          dc->setForeground(textColor);
          drawTextFragment(xx,yy,si,ei);
          }
        }
      }
    }
  }


// Handle repaint 
long FXTextField::onPaint(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *ev=(FXEvent*)ptr;
  FXDC *dc=DC();
  dc->begin(this);
  dc->setClipRectangle(ev->rect.x,ev->rect.y,ev->rect.w,ev->rect.h);
  
  // Draw frame
  drawFrame(0,0,width,height);
  
  // Draw background
  dc->setForeground(backColor);
  dc->fillRectangle(border,border,width-(border<<1),height-(border<<1));
  
  // Draw text, clipped against frame interior
  dc->setClipRectangle(border,border,width-(border<<1),height-(border<<1));
  drawTextRange(0,contents.length());
  dc->clearClipRectangle();
  dc->end();
  return 1;
  }


// Move cursor to begin of line
long FXTextField::onCmdCursorHome(FXObject*,FXSelector,void*){
  setCursorPos(0);
  makePositionVisible(0);
  return 1;
  }


// Move cursor to end of line
long FXTextField::onCmdCursorEnd(FXObject*,FXSelector,void*){
  setCursorPos(contents.length());
  makePositionVisible(cursor);
  return 1;
  }


// Move cursor right
long FXTextField::onCmdCursorRight(FXObject*,FXSelector,void*){
  setCursorPos(cursor+1);
  makePositionVisible(cursor);
  return 1;
  }


// Move cursor left
long FXTextField::onCmdCursorLeft(FXObject*,FXSelector,void*){
  setCursorPos(cursor-1);
  makePositionVisible(cursor);
  return 1;
  }


// Mark
long FXTextField::onCmdMark(FXObject*,FXSelector,void*){
  setAnchorPos(cursor);
  return 1;
  }


// Extend
long FXTextField::onCmdExtend(FXObject*,FXSelector,void*){
  extendSelection(cursor);
  return 1;
  }


// Select All
long FXTextField::onCmdSelectAll(FXObject*,FXSelector,void*){
  setAnchorPos(0);
  setCursorPos(contents.length());
  extendSelection(cursor);
  makePositionVisible(cursor);
  return 1;
  }


// Deselect All
long FXTextField::onCmdDeselectAll(FXObject*,FXSelector,void*){
  killSelection();
  return 1;
  }


// Cut
long FXTextField::onCmdCutSel(FXObject*,FXSelector,void*){
  FXuint st,l;
  if(hasSelection()){
    if(isEditable()){
      if(takeClipboard()){
        if(anchor<cursor){st=anchor;l=cursor-anchor;}else{st=cursor;l=anchor-cursor;}
        clipped=contents.mid(st,l);
        handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
        }
      }
    else{
      getApp()->beep();
      }
    }
  return 1;
  }


// Copy onto cliboard
long FXTextField::onCmdCopySel(FXObject*,FXSelector,void*){
  FXuint st,l;
  if(hasSelection()){
    if(takeClipboard()){
      if(anchor<cursor){st=anchor;l=cursor-anchor;}else{st=cursor;l=anchor-cursor;}
      clipped=contents.mid(st,l);
      }
    }
  return 1;
  }
  

// Paste
long FXTextField::onCmdPasteSel(FXObject*,FXSelector,void*){
  FXuchar *data;
  FXuint   size;
  if(isEditable()){
    if(hasSelection()){
      handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
      }
    if(getDNDData(FROM_CLIPBOARD,stringType,data,size)){
      handle(this,MKUINT(ID_INSERT_STRING,SEL_COMMAND),(void*)data);
      FXFREE(&data);
      return 1;
      }
    }
  else{
    getApp()->beep();
    }
  return 1;
  }



// Delete selection
long FXTextField::onCmdDeleteSel(FXObject*,FXSelector,void*){
  int st,en;
  if(!hasSelection()) return 1;
  st=FXMIN(anchor,cursor); 
  en=FXMAX(anchor,cursor);
  setCursorPos(st);
  contents.remove(st,en-st);
  layout();
  makePositionVisible(st);
  handle(this,MKUINT(0,SEL_CHANGED),(void*)contents.text());
  killSelection();
  flags|=FLAG_CHANGED;
  return 1;
  }


// Insert a character
long FXTextField::onCmdInsertChar(FXObject*,FXSelector sel,void* ptr){
  FXchar ch=(FXchar)(long)ptr;
  contents.insert(cursor,&ch,1);
  layout();
  setCursorPos(cursor+1);
  makePositionVisible(cursor);
  handle(this,MKUINT(0,SEL_CHANGED),(void*)contents.text());
  update(0,0,width,height);
  flags|=FLAG_CHANGED;
  return 1;
  }


// Insert a string
long FXTextField::onCmdInsertString(FXObject*,FXSelector sel,void* ptr){
  FXchar* string=(FXchar*)ptr;
  FXint len;
  if(!string) return 1;
  len=strlen(string);
  contents.insert(cursor,string,len);
  layout();
  setCursorPos(cursor+len);
  makePositionVisible(cursor);
  handle(this,MKUINT(0,SEL_CHANGED),(void*)contents.text());
  update(0,0,width,height);
  flags|=FLAG_CHANGED;
  return 1;
  }
  

// Backspace character
long FXTextField::onCmdBackspace(FXObject*,FXSelector,void*){
  if(cursor<1){ getApp()->beep(); return 1; }
  setCursorPos(cursor-1);
  contents.remove(cursor,1);
  layout();
  makePositionVisible(cursor);
  handle(this,MKUINT(0,SEL_CHANGED),(void*)contents.text());
  update(0,0,width,height);
  flags|=FLAG_CHANGED;
  return 1;
  }


// Delete character
long FXTextField::onCmdDelete(FXObject*,FXSelector,void*){
  if(cursor>=contents.length()){ getApp()->beep(); return 1; }
  contents.remove(cursor,1);
  layout();
  setCursorPos(cursor);
  makePositionVisible(cursor);
  handle(this,MKUINT(0,SEL_CHANGED),(void*)contents.text());
  update(0,0,width,height);
  flags|=FLAG_CHANGED;
  return 1;
  }


// Pressed a key
long FXTextField::onKeyPress(FXObject*,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  char entry;
  flags&=~FLAG_TIP;
  if(!isEnabled()) return 0;
  flags&=~FLAG_UPDATE;
  switch(event->code){
    case KEY_Shift_L:
    case KEY_Shift_R:
    case KEY_Control_L:
    case KEY_Control_R:
    case KEY_Caps_Lock:
    case KEY_Shift_Lock:
    case KEY_Meta_L:
    case KEY_Meta_R:
    case KEY_Alt_L:
    case KEY_Alt_R:
    case KEY_Super_L:
    case KEY_Super_R:
    case KEY_Hyper_L:
    case KEY_Hyper_R:
    case KEY_Scroll_Lock:
    case KEY_Sys_Req:
    case KEY_Tab:
    case KEY_Next:
    case KEY_Prior: 
    case KEY_ISO_Left_Tab:
    case KEY_Up:    
    case KEY_Down:  
      return 0;
    case KEY_Right:
    case KEY_KP_Right:
      if(!(event->state&SHIFTMASK)){
        handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
        }
      handle(this,MKUINT(ID_CURSOR_RIGHT,SEL_COMMAND),NULL);
      if(event->state&SHIFTMASK){
        handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
        }
      else{
        handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
        }
      return 1;
    case KEY_Left:
    case KEY_KP_Left:
      if(!(event->state&SHIFTMASK)){
        handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
        }
      handle(this,MKUINT(ID_CURSOR_LEFT,SEL_COMMAND),NULL);
      if(event->state&SHIFTMASK){
        handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
        }
      else{
        handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
        }
      return 1;
    case KEY_Home:
    case KEY_KP_Home:
      if(!(event->state&SHIFTMASK)){
        handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
        }
      handle(this,MKUINT(ID_CURSOR_HOME,SEL_COMMAND),NULL);
      if(event->state&SHIFTMASK){
        handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
        }
      else{
        handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
        }
      return 1;
    case KEY_End:
    case KEY_KP_End:
      if(!(event->state&SHIFTMASK)){
        handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
        }
      handle(this,MKUINT(ID_CURSOR_END,SEL_COMMAND),NULL);
      if(event->state&SHIFTMASK){
        handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
        }
      else{
        handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
        }
      return 1;
    case KEY_Delete:
    case KEY_KP_Delete:
      if(isEditable()){
        if(hasSelection())
          handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
        else
          handle(this,MKUINT(ID_DELETE,SEL_COMMAND),NULL);
        }
      else{
        getApp()->beep();
        }
      return 1;
    case KEY_BackSpace:
      if(isEditable()){
        if(hasSelection()){
          handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
          }
        else{
          handle(this,MKUINT(ID_BACKSPACE,SEL_COMMAND),NULL);
          }
        }
      else{
        getApp()->beep();
        }
      return 1;
    case KEY_Return:
    case KEY_KP_Enter:
      if(isEditable()){
        if(flags&FLAG_CHANGED){
          if(target) target->handle(this,MKUINT(message,SEL_COMMAND),(void*)contents.text());
          flags&=~FLAG_CHANGED;
          }
        flags|=FLAG_UPDATE;
        }
      else{
        getApp()->beep();
        }
      return 1;
    default:
      if(!(event->state&ALTMASK)){
        entry=fxkeyval(event->code,event->state);
        switch(entry){
          case 'A'-'@':
            handle(this,MKUINT(ID_SELECT_ALL,SEL_COMMAND),NULL);
            return 1;
          case 'X'-'@':
            if(isEditable()){
              handle(this,MKUINT(ID_CUT_SEL,SEL_COMMAND),NULL);
              }
            else{
              getApp()->beep();
              }
            return 1;
          case 'C'-'@':
            handle(this,MKUINT(ID_COPY_SEL,SEL_COMMAND),NULL);
            return 1;
          case 'V'-'@':
            if(isEditable()){
              handle(this,MKUINT(ID_PASTE_SEL,SEL_COMMAND),NULL);
              }
            else{
              getApp()->beep();
              }
            return 1;
          default:
            if(isEditable()){
              if(hasSelection()){
                handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
                }
              if(' '<=entry && entry<='~'){
                //FXTRACE((10,"Hit code=%x state=%x entry=%x\n",event->code,event->state,entry));
                handle(this,MKUINT(ID_INSERT_CHAR,SEL_COMMAND),(void*)entry);
                return 1;
                }
              }
            else{
              getApp()->beep();
              }
            break;
          }
        }
      break;
    }
  return 0;
  }


// Key Release 
long FXTextField::onKeyRelease(FXObject*,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  char entry;
  if(!isEnabled()) return 0;
  switch(event->code){
    case KEY_Shift_L:
    case KEY_Shift_R:
    case KEY_Control_L:
    case KEY_Control_R:
    case KEY_Caps_Lock:
    case KEY_Shift_Lock:
    case KEY_Meta_L:
    case KEY_Meta_R:
    case KEY_Alt_L:
    case KEY_Alt_R:
    case KEY_Super_L:
    case KEY_Super_R:
    case KEY_Hyper_L:
    case KEY_Hyper_R:
    case KEY_Scroll_Lock:
    case KEY_Sys_Req:
    case KEY_Tab:
    case KEY_Next:
    case KEY_Prior: 
    case KEY_ISO_Left_Tab:
    case KEY_Up:    
    case KEY_Down:  
      return 0;
    case KEY_Right:
    case KEY_KP_Right:
    case KEY_Left:
    case KEY_KP_Left:
    case KEY_Home:
    case KEY_KP_Home:
    case KEY_End:
    case KEY_KP_End:
    case KEY_Delete:
    case KEY_KP_Delete:
    case KEY_BackSpace:
    case KEY_Return:
    case KEY_KP_Enter:
      return 1;
    default:
      if(!(event->state&ALTMASK)){
        entry=fxkeyval(event->code,event->state);
        if(entry<' ' || entry>'~') return 0;
        return 1;
        }
      break;
    }
  return 0;
  }

// Kill the selection
FXbool FXTextField::killSelection(){
  if(hasSelection()){
    releaseSelection();
    update(0,0,width,height);
    return TRUE;
    }
  return FALSE;
  }


// Extend selection
FXbool FXTextField::extendSelection(FXuint pos){
  
  // Validate position
  if(pos>contents.length()) pos=contents.length(); else if(pos<0) pos=0; 
   
  // Got a selection at all?
  if(anchor!=pos){
    acquireSelection();
    }
  else{
    releaseSelection();
    }
  
  update(0,0,width,height);
  return TRUE;
  }


// Set text being displayed
void FXTextField::setText(const FXString& text){
  releaseSelection();
  if(contents!=text){
    setCursorPos(0);
    contents=text;
    scroll=0;
    layout();
    makePositionVisible(0);
    update(border,border,width-(border<<1),height-(border<<1));
    }
  }


// Set text color
long FXTextField::onCmdTextColor(FXObject*,FXSelector,void* ptr){
  setTextColor((FXColor)ptr);
  return 1;
  }


// Set selection color
long FXTextField::onCmdSelTextColor(FXObject*,FXSelector,void* ptr){
  setSelTextColor((FXColor)ptr);
  return 1;
  }


// Set selection background color
long FXTextField::onCmdSelBackColor(FXObject*,FXSelector,void* ptr){
  setSelBackColor((FXColor)ptr);
  return 1;
  }


// Update text color
long FXTextField::onUpdTextColor(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(FXWindow::ID_SETVALUE,SEL_COMMAND),(void*)textColor);
  return 1;
  }


// Update selection color
long FXTextField::onUpdSelTextColor(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(FXWindow::ID_SETVALUE,SEL_COMMAND),(void*)seltextColor);
  return 1;
  }


// Update selection background color
long FXTextField::onUpdSelBackColor(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(FXWindow::ID_SETVALUE,SEL_COMMAND),(void*)selbackColor);
  return 1;
  }


// Set text color
void FXTextField::setTextColor(FXColor clr){
  textColor=clr;
  update(0,0,width,height);
  }


// Set select background color
void FXTextField::setSelBackColor(FXColor clr){
  selbackColor=clr;
  update(0,0,width,height);
  }


// Set selected text color
void FXTextField::setSelTextColor(FXColor clr){
  seltextColor=clr;
  update(0,0,width,height);
  }


// Change number of columns
void FXTextField::setNumColumns(FXuint cols){
  if(columns!=cols){
    scroll=0;
    columns=cols;
    layout();
    update(0,0,width,height);
    }
  }


// Change help text
void FXTextField::setHelpText(const FXString& text){
  help=text;
  }


// Change tip text
void FXTextField::setTipText(const FXString& text){
  tip=text;
  }


// Return true if editable
FXbool FXTextField::isEditable() const { 
  return (options&TEXTFIELD_READONLY)==0; 
  }


// Set widget is editable or not
void FXTextField::setEditable(FXbool edit){
  if(edit) options&=~TEXTFIELD_READONLY; else options|=TEXTFIELD_READONLY;
  }


// Change text style
void FXTextField::setTextStyle(FXuint style){
  FXuint opts=(options&~TEXTFIELD_MASK) | (style&TEXTFIELD_MASK);
  if(options!=opts){
    scroll=0;
    options=opts;
    recalc();
    update(0,0,width,height);
    }
  }
  
  
// Get text style
FXuint FXTextField::getTextStyle() const {
  return (options&TEXTFIELD_MASK); 
  }


// Set text justify style
void FXTextField::setJustify(FXuint style){
  FXuint opts=(options&~JUSTIFY_MASK) | (style&JUSTIFY_MASK);
  if(options!=opts){
    scroll=0;
    options=opts;
    recalc();
    update(0,0,width,height);
    }
  }


// Get text justify style
FXuint FXTextField::getJustify() const { 
  return (options&JUSTIFY_MASK); 
  }


// Save object to stream
void FXTextField::save(FXStream& store) const {
  FXCell::save(store);
  store << contents;
  store << font;
  store << textColor;
  store << selbackColor;
  store << seltextColor;
  store << columns;
  store << help;
  store << tip;
  }


// Load object from stream
void FXTextField::load(FXStream& store){
  FXCell::load(store);
  store >> contents;
  store >> font;
  store >> textColor;
  store >> selbackColor;
  store >> seltextColor;
  store >> columns;
  store >> help;
  store >> tip;
  }  


// Clean up
FXTextField::~FXTextField(){
  if(blinker) getApp()->removeTimeout(blinker);
  font=(FXFont*)-1;
  blinker=(FXTimer*)-1;
  }
