/********************************************************************************
*                                                                               *
*                   T o g g l e    B u t t o n    O b j e c t                   *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXToggleButton.cpp,v 1.15 1999/01/15 04:54:13 jeroen Exp $               *
********************************************************************************/
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDC.h"
#include "FXFont.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXLabel.h"
#include "FXToggleButton.h"

/*
  To do:
*/

/*******************************************************************************/

// Map
FXDEFMAP(FXToggleButton) FXToggleButtonMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXToggleButton::onPaint),
  FXMAPFUNC(SEL_UPDATE,0,FXToggleButton::onUpdate),
  FXMAPFUNC(SEL_ACTIVATE,0,FXToggleButton::onActivate),
  FXMAPFUNC(SEL_DEACTIVATE,0,FXToggleButton::onDeactivate),
  FXMAPFUNC(SEL_ENTER,0,FXToggleButton::onEnter),
  FXMAPFUNC(SEL_LEAVE,0,FXToggleButton::onLeave),
  FXMAPFUNC(SEL_FOCUSIN,0,FXToggleButton::onFocusIn),
  FXMAPFUNC(SEL_FOCUSOUT,0,FXToggleButton::onFocusOut),
  FXMAPFUNC(SEL_KEYPRESS,FXWindow::ID_HOTKEY,FXWindow::onHotKeyPress),
  FXMAPFUNC(SEL_KEYRELEASE,FXWindow::ID_HOTKEY,FXWindow::onHotKeyRelease),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_QUERY_TIP,FXToggleButton::onQueryTip),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_QUERY_HELP,FXToggleButton::onQueryHelp),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_CHECK,FXToggleButton::onCheck),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_UNCHECK,FXToggleButton::onUncheck),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_SETVALUE,FXToggleButton::onCmdSetValue),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_SETINTVALUE,FXToggleButton::onCmdSetIntValue),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_GETINTVALUE,FXToggleButton::onCmdGetIntValue),
  };


// Object implementation
FXIMPLEMENT(FXToggleButton,FXLabel,FXToggleButtonMap,ARRAYNUMBER(FXToggleButtonMap))

  
// Deserialization
FXToggleButton::FXToggleButton(){
  state=FALSE;
  oldstate=FALSE;
  noci=(FXIcon*)-1;
  }


// Construct and init
FXToggleButton::FXToggleButton(FXComposite* p,const FXString& text,FXIcon* ic1,FXIcon* ic2,FXObject* tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h,FXint pl,FXint pr,FXint pt,FXint pb):
  FXLabel(p,text,ic1,opts,x,y,w,h,pl,pr,pt,pb){
  tip=text.extract(1,'\t');
  help=text.extract(2,'\t');
  noci=ic2;
  flags|=FLAG_ENABLED;
  target=tgt;
  message=sel;
  dragCursor=getApp()->rarrowCursor;
  state=FALSE;
  oldstate=FALSE;
  }


// Create X window
void FXToggleButton::create(){
  FXLabel::create();
  if(noci) noci->create();
  }
 

// Enable the window
void FXToggleButton::enable(){
  if(!(flags&FLAG_ENABLED)){
    FXLabel::enable();
    update(0,0,width,height);
    }
  }


// Disable the window
void FXToggleButton::disable(){
  if(flags&FLAG_ENABLED){
    FXLabel::disable();
    update(0,0,width,height);
    }
  }


// Set button state
void FXToggleButton::setState(FXuint s){
  if(state!=s){
    state=s;
    update(0,0,width,height);
    }
  }


// If window can have focus
FXbool FXToggleButton::canFocus() const { return 1; }


// Update value from a message
long FXToggleButton::onCmdSetValue(FXObject*,FXSelector,void* ptr){
  setState((FXint)ptr);
  return 1;
  }


// Update value from a message
long FXToggleButton::onCmdSetIntValue(FXObject*,FXSelector,void* ptr){
  if(ptr==NULL){ fxerror("%s::onCmdSetIntValue: NULL pointer.\n",getClassName()); }
  setState(*((FXint*)ptr));
  return 1;
  }


// Obtain value from text field
long FXToggleButton::onCmdGetIntValue(FXObject*,FXSelector,void* ptr){
  if(ptr==NULL){ fxerror("%s::onCmdGetIntValue: NULL pointer.\n",getClassName()); }
  *((FXint*)ptr)=getState();
  return 1;
  }


// Check the menu button
long FXToggleButton::onCheck(FXObject*,FXSelector,void*){ 
  setState(TRUE); 
  return 1; 
  }


// Check the menu button
long FXToggleButton::onUncheck(FXObject*,FXSelector,void*){ 
  setState(FALSE); 
  return 1; 
  }


// Gained focus
long FXToggleButton::onFocusIn(FXObject* sender,FXSelector sel,void* ptr){
  FXLabel::onFocusIn(sender,sel,ptr);
  update(border,border,width-(border<<1),height-(border<<1));
  return 1;
  }

  
// Lost focus
long FXToggleButton::onFocusOut(FXObject* sender,FXSelector sel,void* ptr){
  FXLabel::onFocusOut(sender,sel,ptr);
  update(border,border,width-(border<<1),height-(border<<1));
  return 1;
  }


// Implement auto-hide or auto-gray modes
long FXToggleButton::onUpdate(FXObject* sender,FXSelector sel,void* ptr){
  if(!FXLabel::onUpdate(sender,sel,ptr)){
    if(options&TOGGLEBUTTON_AUTOHIDE){if(shown()){hide();recalc();}}
    if(options&TOGGLEBUTTON_AUTOGRAY){disable();}
    }
  return 1;
  }


// Button being activated
long FXToggleButton::onActivate(FXObject*,FXSelector,void*){
  flags|=FLAG_PRESSED;
  flags&=~FLAG_UPDATE;
  oldstate=state;
  setState(!oldstate);
  return 0;
  }
  

// Button being deactivated
long FXToggleButton::onDeactivate(FXObject*,FXSelector,void*){
  flags&=~FLAG_PRESSED;
  flags|=FLAG_UPDATE;
  update(0,0,width,height);
  if(state!=oldstate){
    if(target && target->handle(this,MKUINT(message,SEL_COMMAND),(void*)state)) return 1;
    }
  return 1;
  }


// We were asked about status text
long FXToggleButton::onQueryHelp(FXObject* sender,FXSelector,void*){
  if(help.text() && (flags&FLAG_HELP)){
    sender->handle(this,MKUINT(ID_SETSTRINGVALUE,SEL_COMMAND),(void*)&help);
    return 1;
    }
  return 0;
  }


// We were asked about tip text
long FXToggleButton::onQueryTip(FXObject* sender,FXSelector,void*){
  if(tip.text() && (flags&FLAG_TIP)){
    sender->handle(this,MKUINT(ID_SETSTRINGVALUE,SEL_COMMAND),(void*)&tip);
    return 1;
    }
  return 0;
  }


// Entered button
long FXToggleButton::onEnter(FXObject* sender,FXSelector sel,void* ptr){
  FXLabel::onEnter(sender,sel,ptr);
  if(isEnabled()){
    if(flags&FLAG_PRESSED) setState(!oldstate);
    if(options&TOGGLEBUTTON_TOOLBAR) update(0,0,width,height);
    return 1;
    }
  return 0;
  }


// Left button
long FXToggleButton::onLeave(FXObject* sender,FXSelector sel,void* ptr){
  FXLabel::onLeave(sender,sel,ptr);
  if(isEnabled()){
    if(flags&FLAG_PRESSED) setState(oldstate);
    if(options&TOGGLEBUTTON_TOOLBAR) update(0,0,width,height);
    return 1;
    }
  return 0;
  }


// Handle repaint 
long FXToggleButton::onPaint(FXObject*,FXSelector,void* ptr){
  FXint tw=0,th=0,iw=0,ih=0,tx,ty,ix,iy;
  FXEvent *ev=(FXEvent*)ptr;
  FXDC *dc=DC();
  dc->begin(this);
  dc->setClipRectangle(ev->rect.x,ev->rect.y,ev->rect.w,ev->rect.h);
  
  // Got a border at all?
  if(options&(FRAME_RAISED|FRAME_SUNKEN)){

    // Toolbar style
    if(options&TOGGLEBUTTON_TOOLBAR){
      
      // Enabled and cursor inside, and up
      if(isEnabled() && underCursor() && !(flags&FLAG_PRESSED)){
        dc->setForeground(backColor);
        dc->fillRectangle(border,border,width-border*2,height-border*2);
        if(options&FRAME_THICK) drawDoubleRaisedRectangle(0,0,width,height);
        else drawRaisedRectangle(0,0,width,height);
        }
      
      // Enabled and cursor inside and down
      else if(isEnabled() && underCursor() && (flags&FLAG_PRESSED)){
        dc->setForeground(hiliteColor);
        dc->fillRectangle(border,border,width-border*2,height-border*2);
        if(options&FRAME_THICK) drawDoubleSunkenRectangle(0,0,width,height);
        else drawSunkenRectangle(0,0,width,height);
        }
      
      // Disabled or unchecked or not under cursor
      else{
        dc->setForeground(backColor);
        dc->fillRectangle(0,0,width,height);
        }
      }
 
    // Normal style
    else{
      
      // Default
      if(isDefault()){
        
        // Black default border
        drawBorderRectangle(0,0,width,height);
        
        // Draw in up state if disabled or up
        if(!(flags&FLAG_PRESSED)){
          dc->setForeground(backColor);
          dc->fillRectangle(border+1,border+1,width-border*2-2,height-border*2-2);
          if(options&FRAME_THICK) drawDoubleRaisedRectangle(1,1,width-1,height-1);
          else drawRaisedRectangle(1,1,width-1,height-1);
          }
        
        // Draw sunken if enabled and either checked or pressed
        else{
          dc->setForeground(hiliteColor);
          dc->fillRectangle(border,border,width-border*2-1,height-border*2-1);
          if(options&FRAME_THICK) drawDoubleSunkenRectangle(0,0,width-1,height-1);
          else drawSunkenRectangle(0,0,width-1,height-1);
          }
        }
      
      // Non-Default
      else{
        
        // Draw raised if not currently pressed down
        if(!(flags&FLAG_PRESSED)){
          dc->setForeground(backColor);
          dc->fillRectangle(border,border,width-border*2,height-border*2);
          if(options&FRAME_THICK) drawDoubleRaisedRectangle(0,0,width,height);
          else drawRaisedRectangle(0,0,width,height);
          }
        
        // Draw sunken if pressed
        else{
          dc->setForeground(hiliteColor);
          dc->fillRectangle(border,border,width-border*2,height-border*2);
          if(options&FRAME_THICK) drawDoubleSunkenRectangle(0,0,width,height);
          else drawSunkenRectangle(0,0,width,height);
          }
        }
      }
    }
  
  // No borders
  else{
    dc->setForeground(backColor);
    dc->fillRectangle(0,0,width,height);
    }
  
  // Place text & icon
  if(label.text()){
    tw=labelWidth();
    th=labelHeight();
    }
  if(state){
    if(noci){
      iw=noci->getWidth();
      ih=noci->getHeight();
      }
    }
  else{
    if(icon){
      iw=icon->getWidth();
      ih=icon->getHeight();
      }
    }
  
  just_x(tx,ix,tw,iw);
  just_y(ty,iy,th,ih);
  
  // Shift a bit when pressed
  if(flags&FLAG_PRESSED){ ++tx; ++ty; ++ix; ++iy; }

  // Draw the icons
  if(!isEnabled()){
    if(state){
      if(noci) dc->drawIconShaded(noci,ix,iy);
      }
    else{
      if(icon) dc->drawIconShaded(icon,ix,iy);
      }
    }
  else{
    if(state){ 
      if(noci) dc->drawIcon(noci,ix,iy); 
      }
    else{ 
      if(icon) dc->drawIcon(icon,ix,iy); 
      }
    }
  
  // Draw the text
  if(label.text()){
    dc->setTextFont(font);
    if(isEnabled()){
      dc->setForeground(textColor);
      drawLabel(tx,ty,tw,th);
      if(hasFocus()){
        drawFocusRectangle(border+2,border+2,width-2*border-4,height-2*border-4);
        }
      }
    else{
      dc->setForeground(hiliteColor);
      drawLabel(tx+1,ty+1,tw,th);
      dc->setForeground(shadowColor);
      drawLabel(tx,ty,tw,th);
      }
    }
  dc->clearClipRectangle();
  dc->end();
  return 1;
  }


// Change help text
void FXToggleButton::setHelpText(const FXString& text){
  help=text;
  }


// Change tip text
void FXToggleButton::setTipText(const FXString& text){
  tip=text;
  }


// Save object to stream
void FXToggleButton::save(FXStream& store) const {
  FXLabel::save(store);
  store << noci;
  store << tip;
  store << help;
  }

      

// Load object from stream
void FXToggleButton::load(FXStream& store){
  FXLabel::load(store);
  store >> noci;
  store >> tip;
  store >> help;
  }  


// Destruct
FXToggleButton::~FXToggleButton(){
  noci=(FXIcon*)-1;
  }
