/********************************************************************************
*                                                                               *
*                                 Test Tab Book                                 *
*                                                                               *
*********************************************************************************
* Copyright (C) 1997 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* $Id: tabbook.cpp,v 1.6 1999/01/06 07:00:05 jeroen Exp $                       *
********************************************************************************/
#include "fx.h"



  
  
/*******************************************************************************/


// Mini application object
class TabBookApp : public FXApp {
  FXDECLARE(TabBookApp)
protected:
    
  // Member data
  FXMainWindow*      mainwindow;
  FXTooltip*         tooltip;
  FXMenuBar*         menubar;
  FXMenuPane*        filemenu;
  FXMenuPane*        tabmenu;
  FXHorizontalFrame* contents;
  FXTabBook*         tabbook;
  FXTabItem*         tab1;
  FXTabItem*         tab2;
  FXTabItem*         tab3;
  FXHorizontalFrame* listframe;
  FXHorizontalFrame* fileframe;
  FXHorizontalFrame* dirframe;
  FXList*            simplelist;   
  FXFileList*        filelist;
  FXDirList*         dirlist;
public:
  long onCmdTabOrient(FXObject*,FXSelector,void*);
public:
  enum{
    ID_TABS_TOP=FXApp::ID_LAST,
    ID_TABS_BOTTOM,
    ID_TABS_LEFT,
    ID_TABS_RIGHT
    };
public:
  TabBookApp();
  void create();
  };


/*******************************************************************************/
  
// Map
FXDEFMAP(TabBookApp) TabBookAppMap[]={
  FXMAPFUNCS(SEL_COMMAND,TabBookApp::ID_TABS_TOP,TabBookApp::ID_TABS_RIGHT,TabBookApp::onCmdTabOrient),
  };


// Object implementation
FXIMPLEMENT(TabBookApp,FXApp,TabBookAppMap,ARRAYNUMBER(TabBookAppMap))


  
  
/*******************************************************************************/

  
// Make some windows
TabBookApp::TabBookApp():FXApp("TabBook","Test"){

  // Main Window
  mainwindow=new FXMainWindow(this,"Switcher Test",NULL,NULL,DECOR_ALL,0,0,600,400);
  
  // Tooltip
  tooltip=new FXTooltip(this);
  
  // Menubar
  menubar=new FXMenuBar(mainwindow,LAYOUT_SIDE_TOP|LAYOUT_FILL_X);
  
  // Separator
  new FXHorizontalSeparator(mainwindow,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|SEPARATOR_GROOVE);

  // Contents
  contents=new FXHorizontalFrame(mainwindow,LAYOUT_SIDE_TOP|FRAME_NONE|LAYOUT_FILL_X|LAYOUT_FILL_Y|PACK_UNIFORM_WIDTH);
  
  // Switcher
  tabbook=new FXTabBook(contents,NULL,0,LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_RIGHT);
  
  // First item is a list
  tab1=new FXTabItem(tabbook,"&Simple List",NULL);
  listframe=new FXHorizontalFrame(tabbook,FRAME_THICK|FRAME_RAISED);
  simplelist=new FXList(listframe,NULL,0,LIST_EXTENDEDSELECT|LAYOUT_FILL_X|LAYOUT_FILL_Y);
  simplelist->addItemLast("First Entry");
  simplelist->addItemLast("Second Entry");
  simplelist->addItemLast("Third Entry");
  simplelist->addItemLast("Fourth Entry");
    
  // Second item is a file list
  tab2=new FXTabItem(tabbook,"F&ile List",NULL);
  fileframe=new FXHorizontalFrame(tabbook,FRAME_THICK|FRAME_RAISED);
  filelist=new FXFileList(fileframe,NULL,0,ICONLIST_EXTENDEDSELECT|LAYOUT_FILL_X|LAYOUT_FILL_Y);
  
  // Third item is a directory list
  tab3=new FXTabItem(tabbook,"T&ree List",NULL);
  dirframe=new FXHorizontalFrame(tabbook,FRAME_THICK|FRAME_RAISED);
  dirlist=new FXDirList(dirframe,NULL,0,DIRLIST_SHOWFILES|TREELIST_SHOWS_LINES|TREELIST_SHOWS_BOXES|LAYOUT_FILL_X|LAYOUT_FILL_Y);
  
  // File Menu
  filemenu=new FXMenuPane(this);
  new FXMenuCommand(filemenu,"&Simple List",NULL,tabbook,FXTabBar::ID_OPEN_FIRST+0);
  new FXMenuCommand(filemenu,"F&ile List",NULL,tabbook,FXTabBar::ID_OPEN_FIRST+1);
  new FXMenuCommand(filemenu,"T&ree List",NULL,tabbook,FXTabBar::ID_OPEN_FIRST+2);
  new FXMenuCommand(filemenu,"&Quit",NULL,this,ID_QUIT);
  new FXMenuTitle(menubar,"&File",NULL,filemenu);
  
  // Tab side
  tabmenu=new FXMenuPane(this);
  new FXMenuCommand(tabmenu,"&Top Tabs",NULL,this,TabBookApp::ID_TABS_TOP);
  new FXMenuCommand(tabmenu,"&Bottom Tabs",NULL,this,TabBookApp::ID_TABS_BOTTOM);
  new FXMenuCommand(tabmenu,"&Left Tabs",NULL,this,TabBookApp::ID_TABS_LEFT);
  new FXMenuCommand(tabmenu,"&Right Tabs",NULL,this,TabBookApp::ID_TABS_RIGHT);
  new FXMenuTitle(menubar,"&Tab Placement",NULL,tabmenu);
  
  }


// Switch tab orientations
long TabBookApp::onCmdTabOrient(FXObject*,FXSelector sel,void*){
  switch(SELID(sel)){
    case ID_TABS_TOP:
      tabbook->setTabStyle(TABBOOK_TOPTABS);
      tab1->setTabOrientation(TAB_TOP);
      tab2->setTabOrientation(TAB_TOP);
      tab3->setTabOrientation(TAB_TOP);
      break;
    case ID_TABS_BOTTOM:
      tabbook->setTabStyle(TABBOOK_BOTTOMTABS);
      tab1->setTabOrientation(TAB_BOTTOM);
      tab2->setTabOrientation(TAB_BOTTOM);
      tab3->setTabOrientation(TAB_BOTTOM);
      break;
    case ID_TABS_LEFT:
      tabbook->setTabStyle(TABBOOK_LEFTTABS);
      tab1->setTabOrientation(TAB_LEFT);
      tab2->setTabOrientation(TAB_LEFT);
      tab3->setTabOrientation(TAB_LEFT);
      break;
    case ID_TABS_RIGHT:
      tabbook->setTabStyle(TABBOOK_RIGHTTABS);
      tab1->setTabOrientation(TAB_RIGHT);
      tab2->setTabOrientation(TAB_RIGHT);
      tab3->setTabOrientation(TAB_RIGHT);
      break;
    }
  return 1;
  }


// Start
void TabBookApp::create(){
  FXApp::create();
  mainwindow->show();
  }


/*******************************************************************************/


// Start the whole thing
int main(int argc,char *argv[]){

  // Make application
  TabBookApp* application=new TabBookApp;
  
  // Open display
  application->init(argc,argv);
  
  // Create app
  application->create();
  
  // Run
  application->run();
  }


