/********************************************************************************
*                                                                               *
*               Little Helper Program to Compute Color Ramp                     *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: optimalramp.c,v 1.3 1998/11/19 20:07:23 jeroen Exp $                     *
********************************************************************************/

/*

	This program is used exactly once, to build a table.  It is 
	provided basically for your benefit.

	Build color palette based on 6*6*6 cube (216 entries).

	Then sort the colors such that the most important colors
	are at the top of the palette.

	Thus, a reasonable gamut may be established even if we
	don't get them all.

	The two most important colors are black and white, then
	combinations of red, green, blue primaries, and so on.

	Colors are added one at a time, taking the color furthest
	away from the current set each time.

*/

#include <stdio.h>
#include <math.h>


int palette[256][3];

int ramp[256][3];

int same[256];

int nsame;
int nramp;
int npalette;


#define XABS(a) ((a)<0?-(a):(a))

main(){
  int r,g,b,i,j,best;
  double dr,dg,db,dist,d,dd;
  int red,green,blue;


  /* buid 6*6*6 ramp */
  nramp=0;
  for(r=0; r<6; r++){
    red=(255*r)/5;
    for(g=0; g<6; g++){
      green=(255*g)/5;
      for(b=0; b<6; b++){
        blue=(255*b)/5;

        ramp[nramp][0]=red;
        ramp[nramp][1]=green;
        ramp[nramp][2]=blue;
        nramp++;
        }
      }
    }
/*
  for(i=0; i<nramp; i++){
    printf("ramp[%3d]=%3d %3d %3d\n",i,ramp[i][0],ramp[i][1],ramp[i][2]);
    }
*/

  /* Build optimal table, most important colors first */
  palette[0][0]=ramp[0][0];
  palette[0][1]=ramp[0][1];
  palette[0][2]=ramp[0][2];
  npalette=1;
  ramp[0][0]=ramp[nramp-1][0];
  ramp[0][1]=ramp[nramp-1][1];
  ramp[0][2]=ramp[nramp-1][2];
  nramp--;
  while(nramp!=0){
    
    /* Find color farthest from the bunch we have already */
    dist=-10000000.0;
    best=-1;
    nsame=0;
    for(j=0; j<nramp; j++){

      /* Find closest distance to current set */
      dd=10000000.0;
      for(i=0; i<npalette; i++){
        dr=(double)(palette[i][0]-ramp[j][0]);
        dg=(double)(palette[i][1]-ramp[j][1]);
        db=(double)(palette[i][2]-ramp[j][2]);
        dr=fabs(dr); 
        dg=fabs(dg); 
        db=fabs(db);
        d=dr;
        if(dg>d) d=dg;
        if(db>d) d=db;
        if(d<dd){dd=d;}
        }
      if(dd>dist){best=j;dist=dd;}
      }

    /* Should have found one */
    if(best!=-1){
/*
      printf("fartest from:\n");
      for(i=0; i<npalette; i++){
        printf("palette[%3d]=%3d %3d %3d\n",i,palette[i][0],palette[i][1],palette[i][2]);
        }
      printf("best: %3d  [%3d %3d %3d] (dist=%10.1f)\n",best,ramp[best][0],ramp[best][1],ramp[best][2],dist);
*/
      palette[npalette][0]=ramp[best][0];
      palette[npalette][1]=ramp[best][1];
      palette[npalette][2]=ramp[best][2];
      npalette++;
      ramp[best][0]=ramp[nramp-1][0];
      ramp[best][1]=ramp[nramp-1][1];
      ramp[best][2]=ramp[nramp-1][2];
      nramp--;
      }
    }

  /* Spit out the table */
  printf("/* Colors in decreasing order of importance */\n");
  printf("unsigned char palette[][3]={\n");
  for(i=0; i<npalette; i++){
    printf("  {%3d,%3d,%3d},\n",palette[i][0],palette[i][1],palette[i][2]);
    }
  printf("  };\n\n");
  }
 

